/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.
 *
 * Copyright (c) 2010-2017, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file     Contexts.cpp
 *  \brief    Classes providing probability descriptions and contexts (also contains context initialization values)
 */

#include "Contexts.h"

#include <algorithm>
#include <cstring>
#include <limits>



const uint8_t ProbModelTables::m_NextState[128][2] =
{
  {   2,  1 },{   0,  3 },{   4,  0 },{   1,  5 },{   6,  2 },{   3,  7 },{   8,  4 },{   5,  9 },
  {  10,  4 },{   5, 11 },{  12,  8 },{   9, 13 },{  14,  8 },{   9, 15 },{  16, 10 },{  11, 17 },
  {  18, 12 },{  13, 19 },{  20, 14 },{  15, 21 },{  22, 16 },{  17, 23 },{  24, 18 },{  19, 25 },
  {  26, 18 },{  19, 27 },{  28, 22 },{  23, 29 },{  30, 22 },{  23, 31 },{  32, 24 },{  25, 33 },
  {  34, 26 },{  27, 35 },{  36, 26 },{  27, 37 },{  38, 30 },{  31, 39 },{  40, 30 },{  31, 41 },
  {  42, 32 },{  33, 43 },{  44, 32 },{  33, 45 },{  46, 36 },{  37, 47 },{  48, 36 },{  37, 49 },
  {  50, 38 },{  39, 51 },{  52, 38 },{  39, 53 },{  54, 42 },{  43, 55 },{  56, 42 },{  43, 57 },
  {  58, 44 },{  45, 59 },{  60, 44 },{  45, 61 },{  62, 46 },{  47, 63 },{  64, 48 },{  49, 65 },
  {  66, 48 },{  49, 67 },{  68, 50 },{  51, 69 },{  70, 52 },{  53, 71 },{  72, 52 },{  53, 73 },
  {  74, 54 },{  55, 75 },{  76, 54 },{  55, 77 },{  78, 56 },{  57, 79 },{  80, 58 },{  59, 81 },
  {  82, 58 },{  59, 83 },{  84, 60 },{  61, 85 },{  86, 60 },{  61, 87 },{  88, 60 },{  61, 89 },
  {  90, 62 },{  63, 91 },{  92, 64 },{  65, 93 },{  94, 64 },{  65, 95 },{  96, 66 },{  67, 97 },
  {  98, 66 },{  67, 99 },{ 100, 66 },{  67,101 },{ 102, 68 },{  69,103 },{ 104, 68 },{  69,105 },
  { 106, 70 },{  71,107 },{ 108, 70 },{  71,109 },{ 110, 70 },{  71,111 },{ 112, 72 },{  73,113 },
  { 114, 72 },{  73,115 },{ 116, 72 },{  73,117 },{ 118, 74 },{  75,119 },{ 120, 74 },{  75,121 },
  { 122, 74 },{  75,123 },{ 124, 76 },{  77,125 },{ 124, 76 },{  77,125 },{ 126,126 },{ 127,127 }
};

const uint32_t ProbModelTables::m_EstFracBits[128] =
{
  0x07b23, 0x085f9, 0x074a0, 0x08cbc, 0x06ee4, 0x09354, 0x067f4, 0x09c1b, 0x060b0, 0x0a62a, 0x05a9c, 0x0af5b, 0x0548d, 0x0b955, 0x04f56, 0x0c2a9,
  0x04a87, 0x0cbf7, 0x045d6, 0x0d5c3, 0x04144, 0x0e01b, 0x03d88, 0x0e937, 0x039e0, 0x0f2cd, 0x03663, 0x0fc9e, 0x03347, 0x10600, 0x03050, 0x10f95,
  0x02d4d, 0x11a02, 0x02ad3, 0x12333, 0x0286e, 0x12cad, 0x02604, 0x136df, 0x02425, 0x13f48, 0x021f4, 0x149c4, 0x0203e, 0x1527b, 0x01e4d, 0x15d00,
  0x01c99, 0x166de, 0x01b18, 0x17017, 0x019a5, 0x17988, 0x01841, 0x18327, 0x016df, 0x18d50, 0x015d9, 0x19547, 0x0147c, 0x1a083, 0x0138e, 0x1a8a3,
  0x01251, 0x1b418, 0x01166, 0x1bd27, 0x01068, 0x1c77b, 0x00f7f, 0x1d18e, 0x00eda, 0x1d91a, 0x00e19, 0x1e254, 0x00d4f, 0x1ec9a, 0x00c90, 0x1f6e0,
  0x00c01, 0x1fef8, 0x00b5f, 0x208b1, 0x00ab6, 0x21362, 0x00a15, 0x21e46, 0x00988, 0x2285d, 0x00934, 0x22ea8, 0x008a8, 0x239b2, 0x0081d, 0x24577,
  0x007c9, 0x24ce6, 0x00763, 0x25663, 0x00710, 0x25e8f, 0x006a0, 0x26a26, 0x00672, 0x26f23, 0x005e8, 0x27ef8, 0x005ba, 0x284b5, 0x0055e, 0x29057,
  0x0050c, 0x29bab, 0x004c1, 0x2a674, 0x004a7, 0x2aa5e, 0x0046f, 0x2b32f, 0x0041f, 0x2c0ad, 0x003e7, 0x2ca8d, 0x003ba, 0x2d323, 0x0010c, 0x3bfbb
};

const BinFracBits ProbModelTables::m_BinFracBits_128[128] =
{
  {{0x07b23, 0x085f9}}, {{0x085f9, 0x07b23}},   {{0x074a0, 0x08cbc}}, {{0x08cbc, 0x074a0}},   {{0x06ee4, 0x09354}}, {{0x09354, 0x06ee4}},   {{0x067f4, 0x09c1b}}, {{0x09c1b, 0x067f4}},
  {{0x060b0, 0x0a62a}}, {{0x0a62a, 0x060b0}},   {{0x05a9c, 0x0af5b}}, {{0x0af5b, 0x05a9c}},   {{0x0548d, 0x0b955}}, {{0x0b955, 0x0548d}},   {{0x04f56, 0x0c2a9}}, {{0x0c2a9, 0x04f56}},
  {{0x04a87, 0x0cbf7}}, {{0x0cbf7, 0x04a87}},   {{0x045d6, 0x0d5c3}}, {{0x0d5c3, 0x045d6}},   {{0x04144, 0x0e01b}}, {{0x0e01b, 0x04144}},   {{0x03d88, 0x0e937}}, {{0x0e937, 0x03d88}},
  {{0x039e0, 0x0f2cd}}, {{0x0f2cd, 0x039e0}},   {{0x03663, 0x0fc9e}}, {{0x0fc9e, 0x03663}},   {{0x03347, 0x10600}}, {{0x10600, 0x03347}},   {{0x03050, 0x10f95}}, {{0x10f95, 0x03050}},
  {{0x02d4d, 0x11a02}}, {{0x11a02, 0x02d4d}},   {{0x02ad3, 0x12333}}, {{0x12333, 0x02ad3}},   {{0x0286e, 0x12cad}}, {{0x12cad, 0x0286e}},   {{0x02604, 0x136df}}, {{0x136df, 0x02604}},
  {{0x02425, 0x13f48}}, {{0x13f48, 0x02425}},   {{0x021f4, 0x149c4}}, {{0x149c4, 0x021f4}},   {{0x0203e, 0x1527b}}, {{0x1527b, 0x0203e}},   {{0x01e4d, 0x15d00}}, {{0x15d00, 0x01e4d}},
  {{0x01c99, 0x166de}}, {{0x166de, 0x01c99}},   {{0x01b18, 0x17017}}, {{0x17017, 0x01b18}},   {{0x019a5, 0x17988}}, {{0x17988, 0x019a5}},   {{0x01841, 0x18327}}, {{0x18327, 0x01841}},
  {{0x016df, 0x18d50}}, {{0x18d50, 0x016df}},   {{0x015d9, 0x19547}}, {{0x19547, 0x015d9}},   {{0x0147c, 0x1a083}}, {{0x1a083, 0x0147c}},   {{0x0138e, 0x1a8a3}}, {{0x1a8a3, 0x0138e}},
  {{0x01251, 0x1b418}}, {{0x1b418, 0x01251}},   {{0x01166, 0x1bd27}}, {{0x1bd27, 0x01166}},   {{0x01068, 0x1c77b}}, {{0x1c77b, 0x01068}},   {{0x00f7f, 0x1d18e}}, {{0x1d18e, 0x00f7f}},
  {{0x00eda, 0x1d91a}}, {{0x1d91a, 0x00eda}},   {{0x00e19, 0x1e254}}, {{0x1e254, 0x00e19}},   {{0x00d4f, 0x1ec9a}}, {{0x1ec9a, 0x00d4f}},   {{0x00c90, 0x1f6e0}}, {{0x1f6e0, 0x00c90}},
  {{0x00c01, 0x1fef8}}, {{0x1fef8, 0x00c01}},   {{0x00b5f, 0x208b1}}, {{0x208b1, 0x00b5f}},   {{0x00ab6, 0x21362}}, {{0x21362, 0x00ab6}},   {{0x00a15, 0x21e46}}, {{0x21e46, 0x00a15}},
  {{0x00988, 0x2285d}}, {{0x2285d, 0x00988}},   {{0x00934, 0x22ea8}}, {{0x22ea8, 0x00934}},   {{0x008a8, 0x239b2}}, {{0x239b2, 0x008a8}},   {{0x0081d, 0x24577}}, {{0x24577, 0x0081d}},
  {{0x007c9, 0x24ce6}}, {{0x24ce6, 0x007c9}},   {{0x00763, 0x25663}}, {{0x25663, 0x00763}},   {{0x00710, 0x25e8f}}, {{0x25e8f, 0x00710}},   {{0x006a0, 0x26a26}}, {{0x26a26, 0x006a0}},
  {{0x00672, 0x26f23}}, {{0x26f23, 0x00672}},   {{0x005e8, 0x27ef8}}, {{0x27ef8, 0x005e8}},   {{0x005ba, 0x284b5}}, {{0x284b5, 0x005ba}},   {{0x0055e, 0x29057}}, {{0x29057, 0x0055e}},
  {{0x0050c, 0x29bab}}, {{0x29bab, 0x0050c}},   {{0x004c1, 0x2a674}}, {{0x2a674, 0x004c1}},   {{0x004a7, 0x2aa5e}}, {{0x2aa5e, 0x004a7}},   {{0x0046f, 0x2b32f}}, {{0x2b32f, 0x0046f}},
  {{0x0041f, 0x2c0ad}}, {{0x2c0ad, 0x0041f}},   {{0x003e7, 0x2ca8d}}, {{0x2ca8d, 0x003e7}},   {{0x003ba, 0x2d323}}, {{0x2d323, 0x003ba}},   {{0x0010c, 0x3bfbb}}, {{0x3bfbb, 0x0010c}}
};

#if JEM_TOOLS
const BinFracBits ProbModelTables::m_BinFracBits_256[256] =
{
  {{0x0005c, 0x48000}}, {{0x00116, 0x3b520}}, {{0x001d0, 0x356cb}}, {{0x0028b, 0x318a9}}, {{0x00346, 0x2ea40}}, {{0x00403, 0x2c531}}, {{0x004c0, 0x2a658}}, {{0x0057e, 0x28beb}},
  {{0x0063c, 0x274ce}}, {{0x006fc, 0x26044}}, {{0x007bc, 0x24dc9}}, {{0x0087d, 0x23cfc}}, {{0x0093f, 0x22d96}}, {{0x00a01, 0x21f60}}, {{0x00ac4, 0x2122e}}, {{0x00b89, 0x205dd}},
  {{0x00c4e, 0x1fa51}}, {{0x00d13, 0x1ef74}}, {{0x00dda, 0x1e531}}, {{0x00ea2, 0x1db78}}, {{0x00f6a, 0x1d23c}}, {{0x01033, 0x1c970}}, {{0x010fd, 0x1c10b}}, {{0x011c8, 0x1b903}},
  {{0x01294, 0x1b151}}, {{0x01360, 0x1a9ee}}, {{0x0142e, 0x1a2d4}}, {{0x014fc, 0x19bfc}}, {{0x015cc, 0x19564}}, {{0x0169c, 0x18f06}}, {{0x0176d, 0x188de}}, {{0x0183f, 0x182e8}},
  {{0x01912, 0x17d23}}, {{0x019e6, 0x1778a}}, {{0x01abb, 0x1721c}}, {{0x01b91, 0x16cd5}}, {{0x01c68, 0x167b4}}, {{0x01d40, 0x162b6}}, {{0x01e19, 0x15dda}}, {{0x01ef3, 0x1591e}},
  {{0x01fcd, 0x15480}}, {{0x020a9, 0x14fff}}, {{0x02186, 0x14b99}}, {{0x02264, 0x1474e}}, {{0x02343, 0x1431b}}, {{0x02423, 0x13f01}}, {{0x02504, 0x13afd}}, {{0x025e6, 0x1370f}},
  {{0x026ca, 0x13336}}, {{0x027ae, 0x12f71}}, {{0x02894, 0x12bc0}}, {{0x0297a, 0x12821}}, {{0x02a62, 0x12494}}, {{0x02b4b, 0x12118}}, {{0x02c35, 0x11dac}}, {{0x02d20, 0x11a51}},
  {{0x02e0c, 0x11704}}, {{0x02efa, 0x113c7}}, {{0x02fe9, 0x11098}}, {{0x030d9, 0x10d77}}, {{0x031ca, 0x10a63}}, {{0x032bc, 0x1075c}}, {{0x033b0, 0x10461}}, {{0x034a5, 0x10173}},
  {{0x0359b, 0x0fe90}}, {{0x03693, 0x0fbb9}}, {{0x0378c, 0x0f8ed}}, {{0x03886, 0x0f62b}}, {{0x03981, 0x0f374}}, {{0x03a7e, 0x0f0c7}}, {{0x03b7c, 0x0ee23}}, {{0x03c7c, 0x0eb89}},
  {{0x03d7d, 0x0e8f9}}, {{0x03e7f, 0x0e671}}, {{0x03f83, 0x0e3f2}}, {{0x04088, 0x0e17c}}, {{0x0418e, 0x0df0e}}, {{0x04297, 0x0dca8}}, {{0x043a0, 0x0da4a}}, {{0x044ab, 0x0d7f3}},
  {{0x045b8, 0x0d5a5}}, {{0x046c6, 0x0d35d}}, {{0x047d6, 0x0d11c}}, {{0x048e7, 0x0cee3}}, {{0x049fa, 0x0ccb0}}, {{0x04b0e, 0x0ca84}}, {{0x04c24, 0x0c85e}}, {{0x04d3c, 0x0c63f}},
  {{0x04e55, 0x0c426}}, {{0x04f71, 0x0c212}}, {{0x0508d, 0x0c005}}, {{0x051ac, 0x0bdfe}}, {{0x052cc, 0x0bbfc}}, {{0x053ee, 0x0b9ff}}, {{0x05512, 0x0b808}}, {{0x05638, 0x0b617}},
  {{0x0575f, 0x0b42a}}, {{0x05888, 0x0b243}}, {{0x059b4, 0x0b061}}, {{0x05ae1, 0x0ae83}}, {{0x05c10, 0x0acaa}}, {{0x05d41, 0x0aad6}}, {{0x05e74, 0x0a907}}, {{0x05fa9, 0x0a73c}},
  {{0x060e0, 0x0a575}}, {{0x06219, 0x0a3b3}}, {{0x06354, 0x0a1f5}}, {{0x06491, 0x0a03b}}, {{0x065d1, 0x09e85}}, {{0x06712, 0x09cd4}}, {{0x06856, 0x09b26}}, {{0x0699c, 0x0997c}},
  {{0x06ae4, 0x097d6}}, {{0x06c2f, 0x09634}}, {{0x06d7c, 0x09495}}, {{0x06ecb, 0x092fa}}, {{0x0701d, 0x09162}}, {{0x07171, 0x08fce}}, {{0x072c7, 0x08e3e}}, {{0x07421, 0x08cb0}},
  {{0x0757c, 0x08b26}}, {{0x076da, 0x089a0}}, {{0x0783b, 0x0881c}}, {{0x0799f, 0x0869c}}, {{0x07b05, 0x0851f}}, {{0x07c6e, 0x083a4}}, {{0x07dd9, 0x0822d}}, {{0x07f48, 0x080b9}},
  {{0x080b9, 0x07f48}}, {{0x0822d, 0x07dd9}}, {{0x083a4, 0x07c6e}}, {{0x0851f, 0x07b05}}, {{0x0869c, 0x0799f}}, {{0x0881c, 0x0783b}}, {{0x089a0, 0x076da}}, {{0x08b26, 0x0757c}},
  {{0x08cb0, 0x07421}}, {{0x08e3e, 0x072c7}}, {{0x08fce, 0x07171}}, {{0x09162, 0x0701d}}, {{0x092fa, 0x06ecb}}, {{0x09495, 0x06d7c}}, {{0x09634, 0x06c2f}}, {{0x097d6, 0x06ae4}},
  {{0x0997c, 0x0699c}}, {{0x09b26, 0x06856}}, {{0x09cd4, 0x06712}}, {{0x09e85, 0x065d1}}, {{0x0a03b, 0x06491}}, {{0x0a1f5, 0x06354}}, {{0x0a3b3, 0x06219}}, {{0x0a575, 0x060e0}},
  {{0x0a73c, 0x05fa9}}, {{0x0a907, 0x05e74}}, {{0x0aad6, 0x05d41}}, {{0x0acaa, 0x05c10}}, {{0x0ae83, 0x05ae1}}, {{0x0b061, 0x059b4}}, {{0x0b243, 0x05888}}, {{0x0b42a, 0x0575f}},
  {{0x0b617, 0x05638}}, {{0x0b808, 0x05512}}, {{0x0b9ff, 0x053ee}}, {{0x0bbfc, 0x052cc}}, {{0x0bdfe, 0x051ac}}, {{0x0c005, 0x0508d}}, {{0x0c212, 0x04f71}}, {{0x0c426, 0x04e55}},
  {{0x0c63f, 0x04d3c}}, {{0x0c85e, 0x04c24}}, {{0x0ca84, 0x04b0e}}, {{0x0ccb0, 0x049fa}}, {{0x0cee3, 0x048e7}}, {{0x0d11c, 0x047d6}}, {{0x0d35d, 0x046c6}}, {{0x0d5a5, 0x045b8}},
  {{0x0d7f3, 0x044ab}}, {{0x0da4a, 0x043a0}}, {{0x0dca8, 0x04297}}, {{0x0df0e, 0x0418e}}, {{0x0e17c, 0x04088}}, {{0x0e3f2, 0x03f83}}, {{0x0e671, 0x03e7f}}, {{0x0e8f9, 0x03d7d}},
  {{0x0eb89, 0x03c7c}}, {{0x0ee23, 0x03b7c}}, {{0x0f0c7, 0x03a7e}}, {{0x0f374, 0x03981}}, {{0x0f62b, 0x03886}}, {{0x0f8ed, 0x0378c}}, {{0x0fbb9, 0x03693}}, {{0x0fe90, 0x0359b}},
  {{0x10173, 0x034a5}}, {{0x10461, 0x033b0}}, {{0x1075c, 0x032bc}}, {{0x10a63, 0x031ca}}, {{0x10d77, 0x030d9}}, {{0x11098, 0x02fe9}}, {{0x113c7, 0x02efa}}, {{0x11704, 0x02e0c}},
  {{0x11a51, 0x02d20}}, {{0x11dac, 0x02c35}}, {{0x12118, 0x02b4b}}, {{0x12494, 0x02a62}}, {{0x12821, 0x0297a}}, {{0x12bc0, 0x02894}}, {{0x12f71, 0x027ae}}, {{0x13336, 0x026ca}},
  {{0x1370f, 0x025e6}}, {{0x13afd, 0x02504}}, {{0x13f01, 0x02423}}, {{0x1431b, 0x02343}}, {{0x1474e, 0x02264}}, {{0x14b99, 0x02186}}, {{0x14fff, 0x020a9}}, {{0x15480, 0x01fcd}},
  {{0x1591e, 0x01ef3}}, {{0x15dda, 0x01e19}}, {{0x162b6, 0x01d40}}, {{0x167b4, 0x01c68}}, {{0x16cd5, 0x01b91}}, {{0x1721c, 0x01abb}}, {{0x1778a, 0x019e6}}, {{0x17d23, 0x01912}},
  {{0x182e8, 0x0183f}}, {{0x188de, 0x0176d}}, {{0x18f06, 0x0169c}}, {{0x19564, 0x015cc}}, {{0x19bfc, 0x014fc}}, {{0x1a2d4, 0x0142e}}, {{0x1a9ee, 0x01360}}, {{0x1b151, 0x01294}},
  {{0x1b903, 0x011c8}}, {{0x1c10b, 0x010fd}}, {{0x1c970, 0x01033}}, {{0x1d23c, 0x00f6a}}, {{0x1db78, 0x00ea2}}, {{0x1e531, 0x00dda}}, {{0x1ef74, 0x00d13}}, {{0x1fa51, 0x00c4e}},
  {{0x205dd, 0x00b89}}, {{0x2122e, 0x00ac4}}, {{0x21f60, 0x00a01}}, {{0x22d96, 0x0093f}}, {{0x23cfc, 0x0087d}}, {{0x24dc9, 0x007bc}}, {{0x26044, 0x006fc}}, {{0x274ce, 0x0063c}},
  {{0x28beb, 0x0057e}}, {{0x2a658, 0x004c0}}, {{0x2c531, 0x00403}}, {{0x2ea40, 0x00346}}, {{0x318a9, 0x0028b}}, {{0x356cb, 0x001d0}}, {{0x3b520, 0x00116}}, {{0x48000, 0x0005c}}
};

#endif
const uint32_t ProbModelTables::m_EstFracProb[128] =
{
  0x041b5, 0x03df6, 0x04410, 0x03bbc, 0x04636, 0x039a3, 0x048e6, 0x036f6, 0x04bd3, 0x0340c, 0x04e5d, 0x03185, 0x050fa, 0x02eeb, 0x0534b, 0x02c9b,
  0x0557e, 0x02a6a, 0x057b1, 0x02839, 0x059e4, 0x02608, 0x05bba, 0x02433, 0x05d8f, 0x0225e, 0x05f58, 0x02097, 0x060f7, 0x01efa, 0x06288, 0x01d69,
  0x06427, 0x01bcb, 0x06581, 0x01a72, 0x066d4, 0x0191f, 0x0682f, 0x017c6, 0x0693e, 0x016b7, 0x06a80, 0x01576, 0x06b7e, 0x01478, 0x06ca1, 0x01356,
  0x06da2, 0x01255, 0x06e88, 0x01170, 0x06f67, 0x01092, 0x0703e, 0x00fba, 0x07116, 0x00ee3, 0x071b7, 0x00e42, 0x0728f, 0x00d6a, 0x07323, 0x00cd6,
  0x073e9, 0x00c11, 0x0747d, 0x00b7d, 0x0751e, 0x00add, 0x075b2, 0x00a49, 0x0761b, 0x009e0, 0x07697, 0x00964, 0x07719, 0x008e2, 0x07794, 0x00867,
  0x077f1, 0x0080b, 0x0785b, 0x007a1, 0x078c9, 0x00733, 0x07932, 0x006ca, 0x0798f, 0x0066d, 0x079c6, 0x00636, 0x07a23, 0x005da, 0x07a7f, 0x0057d,
  0x07ab7, 0x00546, 0x07afb, 0x00502, 0x07b32, 0x004cb, 0x07b7d, 0x00480, 0x07b9c, 0x00461, 0x07bf8, 0x00405, 0x07c17, 0x003e6, 0x07c55, 0x003a9,
  0x07c8c, 0x00371, 0x07cbf, 0x0033f, 0x07cd0, 0x0032e, 0x07cf6, 0x00308, 0x07d2c, 0x002d1, 0x07d52, 0x002ab, 0x07d71, 0x0028c, 0x07f46, 0x000b5
};

const uint8_t ProbModelTables::m_LPSTable_64_4[64][4] =
{
  { 128, 176, 208, 240 },
  { 128, 167, 197, 227 },
  { 128, 158, 187, 216 },
  { 123, 150, 178, 205 },
  { 116, 142, 169, 195 },
  { 111, 135, 160, 185 },
  { 105, 128, 152, 175 },
  { 100, 122, 144, 166 },
  {  95, 116, 137, 158 },
  {  90, 110, 130, 150 },
  {  85, 104, 123, 142 },
  {  81,  99, 117, 135 },
  {  77,  94, 111, 128 },
  {  73,  89, 105, 122 },
  {  69,  85, 100, 116 },
  {  66,  80,  95, 110 },
  {  62,  76,  90, 104 },
  {  59,  72,  86,  99 },
  {  56,  69,  81,  94 },
  {  53,  65,  77,  89 },
  {  51,  62,  73,  85 },
  {  48,  59,  69,  80 },
  {  46,  56,  66,  76 },
  {  43,  53,  63,  72 },
  {  41,  50,  59,  69 },
  {  39,  48,  56,  65 },
  {  37,  45,  54,  62 },
  {  35,  43,  51,  59 },
  {  33,  41,  48,  56 },
  {  32,  39,  46,  53 },
  {  30,  37,  43,  50 },
  {  29,  35,  41,  48 },
  {  27,  33,  39,  45 },
  {  26,  31,  37,  43 },
  {  24,  30,  35,  41 },
  {  23,  28,  33,  39 },
  {  22,  27,  32,  37 },
  {  21,  26,  30,  35 },
  {  20,  24,  29,  33 },
  {  19,  23,  27,  31 },
  {  18,  22,  26,  30 },
  {  17,  21,  25,  28 },
  {  16,  20,  23,  27 },
  {  15,  19,  22,  25 },
  {  14,  18,  21,  24 },
  {  14,  17,  20,  23 },
  {  13,  16,  19,  22 },
  {  12,  15,  18,  21 },
  {  12,  14,  17,  20 },
  {  11,  14,  16,  19 },
  {  11,  13,  15,  18 },
  {  10,  12,  15,  17 },
  {  10,  12,  14,  16 },
  {   9,  11,  13,  15 },
  {   9,  11,  12,  14 },
  {   8,  10,  12,  14 },
  {   8,   9,  11,  13 },
  {   7,   9,  11,  12 },
  {   7,   9,  10,  12 },
  {   7,   8,  10,  11 },
  {   6,   8,   9,  11 },
  {   6,   7,   9,  10 },
  {   6,   7,   8,   9 },
  {   2,   2,   2,   2 }
};

#if JEM_TOOLS
const uint16_t ProbModelTables::m_LPSTable_512_64[512][64] =
{
  {  253,  257,  261,  265,  269,  273,  277,  281,  285,  289,  293,  297,  301,  305,  309,  313,  317,  321,  325,  329,  333,  337,  341,  345,  349,  353,  357,  361,  365,  369,  373,  377,  381,  385,  389,  393,  397,  401,  405,  409,  413,  417,  421,  425,  429,  433,  437,  441,  445,  449,  453,  457,  461,  465,  469,  473,  477,  481,  485,  489,  493,  497,  501,  505,},
  {  253,  257,  261,  265,  269,  273,  277,  281,  285,  289,  293,  297,  301,  305,  309,  313,  317,  321,  325,  329,  333,  337,  341,  345,  349,  353,  357,  361,  365,  369,  373,  377,  381,  385,  389,  393,  397,  401,  405,  409,  413,  417,  421,  425,  429,  433,  437,  441,  445,  449,  453,  457,  461,  465,  469,  473,  477,  481,  485,  489,  493,  497,  501,  505,},
  {  253,  257,  261,  265,  269,  273,  277,  281,  285,  289,  293,  297,  301,  305,  309,  313,  317,  321,  325,  329,  333,  337,  341,  345,  349,  353,  357,  361,  365,  369,  373,  377,  381,  385,  389,  393,  397,  401,  405,  409,  413,  417,  421,  425,  429,  433,  437,  441,  445,  449,  453,  457,  461,  465,  469,  473,  477,  481,  485,  489,  493,  497,  501,  505,},
  {  253,  257,  261,  265,  269,  273,  277,  281,  285,  289,  293,  297,  301,  305,  309,  313,  317,  321,  325,  329,  333,  337,  341,  345,  349,  353,  357,  361,  365,  369,  373,  377,  381,  385,  389,  393,  397,  401,  405,  409,  413,  417,  421,  425,  429,  433,  437,  441,  445,  449,  453,  457,  461,  465,  469,  473,  477,  481,  485,  489,  493,  497,  501,  505,},
  {  253,  257,  261,  265,  269,  273,  277,  281,  285,  289,  293,  297,  301,  305,  309,  313,  317,  321,  325,  329,  333,  337,  341,  345,  349,  353,  357,  361,  365,  369,  373,  377,  381,  385,  389,  393,  396,  400,  404,  408,  412,  416,  420,  424,  428,  432,  436,  440,  444,  448,  452,  456,  460,  464,  468,  472,  476,  480,  484,  488,  492,  496,  500,  504,},
  {  253,  257,  261,  265,  269,  273,  277,  281,  285,  289,  293,  297,  301,  305,  309,  313,  317,  321,  324,  328,  332,  336,  340,  344,  348,  352,  356,  360,  364,  368,  372,  376,  380,  384,  388,  392,  396,  400,  404,  408,  412,  415,  419,  423,  427,  431,  435,  439,  443,  447,  451,  455,  459,  463,  467,  471,  475,  479,  483,  487,  491,  495,  499,  503,},
  {  253,  257,  261,  265,  269,  272,  276,  280,  284,  288,  292,  296,  300,  304,  308,  312,  316,  320,  324,  328,  332,  336,  340,  344,  348,  351,  355,  359,  363,  367,  371,  375,  379,  383,  387,  391,  395,  399,  403,  407,  411,  415,  419,  423,  427,  430,  434,  438,  442,  446,  450,  454,  458,  462,  466,  470,  474,  478,  482,  486,  490,  494,  498,  502,},
  {  252,  256,  260,  264,  268,  272,  276,  280,  284,  288,  292,  296,  300,  303,  307,  311,  315,  319,  323,  327,  331,  335,  339,  343,  347,  351,  355,  359,  363,  367,  370,  374,  378,  382,  386,  390,  394,  398,  402,  406,  410,  414,  418,  422,  426,  430,  434,  437,  441,  445,  449,  453,  457,  461,  465,  469,  473,  477,  481,  485,  489,  493,  497,  501,},
  {  252,  256,  260,  264,  267,  271,  275,  279,  283,  287,  291,  295,  299,  303,  307,  311,  315,  319,  323,  326,  330,  334,  338,  342,  346,  350,  354,  358,  362,  366,  370,  374,  378,  382,  385,  389,  393,  397,  401,  405,  409,  413,  417,  421,  425,  429,  433,  437,  441,  444,  448,  452,  456,  460,  464,  468,  472,  476,  480,  484,  488,  492,  496,  500,},
  {  251,  255,  259,  263,  267,  271,  275,  279,  283,  287,  291,  294,  298,  302,  306,  310,  314,  318,  322,  326,  330,  334,  338,  342,  345,  349,  353,  357,  361,  365,  369,  373,  377,  381,  385,  389,  393,  397,  400,  404,  408,  412,  416,  420,  424,  428,  432,  436,  440,  444,  448,  451,  455,  459,  463,  467,  471,  475,  479,  483,  487,  491,  495,  499,},
  {  251,  255,  259,  263,  266,  270,  274,  278,  282,  286,  290,  294,  298,  302,  306,  310,  313,  317,  321,  325,  329,  333,  337,  341,  345,  349,  353,  357,  360,  364,  368,  372,  376,  380,  384,  388,  392,  396,  400,  404,  407,  411,  415,  419,  423,  427,  431,  435,  439,  443,  447,  451,  454,  458,  462,  466,  470,  474,  478,  482,  486,  490,  494,  498,},
  {  250,  254,  258,  262,  266,  270,  274,  278,  282,  285,  289,  293,  297,  301,  305,  309,  313,  317,  321,  325,  328,  332,  336,  340,  344,  348,  352,  356,  360,  364,  368,  371,  375,  379,  383,  387,  391,  395,  399,  403,  407,  411,  414,  418,  422,  426,  430,  434,  438,  442,  446,  450,  454,  457,  461,  465,  469,  473,  477,  481,  485,  489,  493,  497,},
  {  250,  254,  258,  261,  265,  269,  273,  277,  281,  285,  289,  293,  297,  300,  304,  308,  312,  316,  320,  324,  328,  332,  336,  340,  343,  347,  351,  355,  359,  363,  367,  371,  375,  379,  382,  386,  390,  394,  398,  402,  406,  410,  414,  418,  421,  425,  429,  433,  437,  441,  445,  449,  453,  457,  460,  464,  468,  472,  476,  480,  484,  488,  492,  496,},
  {  249,  253,  257,  261,  265,  269,  273,  277,  280,  284,  288,  292,  296,  300,  304,  308,  312,  315,  319,  323,  327,  331,  335,  339,  343,  347,  351,  354,  358,  362,  366,  370,  374,  378,  382,  386,  389,  393,  397,  401,  405,  409,  413,  417,  421,  425,  428,  432,  436,  440,  444,  448,  452,  456,  460,  463,  467,  471,  475,  479,  483,  487,  491,  495,},
  {  249,  253,  257,  260,  264,  268,  272,  276,  280,  284,  288,  292,  295,  299,  303,  307,  311,  315,  319,  323,  326,  330,  334,  338,  342,  346,  350,  354,  358,  361,  365,  369,  373,  377,  381,  385,  389,  393,  396,  400,  404,  408,  412,  416,  420,  424,  428,  431,  435,  439,  443,  447,  451,  455,  459,  463,  466,  470,  474,  478,  482,  486,  490,  494,},
  {  248,  252,  256,  260,  264,  268,  272,  275,  279,  283,  287,  291,  295,  299,  303,  306,  310,  314,  318,  322,  326,  330,  334,  337,  341,  345,  349,  353,  357,  361,  365,  368,  372,  376,  380,  384,  388,  392,  396,  400,  403,  407,  411,  415,  419,  423,  427,  431,  434,  438,  442,  446,  450,  454,  458,  462,  465,  469,  473,  477,  481,  485,  489,  493,},
  {  248,  252,  255,  259,  263,  267,  271,  275,  279,  283,  286,  290,  294,  298,  302,  306,  310,  314,  317,  321,  325,  329,  333,  337,  341,  345,  348,  352,  356,  360,  364,  368,  372,  375,  379,  383,  387,  391,  395,  399,  403,  406,  410,  414,  418,  422,  426,  430,  434,  437,  441,  445,  449,  453,  457,  461,  465,  468,  472,  476,  480,  484,  488,  492,},
  {  247,  251,  255,  259,  263,  267,  270,  274,  278,  282,  286,  290,  294,  297,  301,  305,  309,  313,  317,  321,  325,  328,  332,  336,  340,  344,  348,  352,  355,  359,  363,  367,  371,  375,  379,  382,  386,  390,  394,  398,  402,  406,  410,  413,  417,  421,  425,  429,  433,  437,  440,  444,  448,  452,  456,  460,  464,  467,  471,  475,  479,  483,  487,  491,},
  {  247,  251,  254,  258,  262,  266,  270,  274,  278,  281,  285,  289,  293,  297,  301,  305,  308,  312,  316,  320,  324,  328,  332,  335,  339,  343,  347,  351,  355,  359,  362,  366,  370,  374,  378,  382,  386,  389,  393,  397,  401,  405,  409,  413,  416,  420,  424,  428,  432,  436,  440,  443,  447,  451,  455,  459,  463,  467,  470,  474,  478,  482,  486,  490,},
  {  246,  250,  254,  258,  262,  265,  269,  273,  277,  281,  285,  289,  292,  296,  300,  304,  308,  312,  316,  319,  323,  327,  331,  335,  339,  342,  346,  350,  354,  358,  362,  366,  369,  373,  377,  381,  385,  389,  392,  396,  400,  404,  408,  412,  416,  419,  423,  427,  431,  435,  439,  442,  446,  450,  454,  458,  462,  466,  469,  473,  477,  481,  485,  489,},
  {  246,  250,  253,  257,  261,  265,  269,  273,  276,  280,  284,  288,  292,  296,  300,  303,  307,  311,  315,  319,  323,  326,  330,  334,  338,  342,  346,  349,  353,  357,  361,  365,  369,  372,  376,  380,  384,  388,  392,  396,  399,  403,  407,  411,  415,  419,  422,  426,  430,  434,  438,  442,  445,  449,  453,  457,  461,  465,  468,  472,  476,  480,  484,  488,},
  {  245,  249,  253,  257,  261,  264,  268,  272,  276,  280,  284,  287,  291,  295,  299,  303,  307,  310,  314,  318,  322,  326,  330,  333,  337,  341,  345,  349,  353,  356,  360,  364,  368,  372,  376,  379,  383,  387,  391,  395,  399,  402,  406,  410,  414,  418,  422,  425,  429,  433,  437,  441,  445,  448,  452,  456,  460,  464,  468,  471,  475,  479,  483,  487,},
  {  245,  249,  252,  256,  260,  264,  268,  272,  275,  279,  283,  287,  291,  294,  298,  302,  306,  310,  314,  317,  321,  325,  329,  333,  337,  340,  344,  348,  352,  356,  359,  363,  367,  371,  375,  379,  382,  386,  390,  394,  398,  402,  405,  409,  413,  417,  421,  424,  428,  432,  436,  440,  444,  447,  451,  455,  459,  463,  467,  470,  474,  478,  482,  486,},
  {  244,  248,  252,  256,  260,  263,  267,  271,  275,  279,  282,  286,  290,  294,  298,  301,  305,  309,  313,  317,  321,  324,  328,  332,  336,  340,  343,  347,  351,  355,  359,  363,  366,  370,  374,  378,  382,  385,  389,  393,  397,  401,  405,  408,  412,  416,  420,  424,  427,  431,  435,  439,  443,  447,  450,  454,  458,  462,  466,  469,  473,  477,  481,  485,},
  {  244,  248,  251,  255,  259,  263,  267,  270,  274,  278,  282,  286,  289,  293,  297,  301,  305,  308,  312,  316,  320,  324,  328,  331,  335,  339,  343,  347,  350,  354,  358,  362,  366,  369,  373,  377,  381,  385,  388,  392,  396,  400,  404,  408,  411,  415,  419,  423,  427,  430,  434,  438,  442,  446,  449,  453,  457,  461,  465,  468,  472,  476,  480,  484,},
  {  243,  247,  251,  255,  258,  262,  266,  270,  274,  277,  281,  285,  289,  293,  296,  300,  304,  308,  312,  315,  319,  323,  327,  331,  334,  338,  342,  346,  350,  353,  357,  361,  365,  369,  372,  376,  380,  384,  388,  391,  395,  399,  403,  407,  410,  414,  418,  422,  426,  429,  433,  437,  441,  445,  448,  452,  456,  460,  464,  467,  471,  475,  479,  483,},
  {  243,  247,  250,  254,  258,  262,  266,  269,  273,  277,  281,  284,  288,  292,  296,  300,  303,  307,  311,  315,  319,  322,  326,  330,  334,  338,  341,  345,  349,  353,  357,  360,  364,  368,  372,  376,  379,  383,  387,  391,  394,  398,  402,  406,  410,  413,  417,  421,  425,  429,  432,  436,  440,  444,  448,  451,  455,  459,  463,  467,  470,  474,  478,  482,},
  {  242,  246,  250,  254,  257,  261,  265,  269,  273,  276,  280,  284,  288,  291,  295,  299,  303,  307,  310,  314,  318,  322,  326,  329,  333,  337,  341,  344,  348,  352,  356,  360,  363,  367,  371,  375,  379,  382,  386,  390,  394,  397,  401,  405,  409,  413,  416,  420,  424,  428,  432,  435,  439,  443,  447,  450,  454,  458,  462,  466,  469,  473,  477,  481,},
  {  242,  246,  249,  253,  257,  261,  264,  268,  272,  276,  280,  283,  287,  291,  295,  298,  302,  306,  310,  314,  317,  321,  325,  329,  332,  336,  340,  344,  348,  351,  355,  359,  363,  366,  370,  374,  378,  382,  385,  389,  393,  397,  400,  404,  408,  412,  416,  419,  423,  427,  431,  434,  438,  442,  446,  450,  453,  457,  461,  465,  468,  472,  476,  480,},
  {  241,  245,  249,  253,  256,  260,  264,  268,  271,  275,  279,  283,  286,  290,  294,  298,  302,  305,  309,  313,  317,  320,  324,  328,  332,  335,  339,  343,  347,  351,  354,  358,  362,  366,  369,  373,  377,  381,  384,  388,  392,  396,  400,  403,  407,  411,  415,  418,  422,  426,  430,  433,  437,  441,  445,  449,  452,  456,  460,  464,  467,  471,  475,  479,},
  {  241,  245,  248,  252,  256,  260,  263,  267,  271,  275,  278,  282,  286,  290,  293,  297,  301,  305,  308,  312,  316,  320,  324,  327,  331,  335,  339,  342,  346,  350,  354,  357,  361,  365,  369,  372,  376,  380,  384,  387,  391,  395,  399,  403,  406,  410,  414,  418,  421,  425,  429,  433,  436,  440,  444,  448,  451,  455,  459,  463,  466,  470,  474,  478,},
  {  240,  244,  248,  252,  255,  259,  263,  267,  270,  274,  278,  282,  285,  289,  293,  297,  300,  304,  308,  312,  315,  319,  323,  327,  330,  334,  338,  342,  345,  349,  353,  357,  360,  364,  368,  372,  375,  379,  383,  387,  390,  394,  398,  402,  405,  409,  413,  417,  420,  424,  428,  432,  435,  439,  443,  447,  450,  454,  458,  462,  465,  469,  473,  477,},
  {  240,  243,  247,  251,  255,  258,  262,  266,  270,  273,  277,  281,  285,  288,  292,  296,  300,  303,  307,  311,  315,  318,  322,  326,  330,  333,  337,  341,  345,  348,  352,  356,  360,  363,  367,  371,  375,  378,  382,  386,  390,  393,  397,  401,  405,  408,  412,  416,  420,  423,  427,  431,  435,  438,  442,  446,  450,  453,  457,  461,  465,  468,  472,  476,},
  {  239,  243,  247,  250,  254,  258,  262,  265,  269,  273,  277,  280,  284,  288,  292,  295,  299,  303,  307,  310,  314,  318,  321,  325,  329,  333,  336,  340,  344,  348,  351,  355,  359,  363,  366,  370,  374,  378,  381,  385,  389,  393,  396,  400,  404,  407,  411,  415,  419,  422,  426,  430,  434,  437,  441,  445,  449,  452,  456,  460,  464,  467,  471,  475,},
  {  239,  242,  246,  250,  254,  257,  261,  265,  269,  272,  276,  280,  284,  287,  291,  295,  298,  302,  306,  310,  313,  317,  321,  325,  328,  332,  336,  339,  343,  347,  351,  354,  358,  362,  366,  369,  373,  377,  381,  384,  388,  392,  395,  399,  403,  407,  410,  414,  418,  422,  425,  429,  433,  436,  440,  444,  448,  451,  455,  459,  463,  466,  470,  474,},
  {  238,  242,  246,  249,  253,  257,  261,  264,  268,  272,  275,  279,  283,  287,  290,  294,  298,  302,  305,  309,  313,  316,  320,  324,  328,  331,  335,  339,  342,  346,  350,  354,  357,  361,  365,  369,  372,  376,  380,  383,  387,  391,  395,  398,  402,  406,  409,  413,  417,  421,  424,  428,  432,  436,  439,  443,  447,  450,  454,  458,  462,  465,  469,  473,},
  {  238,  241,  245,  249,  253,  256,  260,  264,  267,  271,  275,  279,  282,  286,  290,  293,  297,  301,  305,  308,  312,  316,  319,  323,  327,  331,  334,  338,  342,  345,  349,  353,  357,  360,  364,  368,  371,  375,  379,  383,  386,  390,  394,  397,  401,  405,  409,  412,  416,  420,  423,  427,  431,  435,  438,  442,  446,  449,  453,  457,  461,  464,  468,  472,},
  {  237,  241,  245,  248,  252,  256,  259,  263,  267,  271,  274,  278,  282,  285,  289,  293,  297,  300,  304,  308,  311,  315,  319,  323,  326,  330,  334,  337,  341,  345,  348,  352,  356,  360,  363,  367,  371,  374,  378,  382,  386,  389,  393,  397,  400,  404,  408,  411,  415,  419,  423,  426,  430,  434,  437,  441,  445,  449,  452,  456,  460,  463,  467,  471,},
  {  237,  240,  244,  248,  252,  255,  259,  263,  266,  270,  274,  277,  281,  285,  289,  292,  296,  300,  303,  307,  311,  314,  318,  322,  326,  329,  333,  337,  340,  344,  348,  351,  355,  359,  363,  366,  370,  374,  377,  381,  385,  388,  392,  396,  400,  403,  407,  411,  414,  418,  422,  425,  429,  433,  437,  440,  444,  448,  451,  455,  459,  462,  466,  470,},
  {  236,  240,  244,  247,  251,  255,  258,  262,  266,  269,  273,  277,  281,  284,  288,  292,  295,  299,  303,  306,  310,  314,  317,  321,  325,  329,  332,  336,  340,  343,  347,  351,  354,  358,  362,  365,  369,  373,  377,  380,  384,  388,  391,  395,  399,  402,  406,  410,  413,  417,  421,  425,  428,  432,  436,  439,  443,  447,  450,  454,  458,  461,  465,  469,},
  {  236,  239,  243,  247,  250,  254,  258,  262,  265,  269,  273,  276,  280,  284,  287,  291,  295,  298,  302,  306,  309,  313,  317,  320,  324,  328,  332,  335,  339,  343,  346,  350,  354,  357,  361,  365,  368,  372,  376,  379,  383,  387,  390,  394,  398,  402,  405,  409,  413,  416,  420,  424,  427,  431,  435,  438,  442,  446,  449,  453,  457,  460,  464,  468,},
  {  235,  239,  243,  246,  250,  254,  257,  261,  265,  268,  272,  276,  279,  283,  287,  290,  294,  298,  301,  305,  309,  312,  316,  320,  323,  327,  331,  334,  338,  342,  346,  349,  353,  357,  360,  364,  368,  371,  375,  379,  382,  386,  390,  393,  397,  401,  404,  408,  412,  415,  419,  423,  426,  430,  434,  437,  441,  445,  448,  452,  456,  459,  463,  467,},
  {  235,  238,  242,  246,  249,  253,  257,  260,  264,  268,  271,  275,  279,  282,  286,  290,  293,  297,  301,  304,  308,  312,  315,  319,  323,  326,  330,  334,  337,  341,  345,  348,  352,  356,  359,  363,  367,  370,  374,  378,  381,  385,  389,  392,  396,  400,  403,  407,  411,  414,  418,  422,  425,  429,  433,  436,  440,  444,  447,  451,  455,  458,  462,  466,},
  {  234,  238,  242,  245,  249,  253,  256,  260,  264,  267,  271,  275,  278,  282,  285,  289,  293,  296,  300,  304,  307,  311,  315,  318,  322,  326,  329,  333,  337,  340,  344,  348,  351,  355,  359,  362,  366,  370,  373,  377,  381,  384,  388,  392,  395,  399,  403,  406,  410,  414,  417,  421,  425,  428,  432,  436,  439,  443,  447,  450,  454,  458,  461,  465,},
  {  234,  237,  241,  245,  248,  252,  256,  259,  263,  267,  270,  274,  278,  281,  285,  289,  292,  296,  299,  303,  307,  310,  314,  318,  321,  325,  329,  332,  336,  340,  343,  347,  351,  354,  358,  362,  365,  369,  373,  376,  380,  383,  387,  391,  394,  398,  402,  405,  409,  413,  416,  420,  424,  427,  431,  435,  438,  442,  446,  449,  453,  457,  460,  464,},
  {  233,  237,  241,  244,  248,  251,  255,  259,  262,  266,  270,  273,  277,  281,  284,  288,  292,  295,  299,  302,  306,  310,  313,  317,  321,  324,  328,  332,  335,  339,  343,  346,  350,  354,  357,  361,  364,  368,  372,  375,  379,  383,  386,  390,  394,  397,  401,  405,  408,  412,  415,  419,  423,  426,  430,  434,  437,  441,  445,  448,  452,  456,  459,  463,},
  {  233,  236,  240,  244,  247,  251,  255,  258,  262,  265,  269,  273,  276,  280,  284,  287,  291,  295,  298,  302,  305,  309,  313,  316,  320,  324,  327,  331,  335,  338,  342,  345,  349,  353,  356,  360,  364,  367,  371,  375,  378,  382,  385,  389,  393,  396,  400,  404,  407,  411,  415,  418,  422,  425,  429,  433,  436,  440,  444,  447,  451,  455,  458,  462,},
  {  232,  236,  240,  243,  247,  250,  254,  258,  261,  265,  269,  272,  276,  279,  283,  287,  290,  294,  298,  301,  305,  308,  312,  316,  319,  323,  327,  330,  334,  337,  341,  345,  348,  352,  356,  359,  363,  367,  370,  374,  377,  381,  385,  388,  392,  396,  399,  403,  406,  410,  414,  417,  421,  425,  428,  432,  435,  439,  443,  446,  450,  454,  457,  461,},
  {  232,  235,  239,  243,  246,  250,  253,  257,  261,  264,  268,  272,  275,  279,  282,  286,  290,  293,  297,  301,  304,  308,  311,  315,  319,  322,  326,  330,  333,  337,  340,  344,  348,  351,  355,  358,  362,  366,  369,  373,  377,  380,  384,  387,  391,  395,  398,  402,  406,  409,  413,  416,  420,  424,  427,  431,  435,  438,  442,  445,  449,  453,  456,  460,},
  {  231,  235,  238,  242,  246,  249,  253,  257,  260,  264,  267,  271,  275,  278,  282,  285,  289,  293,  296,  300,  304,  307,  311,  314,  318,  322,  325,  329,  332,  336,  340,  343,  347,  350,  354,  358,  361,  365,  369,  372,  376,  379,  383,  387,  390,  394,  397,  401,  405,  408,  412,  416,  419,  423,  426,  430,  434,  437,  441,  444,  448,  452,  455,  459,},
  {  231,  234,  238,  242,  245,  249,  252,  256,  260,  263,  267,  270,  274,  278,  281,  285,  288,  292,  296,  299,  303,  306,  310,  314,  317,  321,  324,  328,  332,  335,  339,  343,  346,  350,  353,  357,  361,  364,  368,  371,  375,  379,  382,  386,  389,  393,  397,  400,  404,  407,  411,  415,  418,  422,  425,  429,  433,  436,  440,  443,  447,  451,  454,  458,},
  {  230,  234,  237,  241,  245,  248,  252,  255,  259,  263,  266,  270,  273,  277,  281,  284,  288,  291,  295,  299,  302,  306,  309,  313,  317,  320,  324,  327,  331,  335,  338,  342,  345,  349,  353,  356,  360,  363,  367,  371,  374,  378,  381,  385,  389,  392,  396,  399,  403,  407,  410,  414,  417,  421,  425,  428,  432,  435,  439,  443,  446,  450,  453,  457,},
  {  230,  233,  237,  241,  244,  248,  251,  255,  258,  262,  266,  269,  273,  276,  280,  284,  287,  291,  294,  298,  302,  305,  309,  312,  316,  319,  323,  327,  330,  334,  337,  341,  345,  348,  352,  355,  359,  363,  366,  370,  373,  377,  381,  384,  388,  391,  395,  398,  402,  406,  409,  413,  416,  420,  424,  427,  431,  434,  438,  442,  445,  449,  452,  456,},
  {  229,  233,  236,  240,  244,  247,  251,  254,  258,  261,  265,  269,  272,  276,  279,  283,  287,  290,  294,  297,  301,  304,  308,  312,  315,  319,  322,  326,  330,  333,  337,  340,  344,  347,  351,  355,  358,  362,  365,  369,  373,  376,  380,  383,  387,  390,  394,  398,  401,  405,  408,  412,  416,  419,  423,  426,  430,  433,  437,  441,  444,  448,  451,  455,},
  {  229,  232,  236,  239,  243,  247,  250,  254,  257,  261,  264,  268,  272,  275,  279,  282,  286,  290,  293,  297,  300,  304,  307,  311,  315,  318,  322,  325,  329,  332,  336,  340,  343,  347,  350,  354,  357,  361,  365,  368,  372,  375,  379,  382,  386,  390,  393,  397,  400,  404,  407,  411,  415,  418,  422,  425,  429,  432,  436,  440,  443,  447,  450,  454,},
  {  228,  232,  235,  239,  243,  246,  250,  253,  257,  260,  264,  267,  271,  275,  278,  282,  285,  289,  292,  296,  300,  303,  307,  310,  314,  317,  321,  325,  328,  332,  335,  339,  342,  346,  350,  353,  357,  360,  364,  367,  371,  374,  378,  382,  385,  389,  392,  396,  399,  403,  407,  410,  414,  417,  421,  424,  428,  432,  435,  439,  442,  446,  449,  453,},
  {  228,  231,  235,  238,  242,  246,  249,  253,  256,  260,  263,  267,  270,  274,  278,  281,  285,  288,  292,  295,  299,  302,  306,  310,  313,  317,  320,  324,  327,  331,  335,  338,  342,  345,  349,  352,  356,  359,  363,  367,  370,  374,  377,  381,  384,  388,  391,  395,  399,  402,  406,  409,  413,  416,  420,  423,  427,  431,  434,  438,  441,  445,  448,  452,},
  {  227,  231,  234,  238,  241,  245,  249,  252,  256,  259,  263,  266,  270,  273,  277,  281,  284,  288,  291,  295,  298,  302,  305,  309,  312,  316,  320,  323,  327,  330,  334,  337,  341,  344,  348,  352,  355,  359,  362,  366,  369,  373,  376,  380,  383,  387,  391,  394,  398,  401,  405,  408,  412,  415,  419,  423,  426,  430,  433,  437,  440,  444,  447,  451,},
  {  227,  230,  234,  237,  241,  244,  248,  252,  255,  259,  262,  266,  269,  273,  276,  280,  283,  287,  291,  294,  298,  301,  305,  308,  312,  315,  319,  322,  326,  329,  333,  337,  340,  344,  347,  351,  354,  358,  361,  365,  368,  372,  376,  379,  383,  386,  390,  393,  397,  400,  404,  407,  411,  415,  418,  422,  425,  429,  432,  436,  439,  443,  446,  450,},
  {  226,  230,  233,  237,  240,  244,  247,  251,  255,  258,  262,  265,  269,  272,  276,  279,  283,  286,  290,  293,  297,  300,  304,  308,  311,  315,  318,  322,  325,  329,  332,  336,  339,  343,  346,  350,  354,  357,  361,  364,  368,  371,  375,  378,  382,  385,  389,  392,  396,  399,  403,  407,  410,  414,  417,  421,  424,  428,  431,  435,  438,  442,  445,  449,},
  {  226,  229,  233,  236,  240,  243,  247,  250,  254,  257,  261,  265,  268,  272,  275,  279,  282,  286,  289,  293,  296,  300,  303,  307,  310,  314,  317,  321,  325,  328,  332,  335,  339,  342,  346,  349,  353,  356,  360,  363,  367,  370,  374,  377,  381,  384,  388,  392,  395,  399,  402,  406,  409,  413,  416,  420,  423,  427,  430,  434,  437,  441,  444,  448,},
  {  225,  229,  232,  236,  239,  243,  246,  250,  253,  257,  260,  264,  267,  271,  275,  278,  282,  285,  289,  292,  296,  299,  303,  306,  310,  313,  317,  320,  324,  327,  331,  334,  338,  341,  345,  348,  352,  355,  359,  363,  366,  370,  373,  377,  380,  384,  387,  391,  394,  398,  401,  405,  408,  412,  415,  419,  422,  426,  429,  433,  436,  440,  443,  447,},
  {  225,  228,  232,  235,  239,  242,  246,  249,  253,  256,  260,  263,  267,  270,  274,  277,  281,  284,  288,  291,  295,  298,  302,  306,  309,  313,  316,  320,  323,  327,  330,  334,  337,  341,  344,  348,  351,  355,  358,  362,  365,  369,  372,  376,  379,  383,  386,  390,  393,  397,  400,  404,  407,  411,  414,  418,  421,  425,  428,  432,  435,  439,  442,  446,},
  {  224,  228,  231,  235,  238,  242,  245,  249,  252,  256,  259,  263,  266,  270,  273,  277,  280,  284,  287,  291,  294,  298,  301,  305,  308,  312,  315,  319,  322,  326,  329,  333,  336,  340,  343,  347,  350,  354,  357,  361,  364,  368,  371,  375,  378,  382,  385,  389,  392,  396,  399,  403,  406,  410,  413,  417,  420,  424,  427,  431,  434,  438,  441,  445,},
  {  224,  227,  231,  234,  238,  241,  245,  248,  252,  255,  259,  262,  266,  269,  273,  276,  280,  283,  287,  290,  294,  297,  301,  304,  308,  311,  315,  318,  322,  325,  329,  332,  336,  339,  343,  346,  350,  353,  357,  360,  364,  367,  371,  374,  378,  381,  385,  388,  392,  395,  399,  402,  406,  409,  413,  416,  420,  423,  427,  430,  434,  437,  441,  444,},
  {  223,  227,  230,  234,  237,  241,  244,  248,  251,  255,  258,  262,  265,  269,  272,  276,  279,  283,  286,  290,  293,  297,  300,  303,  307,  310,  314,  317,  321,  324,  328,  331,  335,  338,  342,  345,  349,  352,  356,  359,  363,  366,  370,  373,  377,  380,  384,  387,  391,  394,  398,  401,  405,  408,  412,  415,  419,  422,  426,  429,  433,  436,  440,  443,},
  {  223,  226,  230,  233,  237,  240,  244,  247,  251,  254,  258,  261,  265,  268,  271,  275,  278,  282,  285,  289,  292,  296,  299,  303,  306,  310,  313,  317,  320,  324,  327,  331,  334,  338,  341,  345,  348,  352,  355,  358,  362,  365,  369,  372,  376,  379,  383,  386,  390,  393,  397,  400,  404,  407,  411,  414,  418,  421,  425,  428,  432,  435,  439,  442,},
  {  222,  226,  229,  233,  236,  240,  243,  247,  250,  254,  257,  260,  264,  267,  271,  274,  278,  281,  285,  288,  292,  295,  299,  302,  306,  309,  313,  316,  319,  323,  326,  330,  333,  337,  340,  344,  347,  351,  354,  358,  361,  365,  368,  372,  375,  379,  382,  385,  389,  392,  396,  399,  403,  406,  410,  413,  417,  420,  424,  427,  431,  434,  438,  441,},
  {  222,  225,  229,  232,  236,  239,  243,  246,  249,  253,  256,  260,  263,  267,  270,  274,  277,  281,  284,  288,  291,  295,  298,  301,  305,  308,  312,  315,  319,  322,  326,  329,  333,  336,  340,  343,  346,  350,  353,  357,  360,  364,  367,  371,  374,  378,  381,  385,  388,  392,  395,  398,  402,  405,  409,  412,  416,  419,  423,  426,  430,  433,  437,  440,},
  {  221,  225,  228,  232,  235,  239,  242,  245,  249,  252,  256,  259,  263,  266,  270,  273,  277,  280,  283,  287,  290,  294,  297,  301,  304,  308,  311,  315,  318,  322,  325,  328,  332,  335,  339,  342,  346,  349,  353,  356,  360,  363,  366,  370,  373,  377,  380,  384,  387,  391,  394,  398,  401,  404,  408,  411,  415,  418,  422,  425,  429,  432,  436,  439,},
  {  221,  224,  228,  231,  235,  238,  241,  245,  248,  252,  255,  259,  262,  266,  269,  272,  276,  279,  283,  286,  290,  293,  297,  300,  304,  307,  310,  314,  317,  321,  324,  328,  331,  335,  338,  341,  345,  348,  352,  355,  359,  362,  366,  369,  373,  376,  379,  383,  386,  390,  393,  397,  400,  404,  407,  410,  414,  417,  421,  424,  428,  431,  435,  438,},
  {  220,  224,  227,  231,  234,  237,  241,  244,  248,  251,  255,  258,  262,  265,  268,  272,  275,  279,  282,  286,  289,  293,  296,  299,  303,  306,  310,  313,  317,  320,  323,  327,  330,  334,  337,  341,  344,  348,  351,  354,  358,  361,  365,  368,  372,  375,  379,  382,  385,  389,  392,  396,  399,  403,  406,  410,  413,  416,  420,  423,  427,  430,  434,  437,},
  {  220,  223,  227,  230,  233,  237,  240,  244,  247,  251,  254,  258,  261,  264,  268,  271,  275,  278,  282,  285,  288,  292,  295,  299,  302,  306,  309,  312,  316,  319,  323,  326,  330,  333,  336,  340,  343,  347,  350,  354,  357,  361,  364,  367,  371,  374,  378,  381,  385,  388,  391,  395,  398,  402,  405,  409,  412,  415,  419,  422,  426,  429,  433,  436,},
  {  219,  223,  226,  230,  233,  236,  240,  243,  247,  250,  254,  257,  260,  264,  267,  271,  274,  277,  281,  284,  288,  291,  295,  298,  301,  305,  308,  312,  315,  319,  322,  325,  329,  332,  336,  339,  343,  346,  349,  353,  356,  360,  363,  367,  370,  373,  377,  380,  384,  387,  391,  394,  397,  401,  404,  408,  411,  415,  418,  421,  425,  428,  432,  435,},
  {  219,  222,  226,  229,  232,  236,  239,  243,  246,  250,  253,  256,  260,  263,  267,  270,  273,  277,  280,  284,  287,  291,  294,  297,  301,  304,  308,  311,  314,  318,  321,  325,  328,  332,  335,  338,  342,  345,  349,  352,  355,  359,  362,  366,  369,  373,  376,  379,  383,  386,  390,  393,  396,  400,  403,  407,  410,  414,  417,  420,  424,  427,  431,  434,},
  {  218,  222,  225,  228,  232,  235,  239,  242,  246,  249,  252,  256,  259,  263,  266,  269,  273,  276,  280,  283,  286,  290,  293,  297,  300,  304,  307,  310,  314,  317,  321,  324,  327,  331,  334,  338,  341,  344,  348,  351,  355,  358,  361,  365,  368,  372,  375,  379,  382,  385,  389,  392,  396,  399,  402,  406,  409,  413,  416,  419,  423,  426,  430,  433,},
  {  218,  221,  225,  228,  231,  235,  238,  242,  245,  248,  252,  255,  259,  262,  265,  269,  272,  276,  279,  282,  286,  289,  293,  296,  299,  303,  306,  310,  313,  316,  320,  323,  327,  330,  333,  337,  340,  344,  347,  350,  354,  357,  361,  364,  367,  371,  374,  378,  381,  384,  388,  391,  395,  398,  401,  405,  408,  412,  415,  418,  422,  425,  429,  432,},
  {  217,  221,  224,  227,  231,  234,  238,  241,  244,  248,  251,  255,  258,  261,  265,  268,  272,  275,  278,  282,  285,  289,  292,  295,  299,  302,  306,  309,  312,  316,  319,  322,  326,  329,  333,  336,  339,  343,  346,  350,  353,  356,  360,  363,  367,  370,  373,  377,  380,  384,  387,  390,  394,  397,  401,  404,  407,  411,  414,  418,  421,  424,  428,  431,},
  {  217,  220,  224,  227,  230,  234,  237,  240,  244,  247,  251,  254,  257,  261,  264,  268,  271,  274,  278,  281,  284,  288,  291,  295,  298,  301,  305,  308,  312,  315,  318,  322,  325,  329,  332,  335,  339,  342,  345,  349,  352,  356,  359,  362,  366,  369,  373,  376,  379,  383,  386,  389,  393,  396,  400,  403,  406,  410,  413,  417,  420,  423,  427,  430,},
  {  216,  220,  223,  226,  230,  233,  237,  240,  243,  247,  250,  253,  257,  260,  264,  267,  270,  274,  277,  280,  284,  287,  291,  294,  297,  301,  304,  307,  311,  314,  318,  321,  324,  328,  331,  335,  338,  341,  345,  348,  351,  355,  358,  362,  365,  368,  372,  375,  378,  382,  385,  389,  392,  395,  399,  402,  405,  409,  412,  416,  419,  422,  426,  429,},
  {  216,  219,  222,  226,  229,  233,  236,  239,  243,  246,  249,  253,  256,  260,  263,  266,  270,  273,  276,  280,  283,  287,  290,  293,  297,  300,  303,  307,  310,  314,  317,  320,  324,  327,  330,  334,  337,  340,  344,  347,  351,  354,  357,  361,  364,  367,  371,  374,  378,  381,  384,  388,  391,  394,  398,  401,  405,  408,  411,  415,  418,  421,  425,  428,},
  {  215,  219,  222,  225,  229,  232,  235,  239,  242,  246,  249,  252,  256,  259,  262,  266,  269,  272,  276,  279,  283,  286,  289,  293,  296,  299,  303,  306,  309,  313,  316,  320,  323,  326,  330,  333,  336,  340,  343,  346,  350,  353,  357,  360,  363,  367,  370,  373,  377,  380,  383,  387,  390,  394,  397,  400,  404,  407,  410,  414,  417,  420,  424,  427,},
  {  215,  218,  221,  225,  228,  232,  235,  238,  242,  245,  248,  252,  255,  258,  262,  265,  268,  272,  275,  279,  282,  285,  289,  292,  295,  299,  302,  305,  309,  312,  315,  319,  322,  325,  329,  332,  336,  339,  342,  346,  349,  352,  356,  359,  362,  366,  369,  372,  376,  379,  383,  386,  389,  393,  396,  399,  403,  406,  409,  413,  416,  419,  423,  426,},
  {  214,  218,  221,  224,  228,  231,  234,  238,  241,  244,  248,  251,  254,  258,  261,  264,  268,  271,  275,  278,  281,  285,  288,  291,  295,  298,  301,  305,  308,  311,  315,  318,  321,  325,  328,  331,  335,  338,  341,  345,  348,  352,  355,  358,  362,  365,  368,  372,  375,  378,  382,  385,  388,  392,  395,  398,  402,  405,  408,  412,  415,  418,  422,  425,},
  {  214,  217,  220,  224,  227,  230,  234,  237,  240,  244,  247,  250,  254,  257,  261,  264,  267,  271,  274,  277,  281,  284,  287,  291,  294,  297,  301,  304,  307,  311,  314,  317,  321,  324,  327,  331,  334,  337,  341,  344,  347,  351,  354,  357,  361,  364,  367,  371,  374,  377,  381,  384,  387,  391,  394,  397,  401,  404,  407,  411,  414,  417,  421,  424,},
  {  213,  217,  220,  223,  227,  230,  233,  237,  240,  243,  247,  250,  253,  257,  260,  263,  267,  270,  273,  277,  280,  283,  287,  290,  293,  297,  300,  303,  307,  310,  313,  317,  320,  323,  327,  330,  333,  337,  340,  343,  347,  350,  353,  357,  360,  363,  367,  370,  373,  377,  380,  383,  387,  390,  393,  397,  400,  403,  407,  410,  413,  417,  420,  423,},
  {  213,  216,  219,  223,  226,  229,  233,  236,  239,  243,  246,  249,  253,  256,  259,  263,  266,  269,  273,  276,  279,  283,  286,  289,  293,  296,  299,  303,  306,  309,  312,  316,  319,  322,  326,  329,  332,  336,  339,  342,  346,  349,  352,  356,  359,  362,  366,  369,  372,  376,  379,  382,  386,  389,  392,  396,  399,  402,  406,  409,  412,  416,  419,  422,},
  {  212,  216,  219,  222,  226,  229,  232,  235,  239,  242,  245,  249,  252,  255,  259,  262,  265,  269,  272,  275,  279,  282,  285,  289,  292,  295,  298,  302,  305,  308,  312,  315,  318,  322,  325,  328,  332,  335,  338,  342,  345,  348,  352,  355,  358,  361,  365,  368,  371,  375,  378,  381,  385,  388,  391,  395,  398,  401,  405,  408,  411,  415,  418,  421,},
  {  212,  215,  218,  222,  225,  228,  232,  235,  238,  242,  245,  248,  251,  255,  258,  261,  265,  268,  271,  275,  278,  281,  285,  288,  291,  294,  298,  301,  304,  308,  311,  314,  318,  321,  324,  328,  331,  334,  337,  341,  344,  347,  351,  354,  357,  361,  364,  367,  371,  374,  377,  380,  384,  387,  390,  394,  397,  400,  404,  407,  410,  414,  417,  420,},
  {  211,  215,  218,  221,  224,  228,  231,  234,  238,  241,  244,  248,  251,  254,  257,  261,  264,  267,  271,  274,  277,  281,  284,  287,  290,  294,  297,  300,  304,  307,  310,  314,  317,  320,  323,  327,  330,  333,  337,  340,  343,  347,  350,  353,  356,  360,  363,  366,  370,  373,  376,  380,  383,  386,  389,  393,  396,  399,  403,  406,  409,  413,  416,  419,},
  {  211,  214,  217,  221,  224,  227,  231,  234,  237,  240,  244,  247,  250,  254,  257,  260,  263,  267,  270,  273,  277,  280,  283,  286,  290,  293,  296,  300,  303,  306,  310,  313,  316,  319,  323,  326,  329,  333,  336,  339,  342,  346,  349,  352,  356,  359,  362,  366,  369,  372,  375,  379,  382,  385,  389,  392,  395,  398,  402,  405,  408,  412,  415,  418,},
  {  210,  214,  217,  220,  223,  227,  230,  233,  237,  240,  243,  246,  250,  253,  256,  260,  263,  266,  269,  273,  276,  279,  283,  286,  289,  292,  296,  299,  302,  306,  309,  312,  315,  319,  322,  325,  329,  332,  335,  338,  342,  345,  348,  352,  355,  358,  361,  365,  368,  371,  375,  378,  381,  384,  388,  391,  394,  398,  401,  404,  407,  411,  414,  417,},
  {  210,  213,  216,  220,  223,  226,  229,  233,  236,  239,  243,  246,  249,  252,  256,  259,  262,  265,  269,  272,  275,  279,  282,  285,  288,  292,  295,  298,  302,  305,  308,  311,  315,  318,  321,  324,  328,  331,  334,  338,  341,  344,  347,  351,  354,  357,  361,  364,  367,  370,  374,  377,  380,  383,  387,  390,  393,  397,  400,  403,  406,  410,  413,  416,},
  {  209,  213,  216,  219,  222,  226,  229,  232,  235,  239,  242,  245,  248,  252,  255,  258,  262,  265,  268,  271,  275,  278,  281,  284,  288,  291,  294,  298,  301,  304,  307,  311,  314,  317,  320,  324,  327,  330,  333,  337,  340,  343,  347,  350,  353,  356,  360,  363,  366,  369,  373,  376,  379,  383,  386,  389,  392,  396,  399,  402,  405,  409,  412,  415,},
  {  209,  212,  215,  219,  222,  225,  228,  232,  235,  238,  241,  245,  248,  251,  254,  258,  261,  264,  267,  271,  274,  277,  281,  284,  287,  290,  294,  297,  300,  303,  307,  310,  313,  316,  320,  323,  326,  329,  333,  336,  339,  342,  346,  349,  352,  356,  359,  362,  365,  369,  372,  375,  378,  382,  385,  388,  391,  395,  398,  401,  404,  408,  411,  414,},
  {  208,  212,  215,  218,  221,  225,  228,  231,  234,  238,  241,  244,  247,  251,  254,  257,  260,  264,  267,  270,  273,  277,  280,  283,  286,  290,  293,  296,  299,  303,  306,  309,  312,  316,  319,  322,  325,  329,  332,  335,  338,  342,  345,  348,  351,  355,  358,  361,  364,  368,  371,  374,  377,  381,  384,  387,  390,  394,  397,  400,  403,  407,  410,  413,},
  {  208,  211,  214,  217,  221,  224,  227,  230,  234,  237,  240,  243,  247,  250,  253,  256,  260,  263,  266,  269,  273,  276,  279,  282,  286,  289,  292,  295,  299,  302,  305,  308,  312,  315,  318,  321,  325,  328,  331,  334,  338,  341,  344,  347,  351,  354,  357,  360,  364,  367,  370,  373,  377,  380,  383,  386,  390,  393,  396,  399,  403,  406,  409,  412,},
  {  207,  210,  214,  217,  220,  223,  227,  230,  233,  236,  240,  243,  246,  249,  253,  256,  259,  262,  266,  269,  272,  275,  278,  282,  285,  288,  291,  295,  298,  301,  304,  308,  311,  314,  317,  321,  324,  327,  330,  334,  337,  340,  343,  346,  350,  353,  356,  359,  363,  366,  369,  372,  376,  379,  382,  385,  389,  392,  395,  398,  402,  405,  408,  411,},
  {  207,  210,  213,  216,  220,  223,  226,  229,  233,  236,  239,  242,  246,  249,  252,  255,  258,  262,  265,  268,  271,  275,  278,  281,  284,  288,  291,  294,  297,  300,  304,  307,  310,  313,  317,  320,  323,  326,  330,  333,  336,  339,  342,  346,  349,  352,  355,  359,  362,  365,  368,  372,  375,  378,  381,  384,  388,  391,  394,  397,  401,  404,  407,  410,},
  {  206,  209,  213,  216,  219,  222,  226,  229,  232,  235,  238,  242,  245,  248,  251,  255,  258,  261,  264,  267,  271,  274,  277,  280,  284,  287,  290,  293,  296,  300,  303,  306,  309,  313,  316,  319,  322,  325,  329,  332,  335,  338,  342,  345,  348,  351,  354,  358,  361,  364,  367,  371,  374,  377,  380,  383,  387,  390,  393,  396,  400,  403,  406,  409,},
  {  206,  209,  212,  215,  219,  222,  225,  228,  231,  235,  238,  241,  244,  248,  251,  254,  257,  260,  264,  267,  270,  273,  276,  280,  283,  286,  289,  293,  296,  299,  302,  305,  309,  312,  315,  318,  321,  325,  328,  331,  334,  338,  341,  344,  347,  350,  354,  357,  360,  363,  366,  370,  373,  376,  379,  383,  386,  389,  392,  395,  399,  402,  405,  408,},
  {  205,  208,  212,  215,  218,  221,  224,  228,  231,  234,  237,  241,  244,  247,  250,  253,  257,  260,  263,  266,  269,  273,  276,  279,  282,  285,  289,  292,  295,  298,  301,  305,  308,  311,  314,  317,  321,  324,  327,  330,  334,  337,  340,  343,  346,  350,  353,  356,  359,  362,  366,  369,  372,  375,  378,  382,  385,  388,  391,  394,  398,  401,  404,  407,},
  {  205,  208,  211,  214,  218,  221,  224,  227,  230,  234,  237,  240,  243,  246,  250,  253,  256,  259,  262,  266,  269,  272,  275,  278,  282,  285,  288,  291,  294,  298,  301,  304,  307,  310,  314,  317,  320,  323,  326,  330,  333,  336,  339,  342,  346,  349,  352,  355,  358,  362,  365,  368,  371,  374,  378,  381,  384,  387,  390,  394,  397,  400,  403,  406,},
  {  204,  207,  211,  214,  217,  220,  223,  227,  230,  233,  236,  239,  243,  246,  249,  252,  255,  259,  262,  265,  268,  271,  274,  278,  281,  284,  287,  290,  294,  297,  300,  303,  306,  310,  313,  316,  319,  322,  326,  329,  332,  335,  338,  341,  345,  348,  351,  354,  357,  361,  364,  367,  370,  373,  377,  380,  383,  386,  389,  393,  396,  399,  402,  405,},
  {  204,  207,  210,  213,  216,  220,  223,  226,  229,  232,  236,  239,  242,  245,  248,  252,  255,  258,  261,  264,  267,  271,  274,  277,  280,  283,  287,  290,  293,  296,  299,  302,  306,  309,  312,  315,  318,  322,  325,  328,  331,  334,  337,  341,  344,  347,  350,  353,  357,  360,  363,  366,  369,  372,  376,  379,  382,  385,  388,  392,  395,  398,  401,  404,},
  {  203,  206,  210,  213,  216,  219,  222,  225,  229,  232,  235,  238,  241,  245,  248,  251,  254,  257,  260,  264,  267,  270,  273,  276,  279,  283,  286,  289,  292,  295,  299,  302,  305,  308,  311,  314,  318,  321,  324,  327,  330,  333,  337,  340,  343,  346,  349,  353,  356,  359,  362,  365,  368,  372,  375,  378,  381,  384,  387,  391,  394,  397,  400,  403,},
  {  203,  206,  209,  212,  215,  219,  222,  225,  228,  231,  234,  238,  241,  244,  247,  250,  253,  257,  260,  263,  266,  269,  272,  276,  279,  282,  285,  288,  291,  295,  298,  301,  304,  307,  310,  314,  317,  320,  323,  326,  329,  333,  336,  339,  342,  345,  348,  352,  355,  358,  361,  364,  367,  371,  374,  377,  380,  383,  386,  390,  393,  396,  399,  402,},
  {  202,  205,  209,  212,  215,  218,  221,  224,  228,  231,  234,  237,  240,  243,  246,  250,  253,  256,  259,  262,  265,  269,  272,  275,  278,  281,  284,  288,  291,  294,  297,  300,  303,  307,  310,  313,  316,  319,  322,  325,  329,  332,  335,  338,  341,  344,  348,  351,  354,  357,  360,  363,  367,  370,  373,  376,  379,  382,  386,  389,  392,  395,  398,  401,},
  {  202,  205,  208,  211,  214,  218,  221,  224,  227,  230,  233,  236,  240,  243,  246,  249,  252,  255,  258,  262,  265,  268,  271,  274,  277,  281,  284,  287,  290,  293,  296,  299,  303,  306,  309,  312,  315,  318,  322,  325,  328,  331,  334,  337,  340,  344,  347,  350,  353,  356,  359,  363,  366,  369,  372,  375,  378,  381,  385,  388,  391,  394,  397,  400,},
  {  201,  204,  208,  211,  214,  217,  220,  223,  226,  230,  233,  236,  239,  242,  245,  248,  252,  255,  258,  261,  264,  267,  270,  274,  277,  280,  283,  286,  289,  292,  296,  299,  302,  305,  308,  311,  314,  318,  321,  324,  327,  330,  333,  336,  340,  343,  346,  349,  352,  355,  358,  362,  365,  368,  371,  374,  377,  380,  384,  387,  390,  393,  396,  399,},
  {  201,  204,  207,  210,  213,  216,  220,  223,  226,  229,  232,  235,  238,  242,  245,  248,  251,  254,  257,  260,  263,  267,  270,  273,  276,  279,  282,  285,  289,  292,  295,  298,  301,  304,  307,  311,  314,  317,  320,  323,  326,  329,  332,  336,  339,  342,  345,  348,  351,  354,  358,  361,  364,  367,  370,  373,  376,  380,  383,  386,  389,  392,  395,  398,},
  {  200,  203,  207,  210,  213,  216,  219,  222,  225,  228,  232,  235,  238,  241,  244,  247,  250,  253,  257,  260,  263,  266,  269,  272,  275,  278,  282,  285,  288,  291,  294,  297,  300,  304,  307,  310,  313,  316,  319,  322,  325,  329,  332,  335,  338,  341,  344,  347,  350,  354,  357,  360,  363,  366,  369,  372,  375,  379,  382,  385,  388,  391,  394,  397,},
  {  200,  203,  206,  209,  212,  215,  218,  222,  225,  228,  231,  234,  237,  240,  243,  247,  250,  253,  256,  259,  262,  265,  268,  272,  275,  278,  281,  284,  287,  290,  293,  297,  300,  303,  306,  309,  312,  315,  318,  321,  325,  328,  331,  334,  337,  340,  343,  346,  350,  353,  356,  359,  362,  365,  368,  371,  375,  378,  381,  384,  387,  390,  393,  396,},
  {  199,  202,  205,  209,  212,  215,  218,  221,  224,  227,  230,  233,  237,  240,  243,  246,  249,  252,  255,  258,  262,  265,  268,  271,  274,  277,  280,  283,  286,  290,  293,  296,  299,  302,  305,  308,  311,  314,  318,  321,  324,  327,  330,  333,  336,  339,  342,  346,  349,  352,  355,  358,  361,  364,  367,  370,  374,  377,  380,  383,  386,  389,  392,  395,},
  {  199,  202,  205,  208,  211,  214,  217,  220,  224,  227,  230,  233,  236,  239,  242,  245,  248,  252,  255,  258,  261,  264,  267,  270,  273,  276,  279,  283,  286,  289,  292,  295,  298,  301,  304,  307,  311,  314,  317,  320,  323,  326,  329,  332,  335,  338,  342,  345,  348,  351,  354,  357,  360,  363,  366,  370,  373,  376,  379,  382,  385,  388,  391,  394,},
  {  198,  201,  204,  208,  211,  214,  217,  220,  223,  226,  229,  232,  235,  239,  242,  245,  248,  251,  254,  257,  260,  263,  266,  269,  273,  276,  279,  282,  285,  288,  291,  294,  297,  300,  304,  307,  310,  313,  316,  319,  322,  325,  328,  331,  335,  338,  341,  344,  347,  350,  353,  356,  359,  362,  366,  369,  372,  375,  378,  381,  384,  387,  390,  393,},
  {  198,  201,  204,  207,  210,  213,  216,  219,  222,  226,  229,  232,  235,  238,  241,  244,  247,  250,  253,  256,  260,  263,  266,  269,  272,  275,  278,  281,  284,  287,  290,  294,  297,  300,  303,  306,  309,  312,  315,  318,  321,  324,  328,  331,  334,  337,  340,  343,  346,  349,  352,  355,  358,  362,  365,  368,  371,  374,  377,  380,  383,  386,  389,  392,},
  {  197,  200,  203,  206,  210,  213,  216,  219,  222,  225,  228,  231,  234,  237,  240,  243,  247,  250,  253,  256,  259,  262,  265,  268,  271,  274,  277,  280,  284,  287,  290,  293,  296,  299,  302,  305,  308,  311,  314,  317,  321,  324,  327,  330,  333,  336,  339,  342,  345,  348,  351,  354,  358,  361,  364,  367,  370,  373,  376,  379,  382,  385,  388,  391,},
  {  197,  200,  203,  206,  209,  212,  215,  218,  221,  224,  227,  231,  234,  237,  240,  243,  246,  249,  252,  255,  258,  261,  264,  267,  271,  274,  277,  280,  283,  286,  289,  292,  295,  298,  301,  304,  307,  310,  314,  317,  320,  323,  326,  329,  332,  335,  338,  341,  344,  347,  350,  354,  357,  360,  363,  366,  369,  372,  375,  378,  381,  384,  387,  390,},
  {  196,  199,  202,  205,  209,  212,  215,  218,  221,  224,  227,  230,  233,  236,  239,  242,  245,  248,  251,  255,  258,  261,  264,  267,  270,  273,  276,  279,  282,  285,  288,  291,  294,  297,  301,  304,  307,  310,  313,  316,  319,  322,  325,  328,  331,  334,  337,  340,  343,  347,  350,  353,  356,  359,  362,  365,  368,  371,  374,  377,  380,  383,  386,  389,},
  {  196,  199,  202,  205,  208,  211,  214,  217,  220,  223,  226,  229,  232,  236,  239,  242,  245,  248,  251,  254,  257,  260,  263,  266,  269,  272,  275,  278,  281,  284,  288,  291,  294,  297,  300,  303,  306,  309,  312,  315,  318,  321,  324,  327,  330,  333,  336,  340,  343,  346,  349,  352,  355,  358,  361,  364,  367,  370,  373,  376,  379,  382,  385,  388,},
  {  195,  198,  201,  204,  207,  211,  214,  217,  220,  223,  226,  229,  232,  235,  238,  241,  244,  247,  250,  253,  256,  259,  262,  265,  268,  272,  275,  278,  281,  284,  287,  290,  293,  296,  299,  302,  305,  308,  311,  314,  317,  320,  323,  326,  329,  333,  336,  339,  342,  345,  348,  351,  354,  357,  360,  363,  366,  369,  372,  375,  378,  381,  384,  387,},
  {  195,  198,  201,  204,  207,  210,  213,  216,  219,  222,  225,  228,  231,  234,  237,  240,  243,  246,  250,  253,  256,  259,  262,  265,  268,  271,  274,  277,  280,  283,  286,  289,  292,  295,  298,  301,  304,  307,  310,  313,  316,  320,  323,  326,  329,  332,  335,  338,  341,  344,  347,  350,  353,  356,  359,  362,  365,  368,  371,  374,  377,  380,  383,  386,},
  {  194,  197,  200,  203,  206,  209,  212,  215,  219,  222,  225,  228,  231,  234,  237,  240,  243,  246,  249,  252,  255,  258,  261,  264,  267,  270,  273,  276,  279,  282,  285,  288,  291,  294,  297,  300,  304,  307,  310,  313,  316,  319,  322,  325,  328,  331,  334,  337,  340,  343,  346,  349,  352,  355,  358,  361,  364,  367,  370,  373,  376,  379,  382,  385,},
  {  194,  197,  200,  203,  206,  209,  212,  215,  218,  221,  224,  227,  230,  233,  236,  239,  242,  245,  248,  251,  254,  257,  260,  263,  266,  269,  272,  275,  279,  282,  285,  288,  291,  294,  297,  300,  303,  306,  309,  312,  315,  318,  321,  324,  327,  330,  333,  336,  339,  342,  345,  348,  351,  354,  357,  360,  363,  366,  369,  372,  375,  378,  381,  384,},
  {  193,  196,  199,  202,  205,  208,  211,  214,  217,  220,  223,  226,  229,  233,  236,  239,  242,  245,  248,  251,  254,  257,  260,  263,  266,  269,  272,  275,  278,  281,  284,  287,  290,  293,  296,  299,  302,  305,  308,  311,  314,  317,  320,  323,  326,  329,  332,  335,  338,  341,  344,  347,  350,  353,  356,  359,  362,  365,  368,  371,  374,  377,  380,  383,},
  {  193,  196,  199,  202,  205,  208,  211,  214,  217,  220,  223,  226,  229,  232,  235,  238,  241,  244,  247,  250,  253,  256,  259,  262,  265,  268,  271,  274,  277,  280,  283,  286,  289,  292,  295,  298,  301,  304,  307,  310,  313,  316,  319,  322,  325,  328,  331,  334,  337,  340,  343,  346,  349,  352,  355,  358,  361,  364,  367,  370,  373,  376,  379,  382,},
  {  192,  195,  198,  201,  204,  207,  210,  213,  216,  219,  222,  225,  228,  231,  234,  237,  240,  243,  246,  249,  252,  255,  258,  261,  264,  267,  270,  273,  276,  279,  282,  285,  288,  291,  294,  297,  300,  303,  306,  309,  312,  315,  318,  321,  324,  327,  330,  333,  336,  339,  342,  345,  348,  351,  354,  357,  360,  363,  366,  369,  372,  375,  378,  381,},
  {  192,  195,  198,  201,  204,  207,  210,  213,  216,  219,  222,  225,  228,  231,  234,  237,  240,  243,  246,  249,  252,  255,  258,  261,  264,  267,  270,  273,  276,  279,  282,  285,  288,  291,  294,  297,  300,  303,  306,  309,  312,  315,  318,  321,  324,  327,  330,  333,  336,  339,  342,  345,  348,  351,  354,  357,  360,  363,  366,  369,  372,  375,  378,  381,},
  {  191,  194,  197,  200,  203,  206,  209,  212,  215,  218,  221,  224,  227,  230,  233,  236,  239,  242,  245,  248,  251,  254,  257,  260,  263,  266,  269,  272,  275,  278,  281,  284,  287,  290,  293,  296,  299,  302,  305,  308,  311,  314,  317,  320,  323,  326,  329,  332,  335,  338,  341,  344,  347,  350,  353,  356,  359,  362,  365,  368,  371,  374,  377,  380,},
  {  191,  194,  197,  200,  203,  206,  209,  212,  215,  218,  221,  224,  227,  229,  232,  235,  238,  241,  244,  247,  250,  253,  256,  259,  262,  265,  268,  271,  274,  277,  280,  283,  286,  289,  292,  295,  298,  301,  304,  307,  310,  313,  316,  319,  322,  325,  328,  331,  334,  337,  340,  343,  346,  349,  352,  355,  358,  361,  364,  367,  370,  373,  376,  379,},
  {  190,  193,  196,  199,  202,  205,  208,  211,  214,  217,  220,  223,  226,  229,  232,  235,  238,  241,  244,  247,  250,  253,  256,  259,  262,  265,  268,  271,  273,  276,  279,  282,  285,  288,  291,  294,  297,  300,  303,  306,  309,  312,  315,  318,  321,  324,  327,  330,  333,  336,  339,  342,  345,  348,  351,  354,  357,  360,  363,  366,  369,  372,  375,  378,},
  {  190,  193,  196,  199,  202,  205,  208,  211,  213,  216,  219,  222,  225,  228,  231,  234,  237,  240,  243,  246,  249,  252,  255,  258,  261,  264,  267,  270,  273,  276,  279,  282,  285,  288,  291,  294,  296,  299,  302,  305,  308,  311,  314,  317,  320,  323,  326,  329,  332,  335,  338,  341,  344,  347,  350,  353,  356,  359,  362,  365,  368,  371,  374,  377,},
  {  189,  192,  195,  198,  201,  204,  207,  210,  213,  216,  219,  222,  225,  228,  231,  234,  237,  240,  242,  245,  248,  251,  254,  257,  260,  263,  266,  269,  272,  275,  278,  281,  284,  287,  290,  293,  296,  299,  302,  305,  308,  310,  313,  316,  319,  322,  325,  328,  331,  334,  337,  340,  343,  346,  349,  352,  355,  358,  361,  364,  367,  370,  373,  376,},
  {  189,  192,  195,  198,  201,  203,  206,  209,  212,  215,  218,  221,  224,  227,  230,  233,  236,  239,  242,  245,  248,  251,  254,  257,  260,  262,  265,  268,  271,  274,  277,  280,  283,  286,  289,  292,  295,  298,  301,  304,  307,  310,  313,  316,  319,  321,  324,  327,  330,  333,  336,  339,  342,  345,  348,  351,  354,  357,  360,  363,  366,  369,  372,  375,},
  {  188,  191,  194,  197,  200,  203,  206,  209,  212,  215,  218,  221,  224,  226,  229,  232,  235,  238,  241,  244,  247,  250,  253,  256,  259,  262,  265,  268,  271,  274,  276,  279,  282,  285,  288,  291,  294,  297,  300,  303,  306,  309,  312,  315,  318,  321,  324,  326,  329,  332,  335,  338,  341,  344,  347,  350,  353,  356,  359,  362,  365,  368,  371,  374,},
  {  188,  191,  194,  197,  199,  202,  205,  208,  211,  214,  217,  220,  223,  226,  229,  232,  235,  238,  241,  243,  246,  249,  252,  255,  258,  261,  264,  267,  270,  273,  276,  279,  282,  285,  287,  290,  293,  296,  299,  302,  305,  308,  311,  314,  317,  320,  323,  326,  329,  331,  334,  337,  340,  343,  346,  349,  352,  355,  358,  361,  364,  367,  370,  373,},
  {  187,  190,  193,  196,  199,  202,  205,  208,  211,  214,  217,  219,  222,  225,  228,  231,  234,  237,  240,  243,  246,  249,  252,  255,  257,  260,  263,  266,  269,  272,  275,  278,  281,  284,  287,  290,  293,  296,  298,  301,  304,  307,  310,  313,  316,  319,  322,  325,  328,  331,  334,  336,  339,  342,  345,  348,  351,  354,  357,  360,  363,  366,  369,  372,},
  {  187,  190,  193,  196,  198,  201,  204,  207,  210,  213,  216,  219,  222,  225,  228,  231,  233,  236,  239,  242,  245,  248,  251,  254,  257,  260,  263,  266,  268,  271,  274,  277,  280,  283,  286,  289,  292,  295,  298,  301,  303,  306,  309,  312,  315,  318,  321,  324,  327,  330,  333,  336,  338,  341,  344,  347,  350,  353,  356,  359,  362,  365,  368,  371,},
  {  186,  189,  192,  195,  198,  201,  204,  207,  210,  212,  215,  218,  221,  224,  227,  230,  233,  236,  239,  242,  244,  247,  250,  253,  256,  259,  262,  265,  268,  271,  274,  276,  279,  282,  285,  288,  291,  294,  297,  300,  303,  306,  308,  311,  314,  317,  320,  323,  326,  329,  332,  335,  338,  340,  343,  346,  349,  352,  355,  358,  361,  364,  367,  370,},
  {  186,  189,  192,  194,  197,  200,  203,  206,  209,  212,  215,  218,  221,  223,  226,  229,  232,  235,  238,  241,  244,  247,  250,  253,  255,  258,  261,  264,  267,  270,  273,  276,  279,  282,  284,  287,  290,  293,  296,  299,  302,  305,  308,  311,  313,  316,  319,  322,  325,  328,  331,  334,  337,  340,  342,  345,  348,  351,  354,  357,  360,  363,  366,  369,},
  {  185,  188,  191,  194,  197,  200,  203,  206,  208,  211,  214,  217,  220,  223,  226,  229,  232,  234,  237,  240,  243,  246,  249,  252,  255,  258,  261,  263,  266,  269,  272,  275,  278,  281,  284,  287,  289,  292,  295,  298,  301,  304,  307,  310,  313,  316,  318,  321,  324,  327,  330,  333,  336,  339,  342,  344,  347,  350,  353,  356,  359,  362,  365,  368,},
  {  185,  188,  191,  193,  196,  199,  202,  205,  208,  211,  214,  217,  219,  222,  225,  228,  231,  234,  237,  240,  242,  245,  248,  251,  254,  257,  260,  263,  266,  268,  271,  274,  277,  280,  283,  286,  289,  292,  294,  297,  300,  303,  306,  309,  312,  315,  318,  320,  323,  326,  329,  332,  335,  338,  341,  344,  346,  349,  352,  355,  358,  361,  364,  367,},
  {  184,  187,  190,  193,  196,  199,  202,  204,  207,  210,  213,  216,  219,  222,  225,  227,  230,  233,  236,  239,  242,  245,  248,  250,  253,  256,  259,  262,  265,  268,  271,  273,  276,  279,  282,  285,  288,  291,  294,  297,  299,  302,  305,  308,  311,  314,  317,  320,  322,  325,  328,  331,  334,  337,  340,  343,  345,  348,  351,  354,  357,  360,  363,  366,},
  {  184,  187,  189,  192,  195,  198,  201,  204,  207,  210,  212,  215,  218,  221,  224,  227,  230,  233,  235,  238,  241,  244,  247,  250,  253,  256,  258,  261,  264,  267,  270,  273,  276,  278,  281,  284,  287,  290,  293,  296,  299,  301,  304,  307,  310,  313,  316,  319,  322,  324,  327,  330,  333,  336,  339,  342,  345,  347,  350,  353,  356,  359,  362,  365,},
  {  183,  186,  189,  192,  195,  198,  200,  203,  206,  209,  212,  215,  218,  220,  223,  226,  229,  232,  235,  238,  241,  243,  246,  249,  252,  255,  258,  261,  263,  266,  269,  272,  275,  278,  281,  283,  286,  289,  292,  295,  298,  301,  304,  306,  309,  312,  315,  318,  321,  324,  326,  329,  332,  335,  338,  341,  344,  346,  349,  352,  355,  358,  361,  364,},
  {  183,  186,  188,  191,  194,  197,  200,  203,  206,  208,  211,  214,  217,  220,  223,  226,  228,  231,  234,  237,  240,  243,  246,  248,  251,  254,  257,  260,  263,  266,  268,  271,  274,  277,  280,  283,  286,  288,  291,  294,  297,  300,  303,  306,  308,  311,  314,  317,  320,  323,  326,  328,  331,  334,  337,  340,  343,  346,  348,  351,  354,  357,  360,  363,},
  {  182,  185,  188,  191,  194,  196,  199,  202,  205,  208,  211,  214,  216,  219,  222,  225,  228,  231,  234,  236,  239,  242,  245,  248,  251,  253,  256,  259,  262,  265,  268,  271,  273,  276,  279,  282,  285,  288,  290,  293,  296,  299,  302,  305,  308,  310,  313,  316,  319,  322,  325,  327,  330,  333,  336,  339,  342,  345,  347,  350,  353,  356,  359,  362,},
  {  182,  185,  187,  190,  193,  196,  199,  202,  204,  207,  210,  213,  216,  219,  222,  224,  227,  230,  233,  236,  239,  241,  244,  247,  250,  253,  256,  258,  261,  264,  267,  270,  273,  275,  278,  281,  284,  287,  290,  293,  295,  298,  301,  304,  307,  310,  312,  315,  318,  321,  324,  327,  329,  332,  335,  338,  341,  344,  346,  349,  352,  355,  358,  361,},
  {  181,  184,  187,  190,  193,  195,  198,  201,  204,  207,  210,  212,  215,  218,  221,  224,  227,  229,  232,  235,  238,  241,  244,  246,  249,  252,  255,  258,  261,  263,  266,  269,  272,  275,  278,  280,  283,  286,  289,  292,  295,  297,  300,  303,  306,  309,  312,  314,  317,  320,  323,  326,  329,  331,  334,  337,  340,  343,  346,  348,  351,  354,  357,  360,},
  {  181,  184,  186,  189,  192,  195,  198,  201,  203,  206,  209,  212,  215,  217,  220,  223,  226,  229,  232,  234,  237,  240,  243,  246,  249,  251,  254,  257,  260,  263,  265,  268,  271,  274,  277,  280,  282,  285,  288,  291,  294,  297,  299,  302,  305,  308,  311,  313,  316,  319,  322,  325,  328,  330,  333,  336,  339,  342,  345,  347,  350,  353,  356,  359,},
  {  180,  183,  186,  189,  192,  194,  197,  200,  203,  206,  208,  211,  214,  217,  220,  222,  225,  228,  231,  234,  237,  239,  242,  245,  248,  251,  253,  256,  259,  262,  265,  268,  270,  273,  276,  279,  282,  284,  287,  290,  293,  296,  299,  301,  304,  307,  310,  313,  315,  318,  321,  324,  327,  330,  332,  335,  338,  341,  344,  346,  349,  352,  355,  358,},
  {  180,  183,  185,  188,  191,  194,  197,  199,  202,  205,  208,  211,  213,  216,  219,  222,  225,  227,  230,  233,  236,  239,  242,  244,  247,  250,  253,  256,  258,  261,  264,  267,  270,  272,  275,  278,  281,  284,  286,  289,  292,  295,  298,  301,  303,  306,  309,  312,  315,  317,  320,  323,  326,  329,  331,  334,  337,  340,  343,  345,  348,  351,  354,  357,},
  {  179,  182,  185,  188,  190,  193,  196,  199,  202,  204,  207,  210,  213,  216,  218,  221,  224,  227,  230,  232,  235,  238,  241,  244,  246,  249,  252,  255,  258,  260,  263,  266,  269,  272,  274,  277,  280,  283,  286,  288,  291,  294,  297,  300,  302,  305,  308,  311,  314,  316,  319,  322,  325,  328,  330,  333,  336,  339,  342,  344,  347,  350,  353,  356,},
  {  179,  182,  184,  187,  190,  193,  196,  198,  201,  204,  207,  209,  212,  215,  218,  221,  223,  226,  229,  232,  235,  237,  240,  243,  246,  249,  251,  254,  257,  260,  263,  265,  268,  271,  274,  277,  279,  282,  285,  288,  290,  293,  296,  299,  302,  304,  307,  310,  313,  316,  318,  321,  324,  327,  330,  332,  335,  338,  341,  344,  346,  349,  352,  355,},
  {  178,  181,  184,  187,  189,  192,  195,  198,  201,  203,  206,  209,  212,  214,  217,  220,  223,  226,  228,  231,  234,  237,  240,  242,  245,  248,  251,  253,  256,  259,  262,  265,  267,  270,  273,  276,  279,  281,  284,  287,  290,  292,  295,  298,  301,  304,  306,  309,  312,  315,  318,  320,  323,  326,  329,  331,  334,  337,  340,  343,  345,  348,  351,  354,},
  {  178,  181,  183,  186,  189,  192,  194,  197,  200,  203,  206,  208,  211,  214,  217,  219,  222,  225,  228,  231,  233,  236,  239,  242,  244,  247,  250,  253,  256,  258,  261,  264,  267,  269,  272,  275,  278,  281,  283,  286,  289,  292,  294,  297,  300,  303,  306,  308,  311,  314,  317,  319,  322,  325,  328,  331,  333,  336,  339,  342,  344,  347,  350,  353,},
  {  177,  180,  183,  186,  188,  191,  194,  197,  199,  202,  205,  208,  210,  213,  216,  219,  222,  224,  227,  230,  233,  235,  238,  241,  244,  246,  249,  252,  255,  258,  260,  263,  266,  269,  271,  274,  277,  280,  282,  285,  288,  291,  294,  296,  299,  302,  305,  307,  310,  313,  316,  318,  321,  324,  327,  330,  332,  335,  338,  341,  343,  346,  349,  352,},
  {  177,  180,  182,  185,  188,  191,  193,  196,  199,  202,  204,  207,  210,  213,  215,  218,  221,  224,  226,  229,  232,  235,  238,  240,  243,  246,  249,  251,  254,  257,  260,  262,  265,  268,  271,  273,  276,  279,  282,  284,  287,  290,  293,  296,  298,  301,  304,  307,  309,  312,  315,  318,  320,  323,  326,  329,  331,  334,  337,  340,  342,  345,  348,  351,},
  {  176,  179,  182,  185,  187,  190,  193,  196,  198,  201,  204,  207,  209,  212,  215,  218,  220,  223,  226,  229,  231,  234,  237,  240,  242,  245,  248,  251,  253,  256,  259,  262,  264,  267,  270,  273,  275,  278,  281,  284,  286,  289,  292,  295,  297,  300,  303,  306,  308,  311,  314,  317,  319,  322,  325,  328,  330,  333,  336,  339,  341,  344,  347,  350,},
  {  176,  178,  181,  184,  187,  189,  192,  195,  198,  200,  203,  206,  209,  211,  214,  217,  220,  222,  225,  228,  231,  233,  236,  239,  242,  244,  247,  250,  253,  255,  258,  261,  264,  266,  269,  272,  275,  277,  280,  283,  286,  288,  291,  294,  297,  299,  302,  305,  308,  310,  313,  316,  319,  321,  324,  327,  330,  332,  335,  338,  341,  343,  346,  349,},
  {  175,  178,  181,  183,  186,  189,  192,  194,  197,  200,  203,  205,  208,  211,  214,  216,  219,  222,  225,  227,  230,  233,  235,  238,  241,  244,  246,  249,  252,  255,  257,  260,  263,  266,  268,  271,  274,  277,  279,  282,  285,  288,  290,  293,  296,  298,  301,  304,  307,  309,  312,  315,  318,  320,  323,  326,  329,  331,  334,  337,  340,  342,  345,  348,},
  {  175,  177,  180,  183,  186,  188,  191,  194,  197,  199,  202,  205,  208,  210,  213,  216,  218,  221,  224,  227,  229,  232,  235,  238,  240,  243,  246,  248,  251,  254,  257,  259,  262,  265,  268,  270,  273,  276,  279,  281,  284,  287,  289,  292,  295,  298,  300,  303,  306,  309,  311,  314,  317,  319,  322,  325,  328,  330,  333,  336,  339,  341,  344,  347,},
  {  174,  177,  180,  182,  185,  188,  191,  193,  196,  199,  201,  204,  207,  210,  212,  215,  218,  221,  223,  226,  229,  231,  234,  237,  240,  242,  245,  248,  250,  253,  256,  259,  261,  264,  267,  270,  272,  275,  278,  280,  283,  286,  289,  291,  294,  297,  299,  302,  305,  308,  310,  313,  316,  319,  321,  324,  327,  329,  332,  335,  338,  340,  343,  346,},
  {  174,  176,  179,  182,  185,  187,  190,  193,  195,  198,  201,  204,  206,  209,  212,  214,  217,  220,  223,  225,  228,  231,  233,  236,  239,  242,  244,  247,  250,  252,  255,  258,  261,  263,  266,  269,  271,  274,  277,  280,  282,  285,  288,  290,  293,  296,  299,  301,  304,  307,  309,  312,  315,  318,  320,  323,  326,  328,  331,  334,  337,  339,  342,  345,},
  {  173,  176,  179,  181,  184,  187,  189,  192,  195,  198,  200,  203,  206,  208,  211,  214,  217,  219,  222,  225,  227,  230,  233,  236,  238,  241,  244,  246,  249,  252,  254,  257,  260,  263,  265,  268,  271,  273,  276,  279,  282,  284,  287,  290,  292,  295,  298,  300,  303,  306,  309,  311,  314,  317,  319,  322,  325,  328,  330,  333,  336,  338,  341,  344,},
  {  173,  175,  178,  181,  184,  186,  189,  192,  194,  197,  200,  202,  205,  208,  211,  213,  216,  219,  221,  224,  227,  229,  232,  235,  238,  240,  243,  246,  248,  251,  254,  256,  259,  262,  265,  267,  270,  273,  275,  278,  281,  283,  286,  289,  292,  294,  297,  300,  302,  305,  308,  310,  313,  316,  319,  321,  324,  327,  329,  332,  335,  337,  340,  343,},
  {  172,  175,  178,  180,  183,  186,  188,  191,  194,  196,  199,  202,  205,  207,  210,  213,  215,  218,  221,  223,  226,  229,  231,  234,  237,  240,  242,  245,  248,  250,  253,  256,  258,  261,  264,  266,  269,  272,  275,  277,  280,  283,  285,  288,  291,  293,  296,  299,  301,  304,  307,  310,  312,  315,  318,  320,  323,  326,  328,  331,  334,  336,  339,  342,},
  {  172,  174,  177,  180,  182,  185,  188,  191,  193,  196,  199,  201,  204,  207,  209,  212,  215,  217,  220,  223,  225,  228,  231,  233,  236,  239,  242,  244,  247,  250,  252,  255,  258,  260,  263,  266,  268,  271,  274,  276,  279,  282,  284,  287,  290,  293,  295,  298,  301,  303,  306,  309,  311,  314,  317,  319,  322,  325,  327,  330,  333,  335,  338,  341,},
  {  171,  174,  177,  179,  182,  185,  187,  190,  193,  195,  198,  201,  203,  206,  209,  211,  214,  217,  219,  222,  225,  227,  230,  233,  235,  238,  241,  243,  246,  249,  252,  254,  257,  260,  262,  265,  268,  270,  273,  276,  278,  281,  284,  286,  289,  292,  294,  297,  300,  302,  305,  308,  310,  313,  316,  318,  321,  324,  326,  329,  332,  334,  337,  340,},
  {  171,  173,  176,  179,  181,  184,  187,  189,  192,  195,  197,  200,  203,  205,  208,  211,  213,  216,  219,  221,  224,  227,  229,  232,  235,  237,  240,  243,  245,  248,  251,  253,  256,  259,  261,  264,  267,  269,  272,  275,  277,  280,  283,  285,  288,  291,  293,  296,  299,  301,  304,  307,  309,  312,  315,  317,  320,  323,  325,  328,  331,  333,  336,  339,},
  {  170,  173,  176,  178,  181,  184,  186,  189,  192,  194,  197,  200,  202,  205,  207,  210,  213,  215,  218,  221,  223,  226,  229,  231,  234,  237,  239,  242,  245,  247,  250,  253,  255,  258,  261,  263,  266,  269,  271,  274,  277,  279,  282,  285,  287,  290,  293,  295,  298,  301,  303,  306,  309,  311,  314,  317,  319,  322,  325,  327,  330,  333,  335,  338,},
  {  170,  172,  175,  178,  180,  183,  186,  188,  191,  194,  196,  199,  202,  204,  207,  210,  212,  215,  217,  220,  223,  225,  228,  231,  233,  236,  239,  241,  244,  247,  249,  252,  255,  257,  260,  263,  265,  268,  271,  273,  276,  278,  281,  284,  286,  289,  292,  294,  297,  300,  302,  305,  308,  310,  313,  316,  318,  321,  324,  326,  329,  332,  334,  337,},
  {  169,  172,  175,  177,  180,  182,  185,  188,  190,  193,  196,  198,  201,  204,  206,  209,  212,  214,  217,  219,  222,  225,  227,  230,  233,  235,  238,  241,  243,  246,  249,  251,  254,  257,  259,  262,  264,  267,  270,  272,  275,  278,  280,  283,  286,  288,  291,  294,  296,  299,  301,  304,  307,  309,  312,  315,  317,  320,  323,  325,  328,  331,  333,  336,},
  {  169,  171,  174,  177,  179,  182,  185,  187,  190,  192,  195,  198,  200,  203,  206,  208,  211,  214,  216,  219,  221,  224,  227,  229,  232,  235,  237,  240,  243,  245,  248,  250,  253,  256,  258,  261,  264,  266,  269,  272,  274,  277,  279,  282,  285,  287,  290,  293,  295,  298,  301,  303,  306,  308,  311,  314,  316,  319,  322,  324,  327,  330,  332,  335,},
  {  168,  171,  174,  176,  179,  181,  184,  187,  189,  192,  195,  197,  200,  202,  205,  208,  210,  213,  216,  218,  221,  223,  226,  229,  231,  234,  237,  239,  242,  244,  247,  250,  252,  255,  258,  260,  263,  266,  268,  271,  273,  276,  279,  281,  284,  287,  289,  292,  294,  297,  300,  302,  305,  308,  310,  313,  315,  318,  321,  323,  326,  329,  331,  334,},
  {  168,  170,  173,  176,  178,  181,  183,  186,  189,  191,  194,  197,  199,  202,  204,  207,  210,  212,  215,  218,  220,  223,  225,  228,  231,  233,  236,  239,  241,  244,  246,  249,  252,  254,  257,  259,  262,  265,  267,  270,  273,  275,  278,  280,  283,  286,  288,  291,  294,  296,  299,  301,  304,  307,  309,  312,  315,  317,  320,  322,  325,  328,  330,  333,},
  {  167,  170,  172,  175,  178,  180,  183,  186,  188,  191,  193,  196,  199,  201,  204,  206,  209,  212,  214,  217,  220,  222,  225,  227,  230,  233,  235,  238,  240,  243,  246,  248,  251,  253,  256,  259,  261,  264,  267,  269,  272,  274,  277,  280,  282,  285,  287,  290,  293,  295,  298,  301,  303,  306,  308,  311,  314,  316,  319,  321,  324,  327,  329,  332,},
  {  167,  169,  172,  175,  177,  180,  182,  185,  188,  190,  193,  195,  198,  201,  203,  206,  208,  211,  214,  216,  219,  221,  224,  227,  229,  232,  234,  237,  240,  242,  245,  248,  250,  253,  255,  258,  261,  263,  266,  268,  271,  274,  276,  279,  281,  284,  287,  289,  292,  294,  297,  300,  302,  305,  307,  310,  313,  315,  318,  320,  323,  326,  328,  331,},
  {  166,  169,  171,  174,  177,  179,  182,  184,  187,  190,  192,  195,  197,  200,  203,  205,  208,  210,  213,  216,  218,  221,  223,  226,  229,  231,  234,  236,  239,  242,  244,  247,  249,  252,  255,  257,  260,  262,  265,  268,  270,  273,  275,  278,  281,  283,  286,  288,  291,  294,  296,  299,  301,  304,  307,  309,  312,  314,  317,  320,  322,  325,  327,  330,},
  {  166,  168,  171,  174,  176,  179,  181,  184,  186,  189,  192,  194,  197,  199,  202,  205,  207,  210,  212,  215,  218,  220,  223,  225,  228,  230,  233,  236,  238,  241,  243,  246,  249,  251,  254,  256,  259,  262,  264,  267,  269,  272,  275,  277,  280,  282,  285,  287,  290,  293,  295,  298,  300,  303,  306,  308,  311,  313,  316,  319,  321,  324,  326,  329,},
  {  165,  168,  170,  173,  176,  178,  181,  183,  186,  188,  191,  194,  196,  199,  201,  204,  207,  209,  212,  214,  217,  219,  222,  225,  227,  230,  232,  235,  238,  240,  243,  245,  248,  250,  253,  256,  258,  261,  263,  266,  269,  271,  274,  276,  279,  281,  284,  287,  289,  292,  294,  297,  300,  302,  305,  307,  310,  312,  315,  318,  320,  323,  325,  328,},
  {  165,  167,  170,  172,  175,  178,  180,  183,  185,  188,  190,  193,  196,  198,  201,  203,  206,  209,  211,  214,  216,  219,  221,  224,  227,  229,  232,  234,  237,  239,  242,  245,  247,  250,  252,  255,  257,  260,  263,  265,  268,  270,  273,  275,  278,  281,  283,  286,  288,  291,  293,  296,  299,  301,  304,  306,  309,  311,  314,  317,  319,  322,  324,  327,},
  {  164,  167,  169,  172,  175,  177,  180,  182,  185,  187,  190,  192,  195,  198,  200,  203,  205,  208,  210,  213,  216,  218,  221,  223,  226,  228,  231,  234,  236,  239,  241,  244,  246,  249,  252,  254,  257,  259,  262,  264,  267,  269,  272,  275,  277,  280,  282,  285,  287,  290,  293,  295,  298,  300,  303,  305,  308,  311,  313,  316,  318,  321,  323,  326,},
  {  164,  166,  169,  171,  174,  177,  179,  182,  184,  187,  189,  192,  194,  197,  200,  202,  205,  207,  210,  212,  215,  217,  220,  223,  225,  228,  230,  233,  235,  238,  241,  243,  246,  248,  251,  253,  256,  258,  261,  264,  266,  269,  271,  274,  276,  279,  281,  284,  287,  289,  292,  294,  297,  299,  302,  304,  307,  310,  312,  315,  317,  320,  322,  325,},
  {  163,  166,  168,  171,  173,  176,  179,  181,  184,  186,  189,  191,  194,  196,  199,  202,  204,  207,  209,  212,  214,  217,  219,  222,  224,  227,  230,  232,  235,  237,  240,  242,  245,  247,  250,  253,  255,  258,  260,  263,  265,  268,  270,  273,  275,  278,  281,  283,  286,  288,  291,  293,  296,  298,  301,  304,  306,  309,  311,  314,  316,  319,  321,  324,},
  {  163,  165,  168,  170,  173,  175,  178,  181,  183,  186,  188,  191,  193,  196,  198,  201,  203,  206,  209,  211,  214,  216,  219,  221,  224,  226,  229,  231,  234,  236,  239,  242,  244,  247,  249,  252,  254,  257,  259,  262,  264,  267,  270,  272,  275,  277,  280,  282,  285,  287,  290,  292,  295,  298,  300,  303,  305,  308,  310,  313,  315,  318,  320,  323,},
  {  162,  165,  167,  170,  172,  175,  177,  180,  183,  185,  188,  190,  193,  195,  198,  200,  203,  205,  208,  210,  213,  215,  218,  221,  223,  226,  228,  231,  233,  236,  238,  241,  243,  246,  248,  251,  254,  256,  259,  261,  264,  266,  269,  271,  274,  276,  279,  281,  284,  286,  289,  292,  294,  297,  299,  302,  304,  307,  309,  312,  314,  317,  319,  322,},
  {  162,  164,  167,  169,  172,  174,  177,  179,  182,  184,  187,  190,  192,  195,  197,  200,  202,  205,  207,  210,  212,  215,  217,  220,  222,  225,  227,  230,  233,  235,  238,  240,  243,  245,  248,  250,  253,  255,  258,  260,  263,  265,  268,  270,  273,  275,  278,  281,  283,  286,  288,  291,  293,  296,  298,  301,  303,  306,  308,  311,  313,  316,  318,  321,},
  {  161,  164,  166,  169,  171,  174,  176,  179,  181,  184,  186,  189,  191,  194,  197,  199,  202,  204,  207,  209,  212,  214,  217,  219,  222,  224,  227,  229,  232,  234,  237,  239,  242,  244,  247,  249,  252,  254,  257,  260,  262,  265,  267,  270,  272,  275,  277,  280,  282,  285,  287,  290,  292,  295,  297,  300,  302,  305,  307,  310,  312,  315,  317,  320,},
  {  161,  163,  166,  168,  171,  173,  176,  178,  181,  183,  186,  188,  191,  193,  196,  198,  201,  203,  206,  208,  211,  213,  216,  219,  221,  224,  226,  229,  231,  234,  236,  239,  241,  244,  246,  249,  251,  254,  256,  259,  261,  264,  266,  269,  271,  274,  276,  279,  281,  284,  286,  289,  291,  294,  296,  299,  301,  304,  306,  309,  311,  314,  316,  319,},
  {  160,  163,  165,  168,  170,  173,  175,  178,  180,  183,  185,  188,  190,  193,  195,  198,  200,  203,  205,  208,  210,  213,  215,  218,  220,  223,  225,  228,  230,  233,  235,  238,  240,  243,  245,  248,  250,  253,  255,  258,  260,  263,  265,  268,  270,  273,  275,  278,  280,  283,  285,  288,  290,  293,  295,  298,  300,  303,  305,  308,  310,  313,  315,  318,},
  {  160,  162,  165,  167,  170,  172,  175,  177,  180,  182,  185,  187,  190,  192,  195,  197,  200,  202,  205,  207,  210,  212,  215,  217,  220,  222,  225,  227,  230,  232,  235,  237,  240,  242,  245,  247,  250,  252,  255,  257,  260,  262,  265,  267,  270,  272,  275,  277,  280,  282,  285,  287,  290,  292,  295,  297,  300,  302,  305,  307,  310,  312,  315,  317,},
  {  159,  162,  164,  167,  169,  172,  174,  177,  179,  182,  184,  187,  189,  192,  194,  197,  199,  202,  204,  207,  209,  212,  214,  216,  219,  221,  224,  226,  229,  231,  234,  236,  239,  241,  244,  246,  249,  251,  254,  256,  259,  261,  264,  266,  269,  271,  274,  276,  279,  281,  284,  286,  289,  291,  294,  296,  299,  301,  304,  306,  309,  311,  314,  316,},
  {  159,  161,  164,  166,  169,  171,  174,  176,  179,  181,  184,  186,  189,  191,  193,  196,  198,  201,  203,  206,  208,  211,  213,  216,  218,  221,  223,  226,  228,  231,  233,  236,  238,  241,  243,  246,  248,  251,  253,  255,  258,  260,  263,  265,  268,  270,  273,  275,  278,  280,  283,  285,  288,  290,  293,  295,  298,  300,  303,  305,  308,  310,  313,  315,},
  {  158,  161,  163,  166,  168,  171,  173,  176,  178,  181,  183,  185,  188,  190,  193,  195,  198,  200,  203,  205,  208,  210,  213,  215,  218,  220,  223,  225,  227,  230,  232,  235,  237,  240,  242,  245,  247,  250,  252,  255,  257,  260,  262,  265,  267,  270,  272,  274,  277,  279,  282,  284,  287,  289,  292,  294,  297,  299,  302,  304,  307,  309,  312,  314,},
  {  158,  160,  163,  165,  168,  170,  173,  175,  177,  180,  182,  185,  187,  190,  192,  195,  197,  200,  202,  205,  207,  210,  212,  214,  217,  219,  222,  224,  227,  229,  232,  234,  237,  239,  242,  244,  246,  249,  251,  254,  256,  259,  261,  264,  266,  269,  271,  274,  276,  279,  281,  283,  286,  288,  291,  293,  296,  298,  301,  303,  306,  308,  311,  313,},
  {  157,  160,  162,  165,  167,  170,  172,  174,  177,  179,  182,  184,  187,  189,  192,  194,  197,  199,  201,  204,  206,  209,  211,  214,  216,  219,  221,  224,  226,  229,  231,  233,  236,  238,  241,  243,  246,  248,  251,  253,  256,  258,  260,  263,  265,  268,  270,  273,  275,  278,  280,  283,  285,  287,  290,  292,  295,  297,  300,  302,  305,  307,  310,  312,},
  {  157,  159,  162,  164,  167,  169,  171,  174,  176,  179,  181,  184,  186,  189,  191,  193,  196,  198,  201,  203,  206,  208,  211,  213,  216,  218,  220,  223,  225,  228,  230,  233,  235,  238,  240,  242,  245,  247,  250,  252,  255,  257,  260,  262,  265,  267,  269,  272,  274,  277,  279,  282,  284,  287,  289,  291,  294,  296,  299,  301,  304,  306,  309,  311,},
  {  156,  159,  161,  164,  166,  168,  171,  173,  176,  178,  181,  183,  186,  188,  190,  193,  195,  198,  200,  203,  205,  208,  210,  212,  215,  217,  220,  222,  225,  227,  229,  232,  234,  237,  239,  242,  244,  247,  249,  251,  254,  256,  259,  261,  264,  266,  269,  271,  273,  276,  278,  281,  283,  286,  288,  291,  293,  295,  298,  300,  303,  305,  308,  310,},
  {  156,  158,  161,  163,  165,  168,  170,  173,  175,  178,  180,  183,  185,  187,  190,  192,  195,  197,  200,  202,  204,  207,  209,  212,  214,  217,  219,  221,  224,  226,  229,  231,  234,  236,  238,  241,  243,  246,  248,  251,  253,  256,  258,  260,  263,  265,  268,  270,  273,  275,  277,  280,  282,  285,  287,  290,  292,  294,  297,  299,  302,  304,  307,  309,},
  {  155,  158,  160,  163,  165,  167,  170,  172,  175,  177,  180,  182,  184,  187,  189,  192,  194,  196,  199,  201,  204,  206,  209,  211,  213,  216,  218,  221,  223,  226,  228,  230,  233,  235,  238,  240,  243,  245,  247,  250,  252,  255,  257,  260,  262,  264,  267,  269,  272,  274,  277,  279,  281,  284,  286,  289,  291,  294,  296,  298,  301,  303,  306,  308,},
  {  155,  157,  160,  162,  164,  167,  169,  172,  174,  177,  179,  181,  184,  186,  189,  191,  193,  196,  198,  201,  203,  206,  208,  210,  213,  215,  218,  220,  222,  225,  227,  230,  232,  235,  237,  239,  242,  244,  247,  249,  251,  254,  256,  259,  261,  264,  266,  268,  271,  273,  276,  278,  280,  283,  285,  288,  290,  293,  295,  297,  300,  302,  305,  307,},
  {  154,  157,  159,  161,  164,  166,  169,  171,  174,  176,  178,  181,  183,  186,  188,  190,  193,  195,  198,  200,  202,  205,  207,  210,  212,  215,  217,  219,  222,  224,  227,  229,  231,  234,  236,  239,  241,  243,  246,  248,  251,  253,  255,  258,  260,  263,  265,  268,  270,  272,  275,  277,  280,  282,  284,  287,  289,  292,  294,  296,  299,  301,  304,  306,},
  {  154,  156,  159,  161,  163,  166,  168,  171,  173,  175,  178,  180,  183,  185,  187,  190,  192,  195,  197,  199,  202,  204,  207,  209,  211,  214,  216,  219,  221,  223,  226,  228,  231,  233,  235,  238,  240,  243,  245,  247,  250,  252,  255,  257,  259,  262,  264,  267,  269,  271,  274,  276,  279,  281,  283,  286,  288,  291,  293,  295,  298,  300,  303,  305,},
  {  153,  156,  158,  160,  163,  165,  168,  170,  172,  175,  177,  180,  182,  184,  187,  189,  192,  194,  196,  199,  201,  204,  206,  208,  211,  213,  216,  218,  220,  223,  225,  227,  230,  232,  235,  237,  239,  242,  244,  247,  249,  251,  254,  256,  259,  261,  263,  266,  268,  271,  273,  275,  278,  280,  283,  285,  287,  290,  292,  295,  297,  299,  302,  304,},
  {  153,  155,  158,  160,  162,  165,  167,  169,  172,  174,  177,  179,  181,  184,  186,  189,  191,  193,  196,  198,  200,  203,  205,  208,  210,  212,  215,  217,  220,  222,  224,  227,  229,  232,  234,  236,  239,  241,  243,  246,  248,  251,  253,  255,  258,  260,  263,  265,  267,  270,  272,  274,  277,  279,  282,  284,  286,  289,  291,  294,  296,  298,  301,  303,},
  {  152,  155,  157,  159,  162,  164,  167,  169,  171,  174,  176,  178,  181,  183,  186,  188,  190,  193,  195,  197,  200,  202,  205,  207,  209,  212,  214,  216,  219,  221,  224,  226,  228,  231,  233,  236,  238,  240,  243,  245,  247,  250,  252,  255,  257,  259,  262,  264,  266,  269,  271,  274,  276,  278,  281,  283,  285,  288,  290,  293,  295,  297,  300,  302,},
  {  152,  154,  156,  159,  161,  164,  166,  168,  171,  173,  175,  178,  180,  183,  185,  187,  190,  192,  194,  197,  199,  202,  204,  206,  209,  211,  213,  216,  218,  221,  223,  225,  228,  230,  232,  235,  237,  239,  242,  244,  247,  249,  251,  254,  256,  258,  261,  263,  266,  268,  270,  273,  275,  277,  280,  282,  285,  287,  289,  292,  294,  296,  299,  301,},
  {  151,  154,  156,  158,  161,  163,  165,  168,  170,  173,  175,  177,  180,  182,  184,  187,  189,  191,  194,  196,  199,  201,  203,  206,  208,  210,  213,  215,  217,  220,  222,  225,  227,  229,  232,  234,  236,  239,  241,  243,  246,  248,  251,  253,  255,  258,  260,  262,  265,  267,  269,  272,  274,  277,  279,  281,  284,  286,  288,  291,  293,  295,  298,  300,},
  {  151,  153,  155,  158,  160,  163,  165,  167,  170,  172,  174,  177,  179,  181,  184,  186,  188,  191,  193,  196,  198,  200,  203,  205,  207,  210,  212,  214,  217,  219,  221,  224,  226,  228,  231,  233,  236,  238,  240,  243,  245,  247,  250,  252,  254,  257,  259,  261,  264,  266,  269,  271,  273,  276,  278,  280,  283,  285,  287,  290,  292,  294,  297,  299,},
  {  150,  153,  155,  157,  160,  162,  164,  167,  169,  171,  174,  176,  178,  181,  183,  185,  188,  190,  193,  195,  197,  200,  202,  204,  207,  209,  211,  214,  216,  218,  221,  223,  225,  228,  230,  232,  235,  237,  239,  242,  244,  247,  249,  251,  254,  256,  258,  261,  263,  265,  268,  270,  272,  275,  277,  279,  282,  284,  286,  289,  291,  293,  296,  298,},
  {  150,  152,  154,  157,  159,  161,  164,  166,  168,  171,  173,  175,  178,  180,  183,  185,  187,  190,  192,  194,  197,  199,  201,  204,  206,  208,  211,  213,  215,  218,  220,  222,  225,  227,  229,  232,  234,  236,  239,  241,  243,  246,  248,  250,  253,  255,  257,  260,  262,  264,  267,  269,  271,  274,  276,  278,  281,  283,  285,  288,  290,  292,  295,  297,},
  {  149,  152,  154,  156,  159,  161,  163,  166,  168,  170,  173,  175,  177,  180,  182,  184,  187,  189,  191,  194,  196,  198,  201,  203,  205,  208,  210,  212,  215,  217,  219,  222,  224,  226,  229,  231,  233,  236,  238,  240,  243,  245,  247,  250,  252,  254,  257,  259,  261,  264,  266,  268,  271,  273,  275,  278,  280,  282,  285,  287,  289,  292,  294,  296,},
  {  149,  151,  153,  156,  158,  160,  163,  165,  167,  170,  172,  174,  177,  179,  181,  184,  186,  188,  191,  193,  195,  198,  200,  202,  205,  207,  209,  212,  214,  216,  218,  221,  223,  225,  228,  230,  232,  235,  237,  239,  242,  244,  246,  249,  251,  253,  256,  258,  260,  263,  265,  267,  270,  272,  274,  277,  279,  281,  284,  286,  288,  291,  293,  295,},
  {  148,  151,  153,  155,  158,  160,  162,  164,  167,  169,  171,  174,  176,  178,  181,  183,  185,  188,  190,  192,  195,  197,  199,  202,  204,  206,  208,  211,  213,  215,  218,  220,  222,  225,  227,  229,  232,  234,  236,  239,  241,  243,  246,  248,  250,  252,  255,  257,  259,  262,  264,  266,  269,  271,  273,  276,  278,  280,  283,  285,  287,  290,  292,  294,},
  {  148,  150,  152,  155,  157,  159,  162,  164,  166,  169,  171,  173,  175,  178,  180,  182,  185,  187,  189,  192,  194,  196,  199,  201,  203,  205,  208,  210,  212,  215,  217,  219,  222,  224,  226,  229,  231,  233,  235,  238,  240,  242,  245,  247,  249,  252,  254,  256,  259,  261,  263,  265,  268,  270,  272,  275,  277,  279,  282,  284,  286,  289,  291,  293,},
  {  147,  150,  152,  154,  156,  159,  161,  163,  166,  168,  170,  173,  175,  177,  179,  182,  184,  186,  189,  191,  193,  196,  198,  200,  202,  205,  207,  209,  212,  214,  216,  219,  221,  223,  225,  228,  230,  232,  235,  237,  239,  242,  244,  246,  248,  251,  253,  255,  258,  260,  262,  265,  267,  269,  271,  274,  276,  278,  281,  283,  285,  288,  290,  292,},
  {  147,  149,  151,  154,  156,  158,  161,  163,  165,  167,  170,  172,  174,  177,  179,  181,  183,  186,  188,  190,  193,  195,  197,  199,  202,  204,  206,  209,  211,  213,  216,  218,  220,  222,  225,  227,  229,  232,  234,  236,  238,  241,  243,  245,  248,  250,  252,  255,  257,  259,  261,  264,  266,  268,  271,  273,  275,  277,  280,  282,  284,  287,  289,  291,},
  {  146,  149,  151,  153,  155,  158,  160,  162,  165,  167,  169,  171,  174,  176,  178,  181,  183,  185,  187,  190,  192,  194,  197,  199,  201,  203,  206,  208,  210,  213,  215,  217,  219,  222,  224,  226,  229,  231,  233,  235,  238,  240,  242,  245,  247,  249,  251,  254,  256,  258,  261,  263,  265,  267,  270,  272,  274,  277,  279,  281,  283,  286,  288,  290,},
  {  146,  148,  150,  153,  155,  157,  159,  162,  164,  166,  169,  171,  173,  175,  178,  180,  182,  184,  187,  189,  191,  194,  196,  198,  200,  203,  205,  207,  210,  212,  214,  216,  219,  221,  223,  225,  228,  230,  232,  235,  237,  239,  241,  244,  246,  248,  251,  253,  255,  257,  260,  262,  264,  266,  269,  271,  273,  276,  278,  280,  282,  285,  287,  289,},
  {  145,  148,  150,  152,  154,  157,  159,  161,  163,  166,  168,  170,  172,  175,  177,  179,  182,  184,  186,  188,  191,  193,  195,  197,  200,  202,  204,  207,  209,  211,  213,  216,  218,  220,  222,  225,  227,  229,  231,  234,  236,  238,  241,  243,  245,  247,  250,  252,  254,  256,  259,  261,  263,  266,  268,  270,  272,  275,  277,  279,  281,  284,  286,  288,},
  {  145,  147,  149,  152,  154,  156,  158,  161,  163,  165,  167,  170,  172,  174,  176,  179,  181,  183,  185,  188,  190,  192,  195,  197,  199,  201,  204,  206,  208,  210,  213,  215,  217,  219,  222,  224,  226,  228,  231,  233,  235,  237,  240,  242,  244,  247,  249,  251,  253,  256,  258,  260,  262,  265,  267,  269,  271,  274,  276,  278,  280,  283,  285,  287,},
  {  144,  147,  149,  151,  153,  156,  158,  160,  162,  165,  167,  169,  171,  174,  176,  178,  180,  183,  185,  187,  189,  192,  194,  196,  198,  201,  203,  205,  207,  210,  212,  214,  216,  219,  221,  223,  225,  228,  230,  232,  234,  237,  239,  241,  243,  246,  248,  250,  252,  255,  257,  259,  261,  264,  266,  268,  270,  273,  275,  277,  279,  282,  284,  286,},
  {  144,  146,  148,  150,  153,  155,  157,  159,  162,  164,  166,  168,  171,  173,  175,  177,  180,  182,  184,  186,  189,  191,  193,  195,  198,  200,  202,  204,  207,  209,  211,  213,  216,  218,  220,  222,  225,  227,  229,  231,  234,  236,  238,  240,  243,  245,  247,  249,  252,  254,  256,  258,  261,  263,  265,  267,  270,  272,  274,  276,  279,  281,  283,  285,},
  {  143,  145,  148,  150,  152,  154,  157,  159,  161,  163,  166,  168,  170,  172,  175,  177,  179,  181,  184,  186,  188,  190,  192,  195,  197,  199,  201,  204,  206,  208,  210,  213,  215,  217,  219,  222,  224,  226,  228,  231,  233,  235,  237,  239,  242,  244,  246,  248,  251,  253,  255,  257,  260,  262,  264,  266,  269,  271,  273,  275,  278,  280,  282,  284,},
  {  143,  145,  147,  149,  152,  154,  156,  158,  161,  163,  165,  167,  170,  172,  174,  176,  178,  181,  183,  185,  187,  190,  192,  194,  196,  199,  201,  203,  205,  207,  210,  212,  214,  216,  219,  221,  223,  225,  228,  230,  232,  234,  236,  239,  241,  243,  245,  248,  250,  252,  254,  257,  259,  261,  263,  265,  268,  270,  272,  274,  277,  279,  281,  283,},
  {  142,  144,  147,  149,  151,  153,  156,  158,  160,  162,  164,  167,  169,  171,  173,  176,  178,  180,  182,  184,  187,  189,  191,  193,  196,  198,  200,  202,  204,  207,  209,  211,  213,  216,  218,  220,  222,  224,  227,  229,  231,  233,  236,  238,  240,  242,  244,  247,  249,  251,  253,  256,  258,  260,  262,  264,  267,  269,  271,  273,  276,  278,  280,  282,},
  {  142,  144,  146,  148,  151,  153,  155,  157,  159,  162,  164,  166,  168,  171,  173,  175,  177,  179,  182,  184,  186,  188,  190,  193,  195,  197,  199,  202,  204,  206,  208,  210,  213,  215,  217,  219,  221,  224,  226,  228,  230,  233,  235,  237,  239,  241,  244,  246,  248,  250,  252,  255,  257,  259,  261,  264,  266,  268,  270,  272,  275,  277,  279,  281,},
  {  141,  143,  146,  148,  150,  152,  154,  157,  159,  161,  163,  166,  168,  170,  172,  174,  177,  179,  181,  183,  185,  188,  190,  192,  194,  196,  199,  201,  203,  205,  207,  210,  212,  214,  216,  218,  221,  223,  225,  227,  230,  232,  234,  236,  238,  241,  243,  245,  247,  249,  252,  254,  256,  258,  260,  263,  265,  267,  269,  271,  274,  276,  278,  280,},
  {  141,  143,  145,  147,  150,  152,  154,  156,  158,  161,  163,  165,  167,  169,  172,  174,  176,  178,  180,  183,  185,  187,  189,  191,  194,  196,  198,  200,  202,  205,  207,  209,  211,  213,  216,  218,  220,  222,  224,  227,  229,  231,  233,  235,  238,  240,  242,  244,  246,  249,  251,  253,  255,  257,  260,  262,  264,  266,  268,  271,  273,  275,  277,  279,},
  {  140,  142,  145,  147,  149,  151,  153,  156,  158,  160,  162,  164,  167,  169,  171,  173,  175,  178,  180,  182,  184,  186,  188,  191,  193,  195,  197,  199,  202,  204,  206,  208,  210,  213,  215,  217,  219,  221,  224,  226,  228,  230,  232,  234,  237,  239,  241,  243,  245,  248,  250,  252,  254,  256,  259,  261,  263,  265,  267,  270,  272,  274,  276,  278,},
  {  140,  142,  144,  146,  148,  151,  153,  155,  157,  159,  162,  164,  166,  168,  170,  173,  175,  177,  179,  181,  183,  186,  188,  190,  192,  194,  197,  199,  201,  203,  205,  207,  210,  212,  214,  216,  218,  221,  223,  225,  227,  229,  231,  234,  236,  238,  240,  242,  245,  247,  249,  251,  253,  255,  258,  260,  262,  264,  266,  269,  271,  273,  275,  277,},
  {  139,  141,  144,  146,  148,  150,  152,  154,  157,  159,  161,  163,  165,  168,  170,  172,  174,  176,  178,  181,  183,  185,  187,  189,  191,  194,  196,  198,  200,  202,  205,  207,  209,  211,  213,  215,  218,  220,  222,  224,  226,  228,  231,  233,  235,  237,  239,  242,  244,  246,  248,  250,  252,  255,  257,  259,  261,  263,  265,  268,  270,  272,  274,  276,},
  {  139,  141,  143,  145,  147,  150,  152,  154,  156,  158,  160,  163,  165,  167,  169,  171,  173,  176,  178,  180,  182,  184,  186,  189,  191,  193,  195,  197,  199,  202,  204,  206,  208,  210,  212,  215,  217,  219,  221,  223,  225,  228,  230,  232,  234,  236,  238,  241,  243,  245,  247,  249,  251,  254,  256,  258,  260,  262,  264,  267,  269,  271,  273,  275,},
  {  138,  140,  143,  145,  147,  149,  151,  153,  156,  158,  160,  162,  164,  166,  168,  171,  173,  175,  177,  179,  181,  184,  186,  188,  190,  192,  194,  197,  199,  201,  203,  205,  207,  210,  212,  214,  216,  218,  220,  222,  225,  227,  229,  231,  233,  235,  238,  240,  242,  244,  246,  248,  251,  253,  255,  257,  259,  261,  264,  266,  268,  270,  272,  274,},
  {  138,  140,  142,  144,  146,  149,  151,  153,  155,  157,  159,  161,  164,  166,  168,  170,  172,  174,  176,  179,  181,  183,  185,  187,  189,  192,  194,  196,  198,  200,  202,  204,  207,  209,  211,  213,  215,  217,  220,  222,  224,  226,  228,  230,  232,  235,  237,  239,  241,  243,  245,  248,  250,  252,  254,  256,  258,  260,  263,  265,  267,  269,  271,  273,},
  {  137,  139,  142,  144,  146,  148,  150,  152,  154,  157,  159,  161,  163,  165,  167,  169,  172,  174,  176,  178,  180,  182,  184,  187,  189,  191,  193,  195,  197,  199,  202,  204,  206,  208,  210,  212,  214,  217,  219,  221,  223,  225,  227,  229,  232,  234,  236,  238,  240,  242,  244,  247,  249,  251,  253,  255,  257,  259,  262,  264,  266,  268,  270,  272,},
  {  137,  139,  141,  143,  145,  147,  150,  152,  154,  156,  158,  160,  162,  165,  167,  169,  171,  173,  175,  177,  179,  182,  184,  186,  188,  190,  192,  194,  197,  199,  201,  203,  205,  207,  209,  212,  214,  216,  218,  220,  222,  224,  226,  229,  231,  233,  235,  237,  239,  241,  244,  246,  248,  250,  252,  254,  256,  259,  261,  263,  265,  267,  269,  271,},
  {  136,  138,  141,  143,  145,  147,  149,  151,  153,  155,  158,  160,  162,  164,  166,  168,  170,  172,  175,  177,  179,  181,  183,  185,  187,  189,  192,  194,  196,  198,  200,  202,  204,  207,  209,  211,  213,  215,  217,  219,  221,  224,  226,  228,  230,  232,  234,  236,  238,  241,  243,  245,  247,  249,  251,  253,  255,  258,  260,  262,  264,  266,  268,  270,},
  {  136,  138,  140,  142,  144,  146,  148,  151,  153,  155,  157,  159,  161,  163,  165,  168,  170,  172,  174,  176,  178,  180,  182,  185,  187,  189,  191,  193,  195,  197,  199,  202,  204,  206,  208,  210,  212,  214,  216,  218,  221,  223,  225,  227,  229,  231,  233,  235,  238,  240,  242,  244,  246,  248,  250,  252,  255,  257,  259,  261,  263,  265,  267,  269,},
  {  135,  137,  139,  142,  144,  146,  148,  150,  152,  154,  156,  158,  161,  163,  165,  167,  169,  171,  173,  175,  178,  180,  182,  184,  186,  188,  190,  192,  194,  197,  199,  201,  203,  205,  207,  209,  211,  213,  216,  218,  220,  222,  224,  226,  228,  230,  232,  235,  237,  239,  241,  243,  245,  247,  249,  251,  254,  256,  258,  260,  262,  264,  266,  268,},
  {  135,  137,  139,  141,  143,  145,  147,  149,  152,  154,  156,  158,  160,  162,  164,  166,  168,  171,  173,  175,  177,  179,  181,  183,  185,  187,  189,  192,  194,  196,  198,  200,  202,  204,  206,  208,  211,  213,  215,  217,  219,  221,  223,  225,  227,  229,  232,  234,  236,  238,  240,  242,  244,  246,  248,  251,  253,  255,  257,  259,  261,  263,  265,  267,},
  {  134,  136,  138,  141,  143,  145,  147,  149,  151,  153,  155,  157,  159,  162,  164,  166,  168,  170,  172,  174,  176,  178,  180,  182,  185,  187,  189,  191,  193,  195,  197,  199,  201,  203,  206,  208,  210,  212,  214,  216,  218,  220,  222,  224,  227,  229,  231,  233,  235,  237,  239,  241,  243,  245,  248,  250,  252,  254,  256,  258,  260,  262,  264,  266,},
  {  134,  136,  138,  140,  142,  144,  146,  148,  150,  153,  155,  157,  159,  161,  163,  165,  167,  169,  171,  173,  176,  178,  180,  182,  184,  186,  188,  190,  192,  194,  196,  199,  201,  203,  205,  207,  209,  211,  213,  215,  217,  219,  222,  224,  226,  228,  230,  232,  234,  236,  238,  240,  242,  245,  247,  249,  251,  253,  255,  257,  259,  261,  263,  265,},
  {  133,  135,  137,  139,  142,  144,  146,  148,  150,  152,  154,  156,  158,  160,  162,  164,  167,  169,  171,  173,  175,  177,  179,  181,  183,  185,  187,  189,  192,  194,  196,  198,  200,  202,  204,  206,  208,  210,  212,  214,  217,  219,  221,  223,  225,  227,  229,  231,  233,  235,  237,  239,  242,  244,  246,  248,  250,  252,  254,  256,  258,  260,  262,  264,},
  {  133,  135,  137,  139,  141,  143,  145,  147,  149,  151,  153,  156,  158,  160,  162,  164,  166,  168,  170,  172,  174,  176,  178,  180,  183,  185,  187,  189,  191,  193,  195,  197,  199,  201,  203,  205,  207,  209,  212,  214,  216,  218,  220,  222,  224,  226,  228,  230,  232,  234,  236,  239,  241,  243,  245,  247,  249,  251,  253,  255,  257,  259,  261,  263,},
  {  132,  134,  136,  138,  141,  143,  145,  147,  149,  151,  153,  155,  157,  159,  161,  163,  165,  167,  169,  172,  174,  176,  178,  180,  182,  184,  186,  188,  190,  192,  194,  196,  198,  200,  203,  205,  207,  209,  211,  213,  215,  217,  219,  221,  223,  225,  227,  229,  231,  234,  236,  238,  240,  242,  244,  246,  248,  250,  252,  254,  256,  258,  260,  262,},
  {  132,  134,  136,  138,  140,  142,  144,  146,  148,  150,  152,  154,  156,  159,  161,  163,  165,  167,  169,  171,  173,  175,  177,  179,  181,  183,  185,  187,  189,  191,  194,  196,  198,  200,  202,  204,  206,  208,  210,  212,  214,  216,  218,  220,  222,  224,  226,  229,  231,  233,  235,  237,  239,  241,  243,  245,  247,  249,  251,  253,  255,  257,  259,  261,},
  {  131,  133,  135,  137,  139,  142,  144,  146,  148,  150,  152,  154,  156,  158,  160,  162,  164,  166,  168,  170,  172,  174,  176,  178,  180,  183,  185,  187,  189,  191,  193,  195,  197,  199,  201,  203,  205,  207,  209,  211,  213,  215,  217,  219,  221,  224,  226,  228,  230,  232,  234,  236,  238,  240,  242,  244,  246,  248,  250,  252,  254,  256,  258,  260,},
  {  131,  133,  135,  137,  139,  141,  143,  145,  147,  149,  151,  153,  155,  157,  159,  161,  163,  165,  168,  170,  172,  174,  176,  178,  180,  182,  184,  186,  188,  190,  192,  194,  196,  198,  200,  202,  204,  206,  208,  210,  212,  215,  217,  219,  221,  223,  225,  227,  229,  231,  233,  235,  237,  239,  241,  243,  245,  247,  249,  251,  253,  255,  257,  259,},
  {  130,  132,  134,  136,  138,  140,  142,  144,  147,  149,  151,  153,  155,  157,  159,  161,  163,  165,  167,  169,  171,  173,  175,  177,  179,  181,  183,  185,  187,  189,  191,  193,  195,  197,  199,  201,  204,  206,  208,  210,  212,  214,  216,  218,  220,  222,  224,  226,  228,  230,  232,  234,  236,  238,  240,  242,  244,  246,  248,  250,  252,  254,  256,  258,},
  {  130,  132,  134,  136,  138,  140,  142,  144,  146,  148,  150,  152,  154,  156,  158,  160,  162,  164,  166,  168,  170,  172,  174,  176,  178,  180,  182,  184,  187,  189,  191,  193,  195,  197,  199,  201,  203,  205,  207,  209,  211,  213,  215,  217,  219,  221,  223,  225,  227,  229,  231,  233,  235,  237,  239,  241,  243,  245,  247,  249,  251,  253,  255,  257,},
  {  129,  131,  133,  135,  137,  139,  141,  143,  145,  147,  149,  151,  153,  156,  158,  160,  162,  164,  166,  168,  170,  172,  174,  176,  178,  180,  182,  184,  186,  188,  190,  192,  194,  196,  198,  200,  202,  204,  206,  208,  210,  212,  214,  216,  218,  220,  222,  224,  226,  228,  230,  232,  234,  236,  238,  240,  242,  244,  246,  248,  250,  252,  254,  256,},
  {  129,  131,  133,  135,  137,  139,  141,  143,  145,  147,  149,  151,  153,  155,  157,  159,  161,  163,  165,  167,  169,  171,  173,  175,  177,  179,  181,  183,  185,  187,  189,  191,  193,  195,  197,  199,  201,  203,  205,  207,  209,  211,  213,  215,  217,  219,  221,  223,  225,  227,  229,  231,  233,  235,  237,  239,  241,  243,  245,  247,  249,  251,  253,  255,},
  {  128,  130,  132,  134,  136,  138,  140,  142,  144,  146,  148,  150,  152,  154,  156,  158,  160,  162,  164,  166,  168,  170,  172,  174,  176,  178,  180,  182,  184,  186,  188,  190,  192,  194,  196,  198,  200,  202,  204,  206,  208,  210,  212,  214,  216,  218,  220,  222,  224,  226,  228,  230,  232,  234,  236,  238,  240,  242,  244,  246,  248,  250,  252,  254,},
  {  128,  130,  132,  134,  136,  138,  140,  142,  144,  146,  148,  150,  152,  154,  156,  158,  160,  162,  164,  166,  168,  170,  172,  174,  176,  178,  180,  182,  184,  186,  188,  190,  192,  194,  196,  198,  200,  202,  204,  206,  208,  210,  212,  214,  216,  218,  220,  222,  224,  226,  228,  230,  232,  234,  236,  238,  240,  242,  244,  246,  248,  250,  252,  254,},
  {  127,  129,  131,  133,  135,  137,  139,  141,  143,  145,  147,  149,  151,  153,  155,  157,  159,  161,  163,  165,  167,  169,  171,  173,  175,  177,  179,  181,  183,  185,  187,  189,  191,  193,  195,  197,  199,  201,  203,  205,  207,  209,  211,  213,  215,  217,  219,  221,  223,  225,  227,  229,  231,  233,  235,  237,  239,  241,  243,  245,  247,  249,  251,  253,},
  {  127,  129,  131,  133,  135,  137,  139,  141,  143,  145,  147,  149,  151,  152,  154,  156,  158,  160,  162,  164,  166,  168,  170,  172,  174,  176,  178,  180,  182,  184,  186,  188,  190,  192,  194,  196,  198,  200,  202,  204,  206,  208,  210,  212,  214,  216,  218,  220,  222,  224,  226,  228,  230,  232,  234,  236,  238,  240,  242,  244,  246,  248,  250,  252,},
  {  126,  128,  130,  132,  134,  136,  138,  140,  142,  144,  146,  148,  150,  152,  154,  156,  158,  160,  162,  164,  166,  168,  170,  172,  174,  176,  178,  180,  181,  183,  185,  187,  189,  191,  193,  195,  197,  199,  201,  203,  205,  207,  209,  211,  213,  215,  217,  219,  221,  223,  225,  227,  229,  231,  233,  235,  237,  239,  241,  243,  245,  247,  249,  251,},
  {  126,  128,  130,  132,  134,  136,  138,  140,  141,  143,  145,  147,  149,  151,  153,  155,  157,  159,  161,  163,  165,  167,  169,  171,  173,  175,  177,  179,  181,  183,  185,  187,  189,  191,  193,  195,  196,  198,  200,  202,  204,  206,  208,  210,  212,  214,  216,  218,  220,  222,  224,  226,  228,  230,  232,  234,  236,  238,  240,  242,  244,  246,  248,  250,},
  {  125,  127,  129,  131,  133,  135,  137,  139,  141,  143,  145,  147,  149,  151,  153,  155,  157,  159,  160,  162,  164,  166,  168,  170,  172,  174,  176,  178,  180,  182,  184,  186,  188,  190,  192,  194,  196,  198,  200,  202,  204,  205,  207,  209,  211,  213,  215,  217,  219,  221,  223,  225,  227,  229,  231,  233,  235,  237,  239,  241,  243,  245,  247,  249,},
  {  125,  127,  129,  131,  133,  134,  136,  138,  140,  142,  144,  146,  148,  150,  152,  154,  156,  158,  160,  162,  164,  166,  168,  170,  172,  173,  175,  177,  179,  181,  183,  185,  187,  189,  191,  193,  195,  197,  199,  201,  203,  205,  207,  209,  211,  212,  214,  216,  218,  220,  222,  224,  226,  228,  230,  232,  234,  236,  238,  240,  242,  244,  246,  248,},
  {  124,  126,  128,  130,  132,  134,  136,  138,  140,  142,  144,  146,  148,  149,  151,  153,  155,  157,  159,  161,  163,  165,  167,  169,  171,  173,  175,  177,  179,  181,  182,  184,  186,  188,  190,  192,  194,  196,  198,  200,  202,  204,  206,  208,  210,  212,  214,  215,  217,  219,  221,  223,  225,  227,  229,  231,  233,  235,  237,  239,  241,  243,  245,  247,},
  {  124,  126,  128,  130,  131,  133,  135,  137,  139,  141,  143,  145,  147,  149,  151,  153,  155,  157,  159,  160,  162,  164,  166,  168,  170,  172,  174,  176,  178,  180,  182,  184,  186,  188,  189,  191,  193,  195,  197,  199,  201,  203,  205,  207,  209,  211,  213,  215,  217,  218,  220,  222,  224,  226,  228,  230,  232,  234,  236,  238,  240,  242,  244,  246,},
  {  123,  125,  127,  129,  131,  133,  135,  137,  139,  141,  143,  144,  146,  148,  150,  152,  154,  156,  158,  160,  162,  164,  166,  168,  169,  171,  173,  175,  177,  179,  181,  183,  185,  187,  189,  191,  193,  195,  196,  198,  200,  202,  204,  206,  208,  210,  212,  214,  216,  218,  220,  221,  223,  225,  227,  229,  231,  233,  235,  237,  239,  241,  243,  245,},
  {  123,  125,  127,  129,  130,  132,  134,  136,  138,  140,  142,  144,  146,  148,  150,  152,  153,  155,  157,  159,  161,  163,  165,  167,  169,  171,  173,  175,  176,  178,  180,  182,  184,  186,  188,  190,  192,  194,  196,  198,  199,  201,  203,  205,  207,  209,  211,  213,  215,  217,  219,  221,  222,  224,  226,  228,  230,  232,  234,  236,  238,  240,  242,  244,},
  {  122,  124,  126,  128,  130,  132,  134,  136,  138,  139,  141,  143,  145,  147,  149,  151,  153,  155,  157,  159,  160,  162,  164,  166,  168,  170,  172,  174,  176,  178,  180,  181,  183,  185,  187,  189,  191,  193,  195,  197,  199,  201,  202,  204,  206,  208,  210,  212,  214,  216,  218,  220,  222,  223,  225,  227,  229,  231,  233,  235,  237,  239,  241,  243,},
  {  122,  124,  126,  127,  129,  131,  133,  135,  137,  139,  141,  143,  145,  146,  148,  150,  152,  154,  156,  158,  160,  162,  164,  166,  167,  169,  171,  173,  175,  177,  179,  181,  183,  185,  186,  188,  190,  192,  194,  196,  198,  200,  202,  204,  205,  207,  209,  211,  213,  215,  217,  219,  221,  223,  224,  226,  228,  230,  232,  234,  236,  238,  240,  242,},
  {  121,  123,  125,  127,  129,  131,  133,  135,  136,  138,  140,  142,  144,  146,  148,  150,  152,  153,  155,  157,  159,  161,  163,  165,  167,  169,  171,  172,  174,  176,  178,  180,  182,  184,  186,  188,  189,  191,  193,  195,  197,  199,  201,  203,  205,  207,  208,  210,  212,  214,  216,  218,  220,  222,  224,  225,  227,  229,  231,  233,  235,  237,  239,  241,},
  {  121,  123,  125,  126,  128,  130,  132,  134,  136,  138,  140,  142,  143,  145,  147,  149,  151,  153,  155,  157,  158,  160,  162,  164,  166,  168,  170,  172,  174,  175,  177,  179,  181,  183,  185,  187,  189,  191,  192,  194,  196,  198,  200,  202,  204,  206,  208,  209,  211,  213,  215,  217,  219,  221,  223,  225,  226,  228,  230,  232,  234,  236,  238,  240,},
  {  120,  122,  124,  126,  128,  130,  132,  133,  135,  137,  139,  141,  143,  145,  147,  148,  150,  152,  154,  156,  158,  160,  162,  163,  165,  167,  169,  171,  173,  175,  177,  178,  180,  182,  184,  186,  188,  190,  192,  194,  195,  197,  199,  201,  203,  205,  207,  209,  210,  212,  214,  216,  218,  220,  222,  224,  225,  227,  229,  231,  233,  235,  237,  239,},
  {  120,  122,  123,  125,  127,  129,  131,  133,  135,  137,  138,  140,  142,  144,  146,  148,  150,  152,  153,  155,  157,  159,  161,  163,  165,  167,  168,  170,  172,  174,  176,  178,  180,  181,  183,  185,  187,  189,  191,  193,  195,  196,  198,  200,  202,  204,  206,  208,  210,  211,  213,  215,  217,  219,  221,  223,  225,  226,  228,  230,  232,  234,  236,  238,},
  {  119,  121,  123,  125,  127,  129,  130,  132,  134,  136,  138,  140,  142,  143,  145,  147,  149,  151,  153,  155,  157,  158,  160,  162,  164,  166,  168,  170,  171,  173,  175,  177,  179,  181,  183,  184,  186,  188,  190,  192,  194,  196,  198,  199,  201,  203,  205,  207,  209,  211,  212,  214,  216,  218,  220,  222,  224,  225,  227,  229,  231,  233,  235,  237,},
  {  119,  121,  122,  124,  126,  128,  130,  132,  134,  135,  137,  139,  141,  143,  145,  147,  148,  150,  152,  154,  156,  158,  160,  161,  163,  165,  167,  169,  171,  173,  174,  176,  178,  180,  182,  184,  186,  187,  189,  191,  193,  195,  197,  199,  200,  202,  204,  206,  208,  210,  212,  213,  215,  217,  219,  221,  223,  225,  226,  228,  230,  232,  234,  236,},
  {  118,  120,  122,  124,  126,  127,  129,  131,  133,  135,  137,  139,  140,  142,  144,  146,  148,  150,  152,  153,  155,  157,  159,  161,  163,  164,  166,  168,  170,  172,  174,  176,  177,  179,  181,  183,  185,  187,  188,  190,  192,  194,  196,  198,  200,  201,  203,  205,  207,  209,  211,  212,  214,  216,  218,  220,  222,  224,  225,  227,  229,  231,  233,  235,},
  {  118,  120,  121,  123,  125,  127,  129,  131,  132,  134,  136,  138,  140,  142,  144,  145,  147,  149,  151,  153,  155,  156,  158,  160,  162,  164,  166,  167,  169,  171,  173,  175,  177,  178,  180,  182,  184,  186,  188,  190,  191,  193,  195,  197,  199,  201,  202,  204,  206,  208,  210,  212,  213,  215,  217,  219,  221,  223,  224,  226,  228,  230,  232,  234,},
  {  117,  119,  121,  123,  125,  126,  128,  130,  132,  134,  136,  137,  139,  141,  143,  145,  147,  148,  150,  152,  154,  156,  158,  159,  161,  163,  165,  167,  169,  170,  172,  174,  176,  178,  180,  181,  183,  185,  187,  189,  191,  192,  194,  196,  198,  200,  202,  203,  205,  207,  209,  211,  213,  214,  216,  218,  220,  222,  224,  225,  227,  229,  231,  233,},
  {  117,  119,  120,  122,  124,  126,  128,  130,  131,  133,  135,  137,  139,  140,  142,  144,  146,  148,  150,  151,  153,  155,  157,  159,  161,  162,  164,  166,  168,  170,  171,  173,  175,  177,  179,  181,  182,  184,  186,  188,  190,  192,  193,  195,  197,  199,  201,  202,  204,  206,  208,  210,  212,  213,  215,  217,  219,  221,  223,  224,  226,  228,  230,  232,},
  {  116,  118,  120,  122,  124,  125,  127,  129,  131,  133,  134,  136,  138,  140,  142,  143,  145,  147,  149,  151,  153,  154,  156,  158,  160,  162,  163,  165,  167,  169,  171,  173,  174,  176,  178,  180,  182,  183,  185,  187,  189,  191,  193,  194,  196,  198,  200,  202,  203,  205,  207,  209,  211,  213,  214,  216,  218,  220,  222,  223,  225,  227,  229,  231,},
  {  116,  118,  119,  121,  123,  125,  127,  128,  130,  132,  134,  136,  137,  139,  141,  143,  145,  146,  148,  150,  152,  154,  156,  157,  159,  161,  163,  165,  166,  168,  170,  172,  174,  175,  177,  179,  181,  183,  184,  186,  188,  190,  192,  194,  195,  197,  199,  201,  203,  204,  206,  208,  210,  212,  213,  215,  217,  219,  221,  222,  224,  226,  228,  230,},
  {  115,  117,  119,  121,  122,  124,  126,  128,  130,  131,  133,  135,  137,  139,  140,  142,  144,  146,  148,  149,  151,  153,  155,  157,  158,  160,  162,  164,  166,  167,  169,  171,  173,  175,  176,  178,  180,  182,  184,  185,  187,  189,  191,  193,  194,  196,  198,  200,  202,  203,  205,  207,  209,  211,  212,  214,  216,  218,  220,  221,  223,  225,  227,  229,},
  {  115,  117,  118,  120,  122,  124,  126,  127,  129,  131,  133,  134,  136,  138,  140,  142,  143,  145,  147,  149,  151,  152,  154,  156,  158,  160,  161,  163,  165,  167,  169,  170,  172,  174,  176,  178,  179,  181,  183,  185,  186,  188,  190,  192,  194,  195,  197,  199,  201,  203,  204,  206,  208,  210,  212,  213,  215,  217,  219,  221,  222,  224,  226,  228,},
  {  114,  116,  118,  120,  121,  123,  125,  127,  129,  130,  132,  134,  136,  137,  139,  141,  143,  145,  146,  148,  150,  152,  154,  155,  157,  159,  161,  162,  164,  166,  168,  170,  171,  173,  175,  177,  179,  180,  182,  184,  186,  187,  189,  191,  193,  195,  196,  198,  200,  202,  204,  205,  207,  209,  211,  212,  214,  216,  218,  220,  221,  223,  225,  227,},
  {  114,  116,  117,  119,  121,  123,  124,  126,  128,  130,  132,  133,  135,  137,  139,  140,  142,  144,  146,  148,  149,  151,  153,  155,  156,  158,  160,  162,  164,  165,  167,  169,  171,  172,  174,  176,  178,  180,  181,  183,  185,  187,  188,  190,  192,  194,  196,  197,  199,  201,  203,  204,  206,  208,  210,  212,  213,  215,  217,  219,  220,  222,  224,  226,},
  {  113,  115,  117,  119,  120,  122,  124,  126,  127,  129,  131,  133,  134,  136,  138,  140,  142,  143,  145,  147,  149,  150,  152,  154,  156,  157,  159,  161,  163,  165,  166,  168,  170,  172,  173,  175,  177,  179,  180,  182,  184,  186,  188,  189,  191,  193,  195,  196,  198,  200,  202,  203,  205,  207,  209,  211,  212,  214,  216,  218,  219,  221,  223,  225,},
  {  113,  115,  116,  118,  120,  122,  123,  125,  127,  129,  130,  132,  134,  136,  137,  139,  141,  143,  144,  146,  148,  150,  152,  153,  155,  157,  159,  160,  162,  164,  166,  167,  169,  171,  173,  174,  176,  178,  180,  181,  183,  185,  187,  189,  190,  192,  194,  196,  197,  199,  201,  203,  204,  206,  208,  210,  211,  213,  215,  217,  218,  220,  222,  224,},
  {  112,  114,  116,  118,  119,  121,  123,  125,  126,  128,  130,  132,  133,  135,  137,  139,  140,  142,  144,  146,  147,  149,  151,  153,  154,  156,  158,  160,  161,  163,  165,  167,  168,  170,  172,  174,  175,  177,  179,  181,  182,  184,  186,  188,  189,  191,  193,  195,  196,  198,  200,  202,  203,  205,  207,  209,  210,  212,  214,  216,  217,  219,  221,  223,},
  {  112,  113,  115,  117,  119,  120,  122,  124,  126,  127,  129,  131,  133,  134,  136,  138,  140,  141,  143,  145,  147,  148,  150,  152,  154,  155,  157,  159,  161,  162,  164,  166,  168,  169,  171,  173,  175,  176,  178,  180,  182,  183,  185,  187,  189,  190,  192,  194,  196,  197,  199,  201,  203,  204,  206,  208,  210,  211,  213,  215,  217,  218,  220,  222,},
  {  111,  113,  115,  116,  118,  120,  122,  123,  125,  127,  129,  130,  132,  134,  136,  137,  139,  141,  143,  144,  146,  148,  149,  151,  153,  155,  156,  158,  160,  162,  163,  165,  167,  169,  170,  172,  174,  176,  177,  179,  181,  183,  184,  186,  188,  189,  191,  193,  195,  196,  198,  200,  202,  203,  205,  207,  209,  210,  212,  214,  216,  217,  219,  221,},
  {  111,  112,  114,  116,  118,  119,  121,  123,  125,  126,  128,  130,  132,  133,  135,  137,  138,  140,  142,  144,  145,  147,  149,  151,  152,  154,  156,  157,  159,  161,  163,  164,  166,  168,  170,  171,  173,  175,  177,  178,  180,  182,  183,  185,  187,  189,  190,  192,  194,  196,  197,  199,  201,  202,  204,  206,  208,  209,  211,  213,  215,  216,  218,  220,},
  {  110,  112,  114,  115,  117,  119,  121,  122,  124,  126,  127,  129,  131,  133,  134,  136,  138,  140,  141,  143,  145,  146,  148,  150,  152,  153,  155,  157,  158,  160,  162,  164,  165,  167,  169,  171,  172,  174,  176,  177,  179,  181,  183,  184,  186,  188,  189,  191,  193,  195,  196,  198,  200,  202,  203,  205,  207,  208,  210,  212,  214,  215,  217,  219,},
  {  110,  111,  113,  115,  117,  118,  120,  122,  123,  125,  127,  129,  130,  132,  134,  135,  137,  139,  141,  142,  144,  146,  147,  149,  151,  153,  154,  156,  158,  159,  161,  163,  165,  166,  168,  170,  171,  173,  175,  177,  178,  180,  182,  183,  185,  187,  189,  190,  192,  194,  195,  197,  199,  201,  202,  204,  206,  207,  209,  211,  213,  214,  216,  218,},
  {  109,  111,  113,  114,  116,  118,  119,  121,  123,  125,  126,  128,  130,  131,  133,  135,  137,  138,  140,  142,  143,  145,  147,  149,  150,  152,  154,  155,  157,  159,  160,  162,  164,  166,  167,  169,  171,  172,  174,  176,  178,  179,  181,  183,  184,  186,  188,  189,  191,  193,  195,  196,  198,  200,  201,  203,  205,  207,  208,  210,  212,  213,  215,  217,},
  {  109,  110,  112,  114,  116,  117,  119,  121,  122,  124,  126,  127,  129,  131,  133,  134,  136,  138,  139,  141,  143,  144,  146,  148,  150,  151,  153,  155,  156,  158,  160,  161,  163,  165,  167,  168,  170,  172,  173,  175,  177,  178,  180,  182,  184,  185,  187,  189,  190,  192,  194,  195,  197,  199,  201,  202,  204,  206,  207,  209,  211,  212,  214,  216,},
  {  108,  110,  112,  113,  115,  117,  118,  120,  122,  123,  125,  127,  129,  130,  132,  134,  135,  137,  139,  140,  142,  144,  145,  147,  149,  151,  152,  154,  156,  157,  159,  161,  162,  164,  166,  167,  169,  171,  173,  174,  176,  178,  179,  181,  183,  184,  186,  188,  189,  191,  193,  195,  196,  198,  200,  201,  203,  205,  206,  208,  210,  211,  213,  215,},
  {  108,  109,  111,  113,  114,  116,  118,  120,  121,  123,  125,  126,  128,  130,  131,  133,  135,  136,  138,  140,  141,  143,  145,  146,  148,  150,  152,  153,  155,  157,  158,  160,  162,  163,  165,  167,  168,  170,  172,  173,  175,  177,  178,  180,  182,  184,  185,  187,  189,  190,  192,  194,  195,  197,  199,  200,  202,  204,  205,  207,  209,  210,  212,  214,},
  {  107,  109,  111,  112,  114,  116,  117,  119,  121,  122,  124,  126,  127,  129,  131,  132,  134,  136,  137,  139,  141,  142,  144,  146,  147,  149,  151,  152,  154,  156,  158,  159,  161,  163,  164,  166,  168,  169,  171,  173,  174,  176,  178,  179,  181,  183,  184,  186,  188,  189,  191,  193,  194,  196,  198,  199,  201,  203,  204,  206,  208,  209,  211,  213,},
  {  107,  108,  110,  112,  113,  115,  117,  118,  120,  122,  123,  125,  127,  128,  130,  132,  133,  135,  137,  138,  140,  142,  143,  145,  147,  148,  150,  152,  153,  155,  157,  158,  160,  162,  163,  165,  167,  168,  170,  172,  173,  175,  177,  178,  180,  182,  183,  185,  187,  188,  190,  192,  193,  195,  197,  198,  200,  202,  203,  205,  207,  208,  210,  212,},
  {  106,  108,  110,  111,  113,  115,  116,  118,  120,  121,  123,  125,  126,  128,  129,  131,  133,  134,  136,  138,  139,  141,  143,  144,  146,  148,  149,  151,  153,  154,  156,  158,  159,  161,  163,  164,  166,  168,  169,  171,  173,  174,  176,  178,  179,  181,  183,  184,  186,  188,  189,  191,  193,  194,  196,  198,  199,  201,  203,  204,  206,  208,  209,  211,},
  {  106,  107,  109,  111,  112,  114,  116,  117,  119,  121,  122,  124,  126,  127,  129,  131,  132,  134,  135,  137,  139,  140,  142,  144,  145,  147,  149,  150,  152,  154,  155,  157,  159,  160,  162,  164,  165,  167,  169,  170,  172,  173,  175,  177,  178,  180,  182,  183,  185,  187,  188,  190,  192,  193,  195,  197,  198,  200,  202,  203,  205,  207,  208,  210,},
  {  105,  107,  109,  110,  112,  113,  115,  117,  118,  120,  122,  123,  125,  127,  128,  130,  132,  133,  135,  136,  138,  140,  141,  143,  145,  146,  148,  150,  151,  153,  155,  156,  158,  160,  161,  163,  164,  166,  168,  169,  171,  173,  174,  176,  178,  179,  181,  183,  184,  186,  187,  189,  191,  192,  194,  196,  197,  199,  201,  202,  204,  206,  207,  209,},
  {  105,  106,  108,  110,  111,  113,  115,  116,  118,  119,  121,  123,  124,  126,  128,  129,  131,  133,  134,  136,  137,  139,  141,  142,  144,  146,  147,  149,  151,  152,  154,  155,  157,  159,  160,  162,  164,  165,  167,  169,  170,  172,  173,  175,  177,  178,  180,  182,  183,  185,  187,  188,  190,  191,  193,  195,  196,  198,  200,  201,  203,  205,  206,  208,},
  {  104,  106,  108,  109,  111,  112,  114,  116,  117,  119,  121,  122,  124,  125,  127,  129,  130,  132,  134,  135,  137,  138,  140,  142,  143,  145,  147,  148,  150,  151,  153,  155,  156,  158,  160,  161,  163,  165,  166,  168,  169,  171,  173,  174,  176,  178,  179,  181,  182,  184,  186,  187,  189,  191,  192,  194,  195,  197,  199,  200,  202,  204,  205,  207,},
  {  104,  105,  107,  109,  110,  112,  113,  115,  117,  118,  120,  122,  123,  125,  126,  128,  130,  131,  133,  135,  136,  138,  139,  141,  143,  144,  146,  148,  149,  151,  152,  154,  156,  157,  159,  160,  162,  164,  165,  167,  169,  170,  172,  173,  175,  177,  178,  180,  182,  183,  185,  186,  188,  190,  191,  193,  195,  196,  198,  199,  201,  203,  204,  206,},
  {  103,  105,  106,  108,  110,  111,  113,  115,  116,  118,  119,  121,  123,  124,  126,  127,  129,  131,  132,  134,  136,  137,  139,  140,  142,  144,  145,  147,  148,  150,  152,  153,  155,  156,  158,  160,  161,  163,  165,  166,  168,  169,  171,  173,  174,  176,  177,  179,  181,  182,  184,  186,  187,  189,  190,  192,  194,  195,  197,  198,  200,  202,  203,  205,},
  {  103,  104,  106,  108,  109,  111,  112,  114,  116,  117,  119,  120,  122,  124,  125,  127,  128,  130,  132,  133,  135,  136,  138,  140,  141,  143,  144,  146,  148,  149,  151,  153,  154,  156,  157,  159,  161,  162,  164,  165,  167,  169,  170,  172,  173,  175,  177,  178,  180,  181,  183,  185,  186,  188,  189,  191,  193,  194,  196,  197,  199,  201,  202,  204,},
  {  102,  104,  105,  107,  109,  110,  112,  113,  115,  117,  118,  120,  121,  123,  125,  126,  128,  129,  131,  133,  134,  136,  137,  139,  141,  142,  144,  145,  147,  149,  150,  152,  153,  155,  157,  158,  160,  161,  163,  165,  166,  168,  169,  171,  173,  174,  176,  177,  179,  181,  182,  184,  185,  187,  189,  190,  192,  193,  195,  197,  198,  200,  201,  203,},
  {  102,  103,  105,  107,  108,  110,  111,  113,  114,  116,  118,  119,  121,  122,  124,  126,  127,  129,  130,  132,  134,  135,  137,  138,  140,  141,  143,  145,  146,  148,  149,  151,  153,  154,  156,  157,  159,  161,  162,  164,  165,  167,  169,  170,  172,  173,  175,  176,  178,  180,  181,  183,  184,  186,  188,  189,  191,  192,  194,  196,  197,  199,  200,  202,},
  {  101,  103,  104,  106,  108,  109,  111,  112,  114,  115,  117,  119,  120,  122,  123,  125,  127,  128,  130,  131,  133,  134,  136,  138,  139,  141,  142,  144,  146,  147,  149,  150,  152,  153,  155,  157,  158,  160,  161,  163,  165,  166,  168,  169,  171,  172,  174,  176,  177,  179,  180,  182,  184,  185,  187,  188,  190,  191,  193,  195,  196,  198,  199,  201,},
  {  101,  102,  104,  105,  107,  109,  110,  112,  113,  115,  116,  118,  120,  121,  123,  124,  126,  128,  129,  131,  132,  134,  135,  137,  139,  140,  142,  143,  145,  146,  148,  150,  151,  153,  154,  156,  157,  159,  161,  162,  164,  165,  167,  168,  170,  172,  173,  175,  176,  178,  179,  181,  183,  184,  186,  187,  189,  190,  192,  194,  195,  197,  198,  200,},
  {  100,  102,  103,  105,  107,  108,  110,  111,  113,  114,  116,  117,  119,  121,  122,  124,  125,  127,  128,  130,  132,  133,  135,  136,  138,  139,  141,  143,  144,  146,  147,  149,  150,  152,  154,  155,  157,  158,  160,  161,  163,  164,  166,  168,  169,  171,  172,  174,  175,  177,  179,  180,  182,  183,  185,  186,  188,  190,  191,  193,  194,  196,  197,  199,},
  {  100,  101,  103,  104,  106,  108,  109,  111,  112,  114,  115,  117,  118,  120,  122,  123,  125,  126,  128,  129,  131,  132,  134,  136,  137,  139,  140,  142,  143,  145,  147,  148,  150,  151,  153,  154,  156,  157,  159,  161,  162,  164,  165,  167,  168,  170,  171,  173,  175,  176,  178,  179,  181,  182,  184,  185,  187,  189,  190,  192,  193,  195,  196,  198,},
  {   99,  101,  102,  104,  105,  107,  109,  110,  112,  113,  115,  116,  118,  119,  121,  123,  124,  126,  127,  129,  130,  132,  133,  135,  136,  138,  140,  141,  143,  144,  146,  147,  149,  150,  152,  154,  155,  157,  158,  160,  161,  163,  164,  166,  167,  169,  171,  172,  174,  175,  177,  178,  180,  181,  183,  185,  186,  188,  189,  191,  192,  194,  195,  197,},
  {   99,  100,  102,  103,  105,  106,  108,  110,  111,  113,  114,  116,  117,  119,  120,  122,  123,  125,  127,  128,  130,  131,  133,  134,  136,  137,  139,  140,  142,  143,  145,  147,  148,  150,  151,  153,  154,  156,  157,  159,  160,  162,  164,  165,  167,  168,  170,  171,  173,  174,  176,  177,  179,  181,  182,  184,  185,  187,  188,  190,  191,  193,  194,  196,},
  {   98,  100,  101,  103,  104,  106,  107,  109,  111,  112,  114,  115,  117,  118,  120,  121,  123,  124,  126,  127,  129,  130,  132,  134,  135,  137,  138,  140,  141,  143,  144,  146,  147,  149,  150,  152,  154,  155,  157,  158,  160,  161,  163,  164,  166,  167,  169,  170,  172,  173,  175,  177,  178,  180,  181,  183,  184,  186,  187,  189,  190,  192,  193,  195,},
  {   98,   99,  101,  102,  104,  105,  107,  108,  110,  111,  113,  115,  116,  118,  119,  121,  122,  124,  125,  127,  128,  130,  131,  133,  134,  136,  137,  139,  141,  142,  144,  145,  147,  148,  150,  151,  153,  154,  156,  157,  159,  160,  162,  163,  165,  166,  168,  170,  171,  173,  174,  176,  177,  179,  180,  182,  183,  185,  186,  188,  189,  191,  192,  194,},
  {   97,   99,  100,  102,  103,  105,  106,  108,  109,  111,  112,  114,  115,  117,  119,  120,  122,  123,  125,  126,  128,  129,  131,  132,  134,  135,  137,  138,  140,  141,  143,  144,  146,  147,  149,  150,  152,  153,  155,  157,  158,  160,  161,  163,  164,  166,  167,  169,  170,  172,  173,  175,  176,  178,  179,  181,  182,  184,  185,  187,  188,  190,  191,  193,},
  {   97,   98,  100,  101,  103,  104,  106,  107,  109,  110,  112,  113,  115,  116,  118,  119,  121,  122,  124,  125,  127,  128,  130,  132,  133,  135,  136,  138,  139,  141,  142,  144,  145,  147,  148,  150,  151,  153,  154,  156,  157,  159,  160,  162,  163,  165,  166,  168,  169,  171,  172,  174,  175,  177,  178,  180,  181,  183,  184,  186,  187,  189,  190,  192,},
  {   96,   98,   99,  101,  102,  104,  105,  107,  108,  110,  111,  113,  114,  116,  117,  119,  120,  122,  123,  125,  126,  128,  129,  131,  132,  134,  135,  137,  138,  140,  141,  143,  144,  146,  147,  149,  150,  152,  153,  155,  156,  158,  159,  161,  162,  164,  165,  167,  168,  170,  171,  173,  174,  176,  177,  179,  180,  182,  183,  185,  186,  188,  189,  191,},
  {   96,   97,   99,  100,  102,  103,  105,  106,  108,  109,  111,  112,  114,  115,  117,  118,  120,  121,  123,  124,  126,  127,  129,  130,  132,  133,  135,  136,  138,  139,  141,  142,  144,  145,  147,  148,  150,  151,  153,  154,  156,  157,  159,  160,  162,  163,  165,  166,  168,  169,  171,  172,  174,  175,  177,  178,  180,  181,  183,  184,  186,  187,  189,  190,},
  {   95,   97,   98,  100,  101,  103,  104,  106,  107,  109,  110,  112,  113,  115,  116,  118,  119,  121,  122,  124,  125,  127,  128,  129,  131,  132,  134,  135,  137,  138,  140,  141,  143,  144,  146,  147,  149,  150,  152,  153,  155,  156,  158,  159,  161,  162,  164,  165,  167,  168,  170,  171,  173,  174,  176,  177,  179,  180,  182,  183,  185,  186,  188,  189,},
  {   95,   96,   98,   99,  101,  102,  104,  105,  107,  108,  110,  111,  113,  114,  115,  117,  118,  120,  121,  123,  124,  126,  127,  129,  130,  132,  133,  135,  136,  138,  139,  141,  142,  144,  145,  147,  148,  150,  151,  152,  154,  155,  157,  158,  160,  161,  163,  164,  166,  167,  169,  170,  172,  173,  175,  176,  178,  179,  181,  182,  184,  185,  187,  188,},
  {   94,   96,   97,   99,  100,  102,  103,  105,  106,  108,  109,  110,  112,  113,  115,  116,  118,  119,  121,  122,  124,  125,  127,  128,  130,  131,  133,  134,  135,  137,  138,  140,  141,  143,  144,  146,  147,  149,  150,  152,  153,  155,  156,  158,  159,  161,  162,  163,  165,  166,  168,  169,  171,  172,  174,  175,  177,  178,  180,  181,  183,  184,  186,  187,},
  {   94,   95,   97,   98,  100,  101,  103,  104,  105,  107,  108,  110,  111,  113,  114,  116,  117,  119,  120,  122,  123,  125,  126,  127,  129,  130,  132,  133,  135,  136,  138,  139,  141,  142,  144,  145,  146,  148,  149,  151,  152,  154,  155,  157,  158,  160,  161,  163,  164,  166,  167,  168,  170,  171,  173,  174,  176,  177,  179,  180,  182,  183,  185,  186,},
  {   93,   95,   96,   98,   99,  101,  102,  103,  105,  106,  108,  109,  111,  112,  114,  115,  117,  118,  119,  121,  122,  124,  125,  127,  128,  130,  131,  133,  134,  136,  137,  138,  140,  141,  143,  144,  146,  147,  149,  150,  152,  153,  154,  156,  157,  159,  160,  162,  163,  165,  166,  168,  169,  170,  172,  173,  175,  176,  178,  179,  181,  182,  184,  185,},
  {   93,   94,   96,   97,   99,  100,  101,  103,  104,  106,  107,  109,  110,  112,  113,  114,  116,  117,  119,  120,  122,  123,  125,  126,  128,  129,  130,  132,  133,  135,  136,  138,  139,  141,  142,  143,  145,  146,  148,  149,  151,  152,  154,  155,  157,  158,  159,  161,  162,  164,  165,  167,  168,  170,  171,  172,  174,  175,  177,  178,  180,  181,  183,  184,},
  {   92,   94,   95,   97,   98,   99,  101,  102,  104,  105,  107,  108,  110,  111,  112,  114,  115,  117,  118,  120,  121,  123,  124,  125,  127,  128,  130,  131,  133,  134,  135,  137,  138,  140,  141,  143,  144,  146,  147,  148,  150,  151,  153,  154,  156,  157,  159,  160,  161,  163,  164,  166,  167,  169,  170,  172,  173,  174,  176,  177,  179,  180,  182,  183,},
  {   92,   93,   95,   96,   97,   99,  100,  102,  103,  105,  106,  108,  109,  110,  112,  113,  115,  116,  118,  119,  120,  122,  123,  125,  126,  128,  129,  130,  132,  133,  135,  136,  138,  139,  140,  142,  143,  145,  146,  148,  149,  151,  152,  153,  155,  156,  158,  159,  161,  162,  163,  165,  166,  168,  169,  171,  172,  173,  175,  176,  178,  179,  181,  182,},
  {   91,   93,   94,   96,   97,   98,  100,  101,  103,  104,  106,  107,  108,  110,  111,  113,  114,  115,  117,  118,  120,  121,  123,  124,  125,  127,  128,  130,  131,  133,  134,  135,  137,  138,  140,  141,  143,  144,  145,  147,  148,  150,  151,  153,  154,  155,  157,  158,  160,  161,  163,  164,  165,  167,  168,  170,  171,  173,  174,  175,  177,  178,  180,  181,},
  {   91,   92,   94,   95,   96,   98,   99,  101,  102,  104,  105,  106,  108,  109,  111,  112,  113,  115,  116,  118,  119,  121,  122,  123,  125,  126,  128,  129,  130,  132,  133,  135,  136,  138,  139,  140,  142,  143,  145,  146,  147,  149,  150,  152,  153,  155,  156,  157,  159,  160,  162,  163,  164,  166,  167,  169,  170,  172,  173,  174,  176,  177,  179,  180,},
  {   90,   92,   93,   94,   96,   97,   99,  100,  102,  103,  104,  106,  107,  109,  110,  111,  113,  114,  116,  117,  118,  120,  121,  123,  124,  126,  127,  128,  130,  131,  133,  134,  135,  137,  138,  140,  141,  142,  144,  145,  147,  148,  149,  151,  152,  154,  155,  157,  158,  159,  161,  162,  164,  165,  166,  168,  169,  171,  172,  173,  175,  176,  178,  179,},
  {   90,   91,   93,   94,   95,   97,   98,  100,  101,  102,  104,  105,  107,  108,  109,  111,  112,  114,  115,  116,  118,  119,  121,  122,  123,  125,  126,  128,  129,  130,  132,  133,  135,  136,  137,  139,  140,  142,  143,  144,  146,  147,  149,  150,  151,  153,  154,  156,  157,  158,  160,  161,  163,  164,  165,  167,  168,  170,  171,  172,  174,  175,  177,  178,},
  {   89,   91,   92,   93,   95,   96,   98,   99,  100,  102,  103,  105,  106,  107,  109,  110,  112,  113,  114,  116,  117,  119,  120,  121,  123,  124,  126,  127,  128,  130,  131,  132,  134,  135,  137,  138,  139,  141,  142,  144,  145,  146,  148,  149,  151,  152,  153,  155,  156,  158,  159,  160,  162,  163,  165,  166,  167,  169,  170,  172,  173,  174,  176,  177,},
  {   89,   90,   92,   93,   94,   96,   97,   98,  100,  101,  103,  104,  105,  107,  108,  110,  111,  112,  114,  115,  116,  118,  119,  121,  122,  123,  125,  126,  128,  129,  130,  132,  133,  135,  136,  137,  139,  140,  141,  143,  144,  146,  147,  148,  150,  151,  153,  154,  155,  157,  158,  159,  161,  162,  164,  165,  166,  168,  169,  171,  172,  173,  175,  176,},
  {   88,   90,   91,   92,   94,   95,   97,   98,   99,  101,  102,  103,  105,  106,  108,  109,  110,  112,  113,  114,  116,  117,  119,  120,  121,  123,  124,  125,  127,  128,  130,  131,  132,  134,  135,  137,  138,  139,  141,  142,  143,  145,  146,  148,  149,  150,  152,  153,  154,  156,  157,  159,  160,  161,  163,  164,  165,  167,  168,  170,  171,  172,  174,  175,},
  {   88,   89,   90,   92,   93,   95,   96,   97,   99,  100,  101,  103,  104,  106,  107,  108,  110,  111,  112,  114,  115,  117,  118,  119,  121,  122,  123,  125,  126,  128,  129,  130,  132,  133,  134,  136,  137,  138,  140,  141,  143,  144,  145,  147,  148,  149,  151,  152,  154,  155,  156,  158,  159,  160,  162,  163,  165,  166,  167,  169,  170,  171,  173,  174,},
  {   87,   89,   90,   91,   93,   94,   95,   97,   98,  100,  101,  102,  104,  105,  106,  108,  109,  110,  112,  113,  115,  116,  117,  119,  120,  121,  123,  124,  125,  127,  128,  130,  131,  132,  134,  135,  136,  138,  139,  140,  142,  143,  145,  146,  147,  149,  150,  151,  153,  154,  155,  157,  158,  160,  161,  162,  164,  165,  166,  168,  169,  170,  172,  173,},
  {   87,   88,   89,   91,   92,   94,   95,   96,   98,   99,  100,  102,  103,  104,  106,  107,  108,  110,  111,  113,  114,  115,  117,  118,  119,  121,  122,  123,  125,  126,  127,  129,  130,  131,  133,  134,  136,  137,  138,  140,  141,  142,  144,  145,  146,  148,  149,  150,  152,  153,  155,  156,  157,  159,  160,  161,  163,  164,  165,  167,  168,  169,  171,  172,},
  {   86,   88,   89,   90,   92,   93,   94,   96,   97,   98,  100,  101,  102,  104,  105,  106,  108,  109,  111,  112,  113,  115,  116,  117,  119,  120,  121,  123,  124,  125,  127,  128,  129,  131,  132,  133,  135,  136,  137,  139,  140,  142,  143,  144,  146,  147,  148,  150,  151,  152,  154,  155,  156,  158,  159,  160,  162,  163,  164,  166,  167,  168,  170,  171,},
  {   86,   87,   88,   90,   91,   92,   94,   95,   96,   98,   99,  100,  102,  103,  105,  106,  107,  109,  110,  111,  113,  114,  115,  117,  118,  119,  121,  122,  123,  125,  126,  127,  129,  130,  131,  133,  134,  135,  137,  138,  139,  141,  142,  143,  145,  146,  147,  149,  150,  151,  153,  154,  155,  157,  158,  159,  161,  162,  163,  165,  166,  167,  169,  170,},
  {   85,   87,   88,   89,   91,   92,   93,   95,   96,   97,   99,  100,  101,  103,  104,  105,  107,  108,  109,  111,  112,  113,  115,  116,  117,  119,  120,  121,  123,  124,  125,  127,  128,  129,  131,  132,  133,  135,  136,  137,  139,  140,  141,  143,  144,  145,  147,  148,  149,  151,  152,  153,  155,  156,  157,  159,  160,  161,  163,  164,  165,  167,  168,  169,},
  {   85,   86,   87,   89,   90,   91,   93,   94,   95,   97,   98,   99,  101,  102,  103,  105,  106,  107,  109,  110,  111,  113,  114,  115,  117,  118,  119,  121,  122,  123,  124,  126,  127,  128,  130,  131,  132,  134,  135,  136,  138,  139,  140,  142,  143,  144,  146,  147,  148,  150,  151,  152,  154,  155,  156,  158,  159,  160,  162,  163,  164,  166,  167,  168,},
  {   84,   86,   87,   88,   90,   91,   92,   93,   95,   96,   97,   99,  100,  101,  103,  104,  105,  107,  108,  109,  111,  112,  113,  115,  116,  117,  118,  120,  121,  122,  124,  125,  126,  128,  129,  130,  132,  133,  134,  136,  137,  138,  140,  141,  142,  143,  145,  146,  147,  149,  150,  151,  153,  154,  155,  157,  158,  159,  161,  162,  163,  165,  166,  167,},
  {   84,   85,   86,   88,   89,   90,   92,   93,   94,   96,   97,   98,   99,  101,  102,  103,  105,  106,  107,  109,  110,  111,  113,  114,  115,  116,  118,  119,  120,  122,  123,  124,  126,  127,  128,  130,  131,  132,  133,  135,  136,  137,  139,  140,  141,  143,  144,  145,  147,  148,  149,  150,  152,  153,  154,  156,  157,  158,  160,  161,  162,  164,  165,  166,},
  {   83,   85,   86,   87,   88,   90,   91,   92,   94,   95,   96,   98,   99,  100,  101,  103,  104,  105,  107,  108,  109,  111,  112,  113,  114,  116,  117,  118,  120,  121,  122,  124,  125,  126,  127,  129,  130,  131,  133,  134,  135,  137,  138,  139,  140,  142,  143,  144,  146,  147,  148,  150,  151,  152,  153,  155,  156,  157,  159,  160,  161,  163,  164,  165,},
  {   83,   84,   85,   87,   88,   89,   91,   92,   93,   94,   96,   97,   98,  100,  101,  102,  103,  105,  106,  107,  109,  110,  111,  112,  114,  115,  116,  118,  119,  120,  122,  123,  124,  125,  127,  128,  129,  131,  132,  133,  134,  136,  137,  138,  140,  141,  142,  144,  145,  146,  147,  149,  150,  151,  153,  154,  155,  156,  158,  159,  160,  162,  163,  164,},
  {   82,   84,   85,   86,   87,   89,   90,   91,   93,   94,   95,   96,   98,   99,  100,  102,  103,  104,  105,  107,  108,  109,  111,  112,  113,  114,  116,  117,  118,  120,  121,  122,  123,  125,  126,  127,  129,  130,  131,  132,  134,  135,  136,  138,  139,  140,  141,  143,  144,  145,  147,  148,  149,  150,  152,  153,  154,  156,  157,  158,  159,  161,  162,  163,},
  {   82,   83,   84,   86,   87,   88,   89,   91,   92,   93,   95,   96,   97,   98,  100,  101,  102,  103,  105,  106,  107,  109,  110,  111,  112,  114,  115,  116,  118,  119,  120,  121,  123,  124,  125,  126,  128,  129,  130,  132,  133,  134,  135,  137,  138,  139,  141,  142,  143,  144,  146,  147,  148,  149,  151,  152,  153,  155,  156,  157,  158,  160,  161,  162,},
  {   81,   83,   84,   85,   86,   88,   89,   90,   91,   93,   94,   95,   96,   98,   99,  100,  102,  103,  104,  105,  107,  108,  109,  110,  112,  113,  114,  116,  117,  118,  119,  121,  122,  123,  124,  126,  127,  128,  129,  131,  132,  133,  135,  136,  137,  138,  140,  141,  142,  143,  145,  146,  147,  149,  150,  151,  152,  154,  155,  156,  157,  159,  160,  161,},
  {   81,   82,   83,   85,   86,   87,   88,   90,   91,   92,   93,   95,   96,   97,   98,  100,  101,  102,  103,  105,  106,  107,  109,  110,  111,  112,  114,  115,  116,  117,  119,  120,  121,  122,  124,  125,  126,  127,  129,  130,  131,  132,  134,  135,  136,  138,  139,  140,  141,  143,  144,  145,  146,  148,  149,  150,  151,  153,  154,  155,  156,  158,  159,  160,},
  {   80,   82,   83,   84,   85,   87,   88,   89,   90,   92,   93,   94,   95,   97,   98,   99,  100,  102,  103,  104,  105,  107,  108,  109,  110,  112,  113,  114,  115,  117,  118,  119,  120,  122,  123,  124,  125,  127,  128,  129,  130,  132,  133,  134,  135,  137,  138,  139,  140,  142,  143,  144,  145,  147,  148,  149,  150,  152,  153,  154,  155,  157,  158,  159,},
  {   80,   81,   82,   83,   85,   86,   87,   88,   90,   91,   92,   93,   95,   96,   97,   98,  100,  101,  102,  103,  105,  106,  107,  108,  110,  111,  112,  113,  115,  116,  117,  118,  120,  121,  122,  123,  125,  126,  127,  128,  130,  131,  132,  133,  135,  136,  137,  138,  140,  141,  142,  143,  145,  146,  147,  148,  150,  151,  152,  153,  155,  156,  157,  158,},
  {   79,   80,   82,   83,   84,   85,   87,   88,   89,   90,   92,   93,   94,   95,   97,   98,   99,  100,  102,  103,  104,  105,  106,  108,  109,  110,  111,  113,  114,  115,  116,  118,  119,  120,  121,  123,  124,  125,  126,  128,  129,  130,  131,  132,  134,  135,  136,  137,  139,  140,  141,  142,  144,  145,  146,  147,  149,  150,  151,  152,  154,  155,  156,  157,},
  {   79,   80,   81,   82,   84,   85,   86,   87,   89,   90,   91,   92,   94,   95,   96,   97,   98,  100,  101,  102,  103,  105,  106,  107,  108,  110,  111,  112,  113,  114,  116,  117,  118,  119,  121,  122,  123,  124,  126,  127,  128,  129,  130,  132,  133,  134,  135,  137,  138,  139,  140,  142,  143,  144,  145,  146,  148,  149,  150,  151,  153,  154,  155,  156,},
  {   78,   79,   81,   82,   83,   84,   86,   87,   88,   89,   90,   92,   93,   94,   95,   97,   98,   99,  100,  101,  103,  104,  105,  106,  108,  109,  110,  111,  112,  114,  115,  116,  117,  119,  120,  121,  122,  123,  125,  126,  127,  128,  130,  131,  132,  133,  134,  136,  137,  138,  139,  141,  142,  143,  144,  145,  147,  148,  149,  150,  152,  153,  154,  155,},
  {   78,   79,   80,   81,   83,   84,   85,   86,   87,   89,   90,   91,   92,   94,   95,   96,   97,   98,  100,  101,  102,  103,  104,  106,  107,  108,  109,  111,  112,  113,  114,  115,  117,  118,  119,  120,  121,  123,  124,  125,  126,  128,  129,  130,  131,  132,  134,  135,  136,  137,  138,  140,  141,  142,  143,  145,  146,  147,  148,  149,  151,  152,  153,  154,},
  {   77,   78,   80,   81,   82,   83,   84,   86,   87,   88,   89,   91,   92,   93,   94,   95,   97,   98,   99,  100,  101,  103,  104,  105,  106,  107,  109,  110,  111,  112,  113,  115,  116,  117,  118,  119,  121,  122,  123,  124,  126,  127,  128,  129,  130,  132,  133,  134,  135,  136,  138,  139,  140,  141,  142,  144,  145,  146,  147,  148,  150,  151,  152,  153,},
  {   77,   78,   79,   80,   82,   83,   84,   85,   86,   88,   89,   90,   91,   92,   94,   95,   96,   97,   98,  100,  101,  102,  103,  104,  106,  107,  108,  109,  110,  112,  113,  114,  115,  116,  118,  119,  120,  121,  122,  124,  125,  126,  127,  128,  130,  131,  132,  133,  134,  136,  137,  138,  139,  140,  142,  143,  144,  145,  146,  148,  149,  150,  151,  152,},
  {   76,   77,   79,   80,   81,   82,   83,   85,   86,   87,   88,   89,   91,   92,   93,   94,   95,   97,   98,   99,  100,  101,  102,  104,  105,  106,  107,  108,  110,  111,  112,  113,  114,  116,  117,  118,  119,  120,  122,  123,  124,  125,  126,  127,  129,  130,  131,  132,  133,  135,  136,  137,  138,  139,  141,  142,  143,  144,  145,  147,  148,  149,  150,  151,},
  {   76,   77,   78,   79,   80,   82,   83,   84,   85,   86,   88,   89,   90,   91,   92,   94,   95,   96,   97,   98,   99,  101,  102,  103,  104,  105,  107,  108,  109,  110,  111,  112,  114,  115,  116,  117,  118,  120,  121,  122,  123,  124,  125,  127,  128,  129,  130,  131,  133,  134,  135,  136,  137,  138,  140,  141,  142,  143,  144,  146,  147,  148,  149,  150,},
  {   75,   76,   78,   79,   80,   81,   82,   83,   85,   86,   87,   88,   89,   91,   92,   93,   94,   95,   96,   98,   99,  100,  101,  102,  103,  105,  106,  107,  108,  109,  111,  112,  113,  114,  115,  116,  118,  119,  120,  121,  122,  123,  125,  126,  127,  128,  129,  131,  132,  133,  134,  135,  136,  138,  139,  140,  141,  142,  143,  145,  146,  147,  148,  149,},
  {   75,   76,   77,   78,   79,   81,   82,   83,   84,   85,   86,   88,   89,   90,   91,   92,   93,   95,   96,   97,   98,   99,  100,  102,  103,  104,  105,  106,  107,  109,  110,  111,  112,  113,  114,  116,  117,  118,  119,  120,  121,  123,  124,  125,  126,  127,  128,  130,  131,  132,  133,  134,  135,  137,  138,  139,  140,  141,  142,  144,  145,  146,  147,  148,},
  {   74,   75,   77,   78,   79,   80,   81,   82,   84,   85,   86,   87,   88,   89,   90,   92,   93,   94,   95,   96,   97,   99,  100,  101,  102,  103,  104,  106,  107,  108,  109,  110,  111,  113,  114,  115,  116,  117,  118,  119,  121,  122,  123,  124,  125,  126,  128,  129,  130,  131,  132,  133,  135,  136,  137,  138,  139,  140,  142,  143,  144,  145,  146,  147,},
  {   74,   75,   76,   77,   78,   80,   81,   82,   83,   84,   85,   86,   88,   89,   90,   91,   92,   93,   94,   96,   97,   98,   99,  100,  101,  103,  104,  105,  106,  107,  108,  109,  111,  112,  113,  114,  115,  116,  118,  119,  120,  121,  122,  123,  124,  126,  127,  128,  129,  130,  131,  133,  134,  135,  136,  137,  138,  139,  141,  142,  143,  144,  145,  146,},
  {   73,   74,   76,   77,   78,   79,   80,   81,   82,   84,   85,   86,   87,   88,   89,   90,   92,   93,   94,   95,   96,   97,   98,  100,  101,  102,  103,  104,  105,  106,  108,  109,  110,  111,  112,  113,  114,  116,  117,  118,  119,  120,  121,  122,  124,  125,  126,  127,  128,  129,  130,  132,  133,  134,  135,  136,  137,  138,  140,  141,  142,  143,  144,  145,},
  {   73,   74,   75,   76,   77,   78,   80,   81,   82,   83,   84,   85,   86,   88,   89,   90,   91,   92,   93,   94,   95,   97,   98,   99,  100,  101,  102,  103,  105,  106,  107,  108,  109,  110,  111,  113,  114,  115,  116,  117,  118,  119,  120,  122,  123,  124,  125,  126,  127,  128,  130,  131,  132,  133,  134,  135,  136,  138,  139,  140,  141,  142,  143,  144,},
  {   72,   73,   75,   76,   77,   78,   79,   80,   81,   82,   84,   85,   86,   87,   88,   89,   90,   91,   93,   94,   95,   96,   97,   98,   99,  100,  102,  103,  104,  105,  106,  107,  108,  110,  111,  112,  113,  114,  115,  116,  117,  119,  120,  121,  122,  123,  124,  125,  126,  128,  129,  130,  131,  132,  133,  134,  135,  137,  138,  139,  140,  141,  142,  143,},
  {   72,   73,   74,   75,   76,   77,   78,   80,   81,   82,   83,   84,   85,   86,   87,   89,   90,   91,   92,   93,   94,   95,   96,   98,   99,  100,  101,  102,  103,  104,  105,  107,  108,  109,  110,  111,  112,  113,  114,  115,  117,  118,  119,  120,  121,  122,  123,  124,  126,  127,  128,  129,  130,  131,  132,  133,  135,  136,  137,  138,  139,  140,  141,  142,},
  {   71,   72,   73,   75,   76,   77,   78,   79,   80,   81,   82,   83,   85,   86,   87,   88,   89,   90,   91,   92,   94,   95,   96,   97,   98,   99,  100,  101,  102,  104,  105,  106,  107,  108,  109,  110,  111,  112,  114,  115,  116,  117,  118,  119,  120,  121,  122,  124,  125,  126,  127,  128,  129,  130,  131,  132,  134,  135,  136,  137,  138,  139,  140,  141,},
  {   71,   72,   73,   74,   75,   76,   77,   78,   80,   81,   82,   83,   84,   85,   86,   87,   88,   90,   91,   92,   93,   94,   95,   96,   97,   98,   99,  101,  102,  103,  104,  105,  106,  107,  108,  109,  111,  112,  113,  114,  115,  116,  117,  118,  119,  120,  122,  123,  124,  125,  126,  127,  128,  129,  130,  132,  133,  134,  135,  136,  137,  138,  139,  140,},
  {   70,   71,   72,   74,   75,   76,   77,   78,   79,   80,   81,   82,   83,   85,   86,   87,   88,   89,   90,   91,   92,   93,   94,   95,   97,   98,   99,  100,  101,  102,  103,  104,  105,  106,  108,  109,  110,  111,  112,  113,  114,  115,  116,  117,  119,  120,  121,  122,  123,  124,  125,  126,  127,  128,  130,  131,  132,  133,  134,  135,  136,  137,  138,  139,},
  {   70,   71,   72,   73,   74,   75,   76,   77,   78,   80,   81,   82,   83,   84,   85,   86,   87,   88,   89,   90,   92,   93,   94,   95,   96,   97,   98,   99,  100,  101,  102,  104,  105,  106,  107,  108,  109,  110,  111,  112,  113,  114,  116,  117,  118,  119,  120,  121,  122,  123,  124,  125,  126,  128,  129,  130,  131,  132,  133,  134,  135,  136,  137,  138,},
  {   69,   70,   71,   72,   74,   75,   76,   77,   78,   79,   80,   81,   82,   83,   84,   85,   87,   88,   89,   90,   91,   92,   93,   94,   95,   96,   97,   98,  100,  101,  102,  103,  104,  105,  106,  107,  108,  109,  110,  111,  113,  114,  115,  116,  117,  118,  119,  120,  121,  122,  123,  124,  126,  127,  128,  129,  130,  131,  132,  133,  134,  135,  136,  137,},
  {   69,   70,   71,   72,   73,   74,   75,   76,   77,   78,   79,   81,   82,   83,   84,   85,   86,   87,   88,   89,   90,   91,   92,   93,   95,   96,   97,   98,   99,  100,  101,  102,  103,  104,  105,  106,  107,  108,  110,  111,  112,  113,  114,  115,  116,  117,  118,  119,  120,  121,  122,  124,  125,  126,  127,  128,  129,  130,  131,  132,  133,  134,  135,  136,},
  {   68,   69,   70,   71,   73,   74,   75,   76,   77,   78,   79,   80,   81,   82,   83,   84,   85,   86,   87,   89,   90,   91,   92,   93,   94,   95,   96,   97,   98,   99,  100,  101,  102,  103,  105,  106,  107,  108,  109,  110,  111,  112,  113,  114,  115,  116,  117,  118,  119,  121,  122,  123,  124,  125,  126,  127,  128,  129,  130,  131,  132,  133,  134,  135,},
  {   68,   69,   70,   71,   72,   73,   74,   75,   76,   77,   78,   79,   80,   82,   83,   84,   85,   86,   87,   88,   89,   90,   91,   92,   93,   94,   95,   96,   97,   98,  100,  101,  102,  103,  104,  105,  106,  107,  108,  109,  110,  111,  112,  113,  114,  115,  116,  118,  119,  120,  121,  122,  123,  124,  125,  126,  127,  128,  129,  130,  131,  132,  133,  134,},
  {   67,   68,   69,   70,   71,   73,   74,   75,   76,   77,   78,   79,   80,   81,   82,   83,   84,   85,   86,   87,   88,   89,   90,   91,   92,   94,   95,   96,   97,   98,   99,  100,  101,  102,  103,  104,  105,  106,  107,  108,  109,  110,  111,  112,  113,  115,  116,  117,  118,  119,  120,  121,  122,  123,  124,  125,  126,  127,  128,  129,  130,  131,  132,  133,},
  {   67,   68,   69,   70,   71,   72,   73,   74,   75,   76,   77,   78,   79,   80,   81,   82,   83,   84,   86,   87,   88,   89,   90,   91,   92,   93,   94,   95,   96,   97,   98,   99,  100,  101,  102,  103,  104,  105,  106,  107,  108,  110,  111,  112,  113,  114,  115,  116,  117,  118,  119,  120,  121,  122,  123,  124,  125,  126,  127,  128,  129,  130,  131,  132,},
  {   66,   67,   68,   69,   70,   71,   72,   73,   75,   76,   77,   78,   79,   80,   81,   82,   83,   84,   85,   86,   87,   88,   89,   90,   91,   92,   93,   94,   95,   96,   97,   98,   99,  100,  101,  102,  104,  105,  106,  107,  108,  109,  110,  111,  112,  113,  114,  115,  116,  117,  118,  119,  120,  121,  122,  123,  124,  125,  126,  127,  128,  129,  130,  131,},
  {   66,   67,   68,   69,   70,   71,   72,   73,   74,   75,   76,   77,   78,   79,   80,   81,   82,   83,   84,   85,   86,   87,   88,   89,   90,   91,   92,   93,   95,   96,   97,   98,   99,  100,  101,  102,  103,  104,  105,  106,  107,  108,  109,  110,  111,  112,  113,  114,  115,  116,  117,  118,  119,  120,  121,  122,  123,  124,  125,  126,  127,  128,  129,  130,},
  {   65,   66,   67,   68,   69,   70,   71,   72,   73,   74,   75,   76,   77,   79,   80,   81,   82,   83,   84,   85,   86,   87,   88,   89,   90,   91,   92,   93,   94,   95,   96,   97,   98,   99,  100,  101,  102,  103,  104,  105,  106,  107,  108,  109,  110,  111,  112,  113,  114,  115,  116,  117,  118,  119,  120,  121,  122,  123,  124,  125,  126,  127,  128,  129,},
  {   65,   66,   67,   68,   69,   70,   71,   72,   73,   74,   75,   76,   77,   78,   79,   80,   81,   82,   83,   84,   85,   86,   87,   88,   89,   90,   91,   92,   93,   94,   95,   96,   97,   98,   99,  100,  101,  102,  103,  104,  105,  106,  107,  108,  109,  110,  111,  112,  113,  114,  115,  116,  117,  118,  119,  120,  121,  122,  123,  124,  125,  126,  127,  128,},
  {   64,   65,   66,   67,   68,   69,   70,   71,   72,   73,   74,   75,   76,   77,   78,   79,   80,   81,   82,   83,   84,   85,   86,   87,   88,   89,   90,   91,   92,   93,   94,   95,   96,   97,   98,   99,  100,  101,  102,  103,  104,  105,  106,  107,  108,  109,  110,  111,  112,  113,  114,  115,  116,  117,  118,  119,  120,  121,  122,  123,  124,  125,  126,  127,},
  {   64,   65,   66,   67,   68,   69,   70,   71,   72,   73,   74,   75,   76,   77,   78,   79,   80,   81,   82,   83,   84,   85,   86,   87,   88,   89,   90,   91,   92,   93,   94,   95,   96,   97,   98,   99,  100,  101,  102,  103,  104,  105,  106,  107,  108,  109,  110,  111,  112,  113,  114,  115,  116,  117,  118,  119,  120,  121,  122,  123,  124,  125,  126,  127,},
  {   63,   64,   65,   66,   67,   68,   69,   70,   71,   72,   73,   74,   75,   76,   77,   78,   79,   80,   81,   82,   83,   84,   85,   86,   87,   88,   89,   90,   91,   92,   93,   94,   95,   96,   97,   98,   99,  100,  101,  102,  103,  104,  105,  106,  107,  108,  109,  110,  111,  112,  113,  114,  115,  116,  117,  118,  119,  120,  121,  122,  123,  124,  125,  126,},
  {   63,   64,   65,   66,   67,   68,   69,   70,   71,   72,   73,   74,   75,   75,   76,   77,   78,   79,   80,   81,   82,   83,   84,   85,   86,   87,   88,   89,   90,   91,   92,   93,   94,   95,   96,   97,   98,   99,  100,  101,  102,  103,  104,  105,  106,  107,  108,  109,  110,  111,  112,  113,  114,  115,  116,  117,  118,  119,  120,  121,  122,  123,  124,  125,},
  {   62,   63,   64,   65,   66,   67,   68,   69,   70,   71,   72,   73,   74,   75,   76,   77,   78,   79,   80,   81,   82,   83,   84,   85,   86,   87,   88,   89,   89,   90,   91,   92,   93,   94,   95,   96,   97,   98,   99,  100,  101,  102,  103,  104,  105,  106,  107,  108,  109,  110,  111,  112,  113,  114,  115,  116,  117,  118,  119,  120,  121,  122,  123,  124,},
  {   62,   63,   64,   65,   66,   67,   68,   69,   69,   70,   71,   72,   73,   74,   75,   76,   77,   78,   79,   80,   81,   82,   83,   84,   85,   86,   87,   88,   89,   90,   91,   92,   93,   94,   95,   96,   96,   97,   98,   99,  100,  101,  102,  103,  104,  105,  106,  107,  108,  109,  110,  111,  112,  113,  114,  115,  116,  117,  118,  119,  120,  121,  122,  123,},
  {   61,   62,   63,   64,   65,   66,   67,   68,   69,   70,   71,   72,   73,   74,   75,   76,   77,   78,   78,   79,   80,   81,   82,   83,   84,   85,   86,   87,   88,   89,   90,   91,   92,   93,   94,   95,   96,   97,   98,   99,  100,  100,  101,  102,  103,  104,  105,  106,  107,  108,  109,  110,  111,  112,  113,  114,  115,  116,  117,  118,  119,  120,  121,  122,},
  {   61,   62,   63,   64,   65,   65,   66,   67,   68,   69,   70,   71,   72,   73,   74,   75,   76,   77,   78,   79,   80,   81,   82,   83,   84,   84,   85,   86,   87,   88,   89,   90,   91,   92,   93,   94,   95,   96,   97,   98,   99,  100,  101,  102,  103,  103,  104,  105,  106,  107,  108,  109,  110,  111,  112,  113,  114,  115,  116,  117,  118,  119,  120,  121,},
  {   60,   61,   62,   63,   64,   65,   66,   67,   68,   69,   70,   71,   72,   72,   73,   74,   75,   76,   77,   78,   79,   80,   81,   82,   83,   84,   85,   86,   87,   88,   88,   89,   90,   91,   92,   93,   94,   95,   96,   97,   98,   99,  100,  101,  102,  103,  104,  104,  105,  106,  107,  108,  109,  110,  111,  112,  113,  114,  115,  116,  117,  118,  119,  120,},
  {   60,   61,   62,   63,   63,   64,   65,   66,   67,   68,   69,   70,   71,   72,   73,   74,   75,   76,   77,   77,   78,   79,   80,   81,   82,   83,   84,   85,   86,   87,   88,   89,   90,   91,   91,   92,   93,   94,   95,   96,   97,   98,   99,  100,  101,  102,  103,  104,  105,  105,  106,  107,  108,  109,  110,  111,  112,  113,  114,  115,  116,  117,  118,  119,},
  {   59,   60,   61,   62,   63,   64,   65,   66,   67,   68,   69,   69,   70,   71,   72,   73,   74,   75,   76,   77,   78,   79,   80,   81,   81,   82,   83,   84,   85,   86,   87,   88,   89,   90,   91,   92,   93,   94,   94,   95,   96,   97,   98,   99,  100,  101,  102,  103,  104,  105,  106,  106,  107,  108,  109,  110,  111,  112,  113,  114,  115,  116,  117,  118,},
  {   59,   60,   61,   62,   62,   63,   64,   65,   66,   67,   68,   69,   70,   71,   72,   73,   73,   74,   75,   76,   77,   78,   79,   80,   81,   82,   83,   84,   84,   85,   86,   87,   88,   89,   90,   91,   92,   93,   94,   95,   95,   96,   97,   98,   99,  100,  101,  102,  103,  104,  105,  106,  106,  107,  108,  109,  110,  111,  112,  113,  114,  115,  116,  117,},
  {   58,   59,   60,   61,   62,   63,   64,   65,   66,   66,   67,   68,   69,   70,   71,   72,   73,   74,   75,   76,   76,   77,   78,   79,   80,   81,   82,   83,   84,   85,   86,   86,   87,   88,   89,   90,   91,   92,   93,   94,   95,   96,   96,   97,   98,   99,  100,  101,  102,  103,  104,  105,  106,  106,  107,  108,  109,  110,  111,  112,  113,  114,  115,  116,},
  {   58,   59,   60,   60,   61,   62,   63,   64,   65,   66,   67,   68,   69,   69,   70,   71,   72,   73,   74,   75,   76,   77,   78,   79,   79,   80,   81,   82,   83,   84,   85,   86,   87,   88,   88,   89,   90,   91,   92,   93,   94,   95,   96,   97,   97,   98,   99,  100,  101,  102,  103,  104,  105,  106,  106,  107,  108,  109,  110,  111,  112,  113,  114,  115,},
  {   57,   58,   59,   60,   61,   62,   63,   64,   64,   65,   66,   67,   68,   69,   70,   71,   72,   72,   73,   74,   75,   76,   77,   78,   79,   80,   81,   81,   82,   83,   84,   85,   86,   87,   88,   89,   89,   90,   91,   92,   93,   94,   95,   96,   97,   98,   98,   99,  100,  101,  102,  103,  104,  105,  106,  106,  107,  108,  109,  110,  111,  112,  113,  114,},
  {   57,   58,   59,   59,   60,   61,   62,   63,   64,   65,   66,   67,   67,   68,   69,   70,   71,   72,   73,   74,   74,   75,   76,   77,   78,   79,   80,   81,   82,   82,   83,   84,   85,   86,   87,   88,   89,   90,   90,   91,   92,   93,   94,   95,   96,   97,   98,   98,   99,  100,  101,  102,  103,  104,  105,  106,  106,  107,  108,  109,  110,  111,  112,  113,},
  {   56,   57,   58,   59,   60,   61,   62,   62,   63,   64,   65,   66,   67,   68,   69,   69,   70,   71,   72,   73,   74,   75,   76,   76,   77,   78,   79,   80,   81,   82,   83,   83,   84,   85,   86,   87,   88,   89,   90,   91,   91,   92,   93,   94,   95,   96,   97,   98,   98,   99,  100,  101,  102,  103,  104,  105,  105,  106,  107,  108,  109,  110,  111,  112,},
  {   56,   57,   57,   58,   59,   60,   61,   62,   63,   64,   64,   65,   66,   67,   68,   69,   70,   71,   71,   72,   73,   74,   75,   76,   77,   78,   78,   79,   80,   81,   82,   83,   84,   84,   85,   86,   87,   88,   89,   90,   91,   91,   92,   93,   94,   95,   96,   97,   98,   98,   99,  100,  101,  102,  103,  104,  105,  105,  106,  107,  108,  109,  110,  111,},
  {   55,   56,   57,   58,   59,   60,   60,   61,   62,   63,   64,   65,   66,   66,   67,   68,   69,   70,   71,   72,   73,   73,   74,   75,   76,   77,   78,   79,   79,   80,   81,   82,   83,   84,   85,   85,   86,   87,   88,   89,   90,   91,   92,   92,   93,   94,   95,   96,   97,   98,   98,   99,  100,  101,  102,  103,  104,  104,  105,  106,  107,  108,  109,  110,},
  {   55,   56,   56,   57,   58,   59,   60,   61,   62,   62,   63,   64,   65,   66,   67,   68,   68,   69,   70,   71,   72,   73,   74,   74,   75,   76,   77,   78,   79,   80,   80,   81,   82,   83,   84,   85,   86,   86,   87,   88,   89,   90,   91,   92,   92,   93,   94,   95,   96,   97,   98,   98,   99,  100,  101,  102,  103,  104,  104,  105,  106,  107,  108,  109,},
  {   54,   55,   56,   57,   58,   58,   59,   60,   61,   62,   63,   64,   64,   65,   66,   67,   68,   69,   70,   70,   71,   72,   73,   74,   75,   75,   76,   77,   78,   79,   80,   81,   81,   82,   83,   84,   85,   86,   86,   87,   88,   89,   90,   91,   92,   92,   93,   94,   95,   96,   97,   97,   98,   99,  100,  101,  102,  103,  103,  104,  105,  106,  107,  108,},
  {   54,   55,   55,   56,   57,   58,   59,   60,   60,   61,   62,   63,   64,   65,   66,   66,   67,   68,   69,   70,   71,   71,   72,   73,   74,   75,   76,   76,   77,   78,   79,   80,   81,   81,   82,   83,   84,   85,   86,   87,   87,   88,   89,   90,   91,   92,   92,   93,   94,   95,   96,   97,   97,   98,   99,  100,  101,  102,  102,  103,  104,  105,  106,  107,},
  {   53,   54,   55,   56,   57,   57,   58,   59,   60,   61,   62,   62,   63,   64,   65,   66,   67,   67,   68,   69,   70,   71,   72,   72,   73,   74,   75,   76,   77,   77,   78,   79,   80,   81,   82,   82,   83,   84,   85,   86,   87,   87,   88,   89,   90,   91,   92,   92,   93,   94,   95,   96,   97,   97,   98,   99,  100,  101,  102,  102,  103,  104,  105,  106,},
  {   53,   54,   54,   55,   56,   57,   58,   59,   59,   60,   61,   62,   63,   63,   64,   65,   66,   67,   68,   68,   69,   70,   71,   72,   73,   73,   74,   75,   76,   77,   77,   78,   79,   80,   81,   82,   82,   83,   84,   85,   86,   87,   87,   88,   89,   90,   91,   91,   92,   93,   94,   95,   96,   96,   97,   98,   99,  100,  101,  101,  102,  103,  104,  105,},
  {   52,   53,   54,   55,   56,   56,   57,   58,   59,   60,   60,   61,   62,   63,   64,   64,   65,   66,   67,   68,   69,   69,   70,   71,   72,   73,   73,   74,   75,   76,   77,   78,   78,   79,   80,   81,   82,   82,   83,   84,   85,   86,   87,   87,   88,   89,   90,   91,   91,   92,   93,   94,   95,   96,   96,   97,   98,   99,  100,  100,  101,  102,  103,  104,},
  {   52,   53,   53,   54,   55,   56,   57,   57,   58,   59,   60,   61,   61,   62,   63,   64,   65,   65,   66,   67,   68,   69,   70,   70,   71,   72,   73,   74,   74,   75,   76,   77,   78,   78,   79,   80,   81,   82,   82,   83,   84,   85,   86,   87,   87,   88,   89,   90,   91,   91,   92,   93,   94,   95,   95,   96,   97,   98,   99,   99,  100,  101,  102,  103,},
  {   51,   52,   53,   54,   54,   55,   56,   57,   58,   58,   59,   60,   61,   62,   62,   63,   64,   65,   66,   66,   67,   68,   69,   70,   70,   71,   72,   73,   74,   74,   75,   76,   77,   78,   78,   79,   80,   81,   82,   82,   83,   84,   85,   86,   86,   87,   88,   89,   90,   90,   91,   92,   93,   94,   94,   95,   96,   97,   98,   98,   99,  100,  101,  102,},
  {   51,   52,   52,   53,   54,   55,   56,   56,   57,   58,   59,   59,   60,   61,   62,   63,   63,   64,   65,   66,   67,   67,   68,   69,   70,   71,   71,   72,   73,   74,   75,   75,   76,   77,   78,   79,   79,   80,   81,   82,   82,   83,   84,   85,   86,   86,   87,   88,   89,   90,   90,   91,   92,   93,   94,   94,   95,   96,   97,   98,   98,   99,  100,  101,},
  {   50,   51,   52,   53,   53,   54,   55,   56,   57,   57,   58,   59,   60,   60,   61,   62,   63,   64,   64,   65,   66,   67,   68,   68,   69,   70,   71,   71,   72,   73,   74,   75,   75,   76,   77,   78,   79,   79,   80,   81,   82,   82,   83,   84,   85,   86,   86,   87,   88,   89,   90,   90,   91,   92,   93,   93,   94,   95,   96,   97,   97,   98,   99,  100,},
  {   50,   51,   51,   52,   53,   54,   54,   55,   56,   57,   58,   58,   59,   60,   61,   61,   62,   63,   64,   65,   65,   66,   67,   68,   68,   69,   70,   71,   72,   72,   73,   74,   75,   75,   76,   77,   78,   79,   79,   80,   81,   82,   82,   83,   84,   85,   86,   86,   87,   88,   89,   89,   90,   91,   92,   93,   93,   94,   95,   96,   96,   97,   98,   99,},
  {   49,   50,   51,   52,   52,   53,   54,   55,   55,   56,   57,   58,   58,   59,   60,   61,   62,   62,   63,   64,   65,   65,   66,   67,   68,   68,   69,   70,   71,   72,   72,   73,   74,   75,   75,   76,   77,   78,   78,   79,   80,   81,   82,   82,   83,   84,   85,   85,   86,   87,   88,   88,   89,   90,   91,   92,   92,   93,   94,   95,   95,   96,   97,   98,},
  {   49,   50,   50,   51,   52,   53,   53,   54,   55,   56,   56,   57,   58,   59,   59,   60,   61,   62,   62,   63,   64,   65,   66,   66,   67,   68,   69,   69,   70,   71,   72,   72,   73,   74,   75,   75,   76,   77,   78,   78,   79,   80,   81,   82,   82,   83,   84,   85,   85,   86,   87,   88,   88,   89,   90,   91,   91,   92,   93,   94,   94,   95,   96,   97,},
  {   48,   49,   50,   51,   51,   52,   53,   54,   54,   55,   56,   57,   57,   58,   59,   60,   60,   61,   62,   63,   63,   64,   65,   66,   66,   67,   68,   69,   69,   70,   71,   72,   72,   73,   74,   75,   75,   76,   77,   78,   78,   79,   80,   81,   81,   82,   83,   84,   84,   85,   86,   87,   87,   88,   89,   90,   90,   91,   92,   93,   93,   94,   95,   96,},
  {   48,   48,   49,   50,   51,   51,   52,   53,   54,   54,   55,   56,   57,   57,   58,   59,   60,   60,   61,   62,   63,   63,   64,   65,   66,   66,   67,   68,   69,   69,   70,   71,   72,   72,   73,   74,   75,   75,   76,   77,   78,   78,   79,   80,   81,   81,   82,   83,   84,   84,   85,   86,   87,   87,   88,   89,   90,   90,   91,   92,   93,   93,   94,   95,},
  {   47,   48,   49,   49,   50,   51,   52,   52,   53,   54,   55,   55,   56,   57,   58,   58,   59,   60,   61,   61,   62,   63,   63,   64,   65,   66,   66,   67,   68,   69,   69,   70,   71,   72,   72,   73,   74,   75,   75,   76,   77,   78,   78,   79,   80,   80,   81,   82,   83,   83,   84,   85,   86,   86,   87,   88,   89,   89,   90,   91,   92,   92,   93,   94,},
  {   47,   47,   48,   49,   50,   50,   51,   52,   53,   53,   54,   55,   56,   56,   57,   58,   58,   59,   60,   61,   61,   62,   63,   64,   64,   65,   66,   66,   67,   68,   69,   69,   70,   71,   72,   72,   73,   74,   75,   75,   76,   77,   77,   78,   79,   80,   80,   81,   82,   83,   83,   84,   85,   85,   86,   87,   88,   88,   89,   90,   91,   91,   92,   93,},
  {   46,   47,   48,   48,   49,   50,   51,   51,   52,   53,   53,   54,   55,   56,   56,   57,   58,   59,   59,   60,   61,   61,   62,   63,   64,   64,   65,   66,   66,   67,   68,   69,   69,   70,   71,   72,   72,   73,   74,   74,   75,   76,   77,   77,   78,   79,   79,   80,   81,   82,   82,   83,   84,   85,   85,   86,   87,   87,   88,   89,   90,   90,   91,   92,},
  {   46,   46,   47,   48,   49,   49,   50,   51,   51,   52,   53,   54,   54,   55,   56,   56,   57,   58,   59,   59,   60,   61,   61,   62,   63,   64,   64,   65,   66,   66,   67,   68,   69,   69,   70,   71,   71,   72,   73,   74,   74,   75,   76,   76,   77,   78,   79,   79,   80,   81,   81,   82,   83,   84,   84,   85,   86,   86,   87,   88,   89,   89,   90,   91,},
  {   45,   46,   47,   47,   48,   49,   49,   50,   51,   52,   52,   53,   54,   54,   55,   56,   57,   57,   58,   59,   59,   60,   61,   62,   62,   63,   64,   64,   65,   66,   66,   67,   68,   69,   69,   70,   71,   71,   72,   73,   74,   74,   75,   76,   76,   77,   78,   78,   79,   80,   81,   81,   82,   83,   83,   84,   85,   86,   86,   87,   88,   88,   89,   90,},
  {   45,   45,   46,   47,   48,   48,   49,   50,   50,   51,   52,   52,   53,   54,   55,   55,   56,   57,   57,   58,   59,   59,   60,   61,   62,   62,   63,   64,   64,   65,   66,   66,   67,   68,   69,   69,   70,   71,   71,   72,   73,   73,   74,   75,   76,   76,   77,   78,   78,   79,   80,   80,   81,   82,   83,   83,   84,   85,   85,   86,   87,   87,   88,   89,},
  {   44,   45,   46,   46,   47,   48,   48,   49,   50,   50,   51,   52,   53,   53,   54,   55,   55,   56,   57,   57,   58,   59,   59,   60,   61,   62,   62,   63,   64,   64,   65,   66,   66,   67,   68,   68,   69,   70,   71,   71,   72,   73,   73,   74,   75,   75,   76,   77,   77,   78,   79,   80,   80,   81,   82,   82,   83,   84,   84,   85,   86,   86,   87,   88,},
  {   44,   44,   45,   46,   46,   47,   48,   49,   49,   50,   51,   51,   52,   53,   53,   54,   55,   55,   56,   57,   57,   58,   59,   59,   60,   61,   62,   62,   63,   64,   64,   65,   66,   66,   67,   68,   68,   69,   70,   70,   71,   72,   72,   73,   74,   75,   75,   76,   77,   77,   78,   79,   79,   80,   81,   81,   82,   83,   83,   84,   85,   85,   86,   87,},
  {   43,   44,   45,   45,   46,   47,   47,   48,   49,   49,   50,   51,   51,   52,   53,   53,   54,   55,   55,   56,   57,   57,   58,   59,   59,   60,   61,   61,   62,   63,   64,   64,   65,   66,   66,   67,   68,   68,   69,   70,   70,   71,   72,   72,   73,   74,   74,   75,   76,   76,   77,   78,   78,   79,   80,   80,   81,   82,   82,   83,   84,   84,   85,   86,},
  {   43,   43,   44,   45,   45,   46,   47,   47,   48,   49,   49,   50,   51,   51,   52,   53,   53,   54,   55,   55,   56,   57,   57,   58,   59,   59,   60,   61,   61,   62,   63,   63,   64,   65,   65,   66,   67,   67,   68,   69,   69,   70,   71,   71,   72,   73,   73,   74,   75,   75,   76,   77,   77,   78,   79,   79,   80,   81,   81,   82,   83,   83,   84,   85,},
  {   42,   43,   44,   44,   45,   46,   46,   47,   48,   48,   49,   50,   50,   51,   51,   52,   53,   53,   54,   55,   55,   56,   57,   57,   58,   59,   59,   60,   61,   61,   62,   63,   63,   64,   65,   65,   66,   67,   67,   68,   69,   69,   70,   71,   71,   72,   73,   73,   74,   75,   75,   76,   77,   77,   78,   79,   79,   80,   81,   81,   82,   83,   83,   84,},
  {   42,   42,   43,   44,   44,   45,   46,   46,   47,   48,   48,   49,   50,   50,   51,   52,   52,   53,   53,   54,   55,   55,   56,   57,   57,   58,   59,   59,   60,   61,   61,   62,   63,   63,   64,   65,   65,   66,   67,   67,   68,   68,   69,   70,   70,   71,   72,   72,   73,   74,   74,   75,   76,   76,   77,   78,   78,   79,   80,   80,   81,   82,   82,   83,},
  {   41,   42,   43,   43,   44,   44,   45,   46,   46,   47,   48,   48,   49,   50,   50,   51,   52,   52,   53,   53,   54,   55,   55,   56,   57,   57,   58,   59,   59,   60,   61,   61,   62,   63,   63,   64,   64,   65,   66,   66,   67,   68,   68,   69,   70,   70,   71,   72,   72,   73,   73,   74,   75,   75,   76,   77,   77,   78,   79,   79,   80,   81,   81,   82,},
  {   41,   41,   42,   43,   43,   44,   45,   45,   46,   46,   47,   48,   48,   49,   50,   50,   51,   52,   52,   53,   53,   54,   55,   55,   56,   57,   57,   58,   59,   59,   60,   60,   61,   62,   62,   63,   64,   64,   65,   66,   66,   67,   67,   68,   69,   69,   70,   71,   71,   72,   73,   73,   74,   74,   75,   76,   76,   77,   78,   78,   79,   80,   80,   81,},
  {   40,   41,   42,   42,   43,   43,   44,   45,   45,   46,   47,   47,   48,   48,   49,   50,   50,   51,   52,   52,   53,   53,   54,   55,   55,   56,   57,   57,   58,   58,   59,   60,   60,   61,   62,   62,   63,   64,   64,   65,   65,   66,   67,   67,   68,   69,   69,   70,   70,   71,   72,   72,   73,   74,   74,   75,   75,   76,   77,   77,   78,   79,   79,   80,},
  {   40,   40,   41,   42,   42,   43,   43,   44,   45,   45,   46,   47,   47,   48,   48,   49,   50,   50,   51,   52,   52,   53,   53,   54,   55,   55,   56,   57,   57,   58,   58,   59,   60,   60,   61,   61,   62,   63,   63,   64,   65,   65,   66,   66,   67,   68,   68,   69,   70,   70,   71,   71,   72,   73,   73,   74,   75,   75,   76,   76,   77,   78,   78,   79,},
  {   39,   40,   40,   41,   42,   42,   43,   44,   44,   45,   45,   46,   47,   47,   48,   48,   49,   50,   50,   51,   52,   52,   53,   53,   54,   55,   55,   56,   56,   57,   58,   58,   59,   59,   60,   61,   61,   62,   63,   63,   64,   64,   65,   66,   66,   67,   67,   68,   69,   69,   70,   71,   71,   72,   72,   73,   74,   74,   75,   75,   76,   77,   77,   78,},
  {   39,   39,   40,   41,   41,   42,   42,   43,   44,   44,   45,   45,   46,   47,   47,   48,   48,   49,   50,   50,   51,   51,   52,   53,   53,   54,   54,   55,   56,   56,   57,   58,   58,   59,   59,   60,   61,   61,   62,   62,   63,   64,   64,   65,   65,   66,   67,   67,   68,   68,   69,   70,   70,   71,   71,   72,   73,   73,   74,   74,   75,   76,   76,   77,},
  {   38,   39,   39,   40,   41,   41,   42,   42,   43,   44,   44,   45,   45,   46,   47,   47,   48,   48,   49,   50,   50,   51,   51,   52,   53,   53,   54,   54,   55,   56,   56,   57,   57,   58,   59,   59,   60,   60,   61,   62,   62,   63,   63,   64,   65,   65,   66,   66,   67,   68,   68,   69,   69,   70,   71,   71,   72,   72,   73,   74,   74,   75,   75,   76,},
  {   38,   38,   39,   40,   40,   41,   41,   42,   42,   43,   44,   44,   45,   45,   46,   47,   47,   48,   48,   49,   50,   50,   51,   51,   52,   52,   53,   54,   54,   55,   55,   56,   57,   57,   58,   58,   59,   60,   60,   61,   61,   62,   63,   63,   64,   64,   65,   65,   66,   67,   67,   68,   68,   69,   70,   70,   71,   71,   72,   73,   73,   74,   74,   75,},
  {   37,   38,   38,   39,   40,   40,   41,   41,   42,   42,   43,   44,   44,   45,   45,   46,   47,   47,   48,   48,   49,   49,   50,   51,   51,   52,   52,   53,   54,   54,   55,   55,   56,   56,   57,   58,   58,   59,   59,   60,   61,   61,   62,   62,   63,   63,   64,   65,   65,   66,   66,   67,   68,   68,   69,   69,   70,   70,   71,   72,   72,   73,   73,   74,},
  {   37,   37,   38,   38,   39,   40,   40,   41,   41,   42,   42,   43,   44,   44,   45,   45,   46,   47,   47,   48,   48,   49,   49,   50,   51,   51,   52,   52,   53,   53,   54,   55,   55,   56,   56,   57,   57,   58,   59,   59,   60,   60,   61,   61,   62,   63,   63,   64,   64,   65,   65,   66,   67,   67,   68,   68,   69,   69,   70,   71,   71,   72,   72,   73,},
  {   36,   37,   37,   38,   39,   39,   40,   40,   41,   41,   42,   42,   43,   44,   44,   45,   45,   46,   46,   47,   48,   48,   49,   49,   50,   50,   51,   52,   52,   53,   53,   54,   54,   55,   56,   56,   57,   57,   58,   58,   59,   59,   60,   61,   61,   62,   62,   63,   63,   64,   65,   65,   66,   66,   67,   67,   68,   69,   69,   70,   70,   71,   71,   72,},
  {   36,   36,   37,   37,   38,   39,   39,   40,   40,   41,   41,   42,   42,   43,   44,   44,   45,   45,   46,   46,   47,   47,   48,   49,   49,   50,   50,   51,   51,   52,   53,   53,   54,   54,   55,   55,   56,   56,   57,   58,   58,   59,   59,   60,   60,   61,   61,   62,   63,   63,   64,   64,   65,   65,   66,   66,   67,   68,   68,   69,   69,   70,   70,   71,},
  {   35,   36,   36,   37,   37,   38,   39,   39,   40,   40,   41,   41,   42,   42,   43,   44,   44,   45,   45,   46,   46,   47,   47,   48,   48,   49,   50,   50,   51,   51,   52,   52,   53,   53,   54,   55,   55,   56,   56,   57,   57,   58,   58,   59,   59,   60,   61,   61,   62,   62,   63,   63,   64,   64,   65,   66,   66,   67,   67,   68,   68,   69,   69,   70,},
  {   35,   35,   36,   36,   37,   37,   38,   39,   39,   40,   40,   41,   41,   42,   42,   43,   43,   44,   45,   45,   46,   46,   47,   47,   48,   48,   49,   49,   50,   50,   51,   52,   52,   53,   53,   54,   54,   55,   55,   56,   56,   57,   58,   58,   59,   59,   60,   60,   61,   61,   62,   62,   63,   64,   64,   65,   65,   66,   66,   67,   67,   68,   68,   69,},
  {   34,   35,   35,   36,   36,   37,   37,   38,   39,   39,   40,   40,   41,   41,   42,   42,   43,   43,   44,   44,   45,   45,   46,   47,   47,   48,   48,   49,   49,   50,   50,   51,   51,   52,   52,   53,   54,   54,   55,   55,   56,   56,   57,   57,   58,   58,   59,   59,   60,   60,   61,   62,   62,   63,   63,   64,   64,   65,   65,   66,   66,   67,   67,   68,},
  {   34,   34,   35,   35,   36,   36,   37,   37,   38,   38,   39,   40,   40,   41,   41,   42,   42,   43,   43,   44,   44,   45,   45,   46,   46,   47,   47,   48,   49,   49,   50,   50,   51,   51,   52,   52,   53,   53,   54,   54,   55,   55,   56,   56,   57,   57,   58,   59,   59,   60,   60,   61,   61,   62,   62,   63,   63,   64,   64,   65,   65,   66,   66,   67,},
  {   33,   34,   34,   35,   35,   36,   36,   37,   37,   38,   38,   39,   39,   40,   41,   41,   42,   42,   43,   43,   44,   44,   45,   45,   46,   46,   47,   47,   48,   48,   49,   49,   50,   50,   51,   51,   52,   52,   53,   54,   54,   55,   55,   56,   56,   57,   57,   58,   58,   59,   59,   60,   60,   61,   61,   62,   62,   63,   63,   64,   64,   65,   65,   66,},
  {   33,   33,   34,   34,   35,   35,   36,   36,   37,   37,   38,   38,   39,   39,   40,   40,   41,   41,   42,   42,   43,   43,   44,   45,   45,   46,   46,   47,   47,   48,   48,   49,   49,   50,   50,   51,   51,   52,   52,   53,   53,   54,   54,   55,   55,   56,   56,   57,   57,   58,   58,   59,   59,   60,   60,   61,   61,   62,   62,   63,   63,   64,   64,   65,},
  {   32,   33,   33,   34,   34,   35,   35,   36,   36,   37,   37,   38,   38,   39,   39,   40,   40,   41,   41,   42,   42,   43,   43,   44,   44,   45,   45,   46,   46,   47,   47,   48,   48,   49,   49,   50,   50,   51,   51,   52,   52,   53,   53,   54,   54,   55,   55,   56,   56,   57,   57,   58,   58,   59,   59,   60,   60,   61,   61,   62,   62,   63,   63,   64,},
  {   32,   32,   33,   33,   34,   34,   35,   35,   36,   36,   37,   37,   38,   38,   39,   39,   40,   40,   41,   41,   42,   42,   43,   43,   44,   44,   45,   45,   46,   46,   47,   47,   48,   48,   49,   49,   50,   50,   51,   51,   52,   52,   53,   53,   54,   54,   55,   55,   56,   56,   57,   57,   58,   58,   59,   59,   60,   60,   61,   61,   62,   62,   63,   63,},
  {   31,   32,   32,   33,   33,   34,   34,   35,   35,   36,   36,   37,   37,   38,   38,   39,   39,   40,   40,   41,   41,   42,   42,   42,   43,   43,   44,   44,   45,   45,   46,   46,   47,   47,   48,   48,   49,   49,   50,   50,   51,   51,   52,   52,   53,   53,   54,   54,   55,   55,   56,   56,   57,   57,   58,   58,   59,   59,   60,   60,   61,   61,   62,   62,},
  {   31,   31,   32,   32,   33,   33,   34,   34,   35,   35,   36,   36,   37,   37,   37,   38,   38,   39,   39,   40,   40,   41,   41,   42,   42,   43,   43,   44,   44,   45,   45,   46,   46,   47,   47,   48,   48,   49,   49,   49,   50,   50,   51,   51,   52,   52,   53,   53,   54,   54,   55,   55,   56,   56,   57,   57,   58,   58,   59,   59,   60,   60,   61,   61,},
  {   30,   31,   31,   32,   32,   33,   33,   34,   34,   35,   35,   35,   36,   36,   37,   37,   38,   38,   39,   39,   40,   40,   41,   41,   42,   42,   43,   43,   43,   44,   44,   45,   45,   46,   46,   47,   47,   48,   48,   49,   49,   50,   50,   51,   51,   52,   52,   52,   53,   53,   54,   54,   55,   55,   56,   56,   57,   57,   58,   58,   59,   59,   60,   60,},
  {   30,   30,   31,   31,   32,   32,   33,   33,   33,   34,   34,   35,   35,   36,   36,   37,   37,   38,   38,   39,   39,   40,   40,   40,   41,   41,   42,   42,   43,   43,   44,   44,   45,   45,   46,   46,   46,   47,   47,   48,   48,   49,   49,   50,   50,   51,   51,   52,   52,   53,   53,   53,   54,   54,   55,   55,   56,   56,   57,   57,   58,   58,   59,   59,},
  {   29,   30,   30,   31,   31,   32,   32,   32,   33,   33,   34,   34,   35,   35,   36,   36,   37,   37,   37,   38,   38,   39,   39,   40,   40,   41,   41,   42,   42,   43,   43,   43,   44,   44,   45,   45,   46,   46,   47,   47,   48,   48,   48,   49,   49,   50,   50,   51,   51,   52,   52,   53,   53,   53,   54,   54,   55,   55,   56,   56,   57,   57,   58,   58,},
  {   29,   29,   30,   30,   31,   31,   31,   32,   32,   33,   33,   34,   34,   35,   35,   35,   36,   36,   37,   37,   38,   38,   39,   39,   40,   40,   40,   41,   41,   42,   42,   43,   43,   44,   44,   44,   45,   45,   46,   46,   47,   47,   48,   48,   49,   49,   49,   50,   50,   51,   51,   52,   52,   53,   53,   53,   54,   54,   55,   55,   56,   56,   57,   57,},
  {   28,   29,   29,   30,   30,   30,   31,   31,   32,   32,   33,   33,   34,   34,   34,   35,   35,   36,   36,   37,   37,   38,   38,   38,   39,   39,   40,   40,   41,   41,   41,   42,   42,   43,   43,   44,   44,   45,   45,   45,   46,   46,   47,   47,   48,   48,   49,   49,   49,   50,   50,   51,   51,   52,   52,   53,   53,   53,   54,   54,   55,   55,   56,   56,},
  {   28,   28,   29,   29,   29,   30,   30,   31,   31,   32,   32,   33,   33,   33,   34,   34,   35,   35,   36,   36,   36,   37,   37,   38,   38,   39,   39,   39,   40,   40,   41,   41,   42,   42,   42,   43,   43,   44,   44,   45,   45,   46,   46,   46,   47,   47,   48,   48,   49,   49,   49,   50,   50,   51,   51,   52,   52,   52,   53,   53,   54,   54,   55,   55,},
  {   27,   28,   28,   29,   29,   29,   30,   30,   31,   31,   32,   32,   32,   33,   33,   34,   34,   34,   35,   35,   36,   36,   37,   37,   37,   38,   38,   39,   39,   40,   40,   40,   41,   41,   42,   42,   43,   43,   43,   44,   44,   45,   45,   46,   46,   46,   47,   47,   48,   48,   49,   49,   49,   50,   50,   51,   51,   52,   52,   52,   53,   53,   54,   54,},
  {   27,   27,   28,   28,   28,   29,   29,   30,   30,   31,   31,   31,   32,   32,   33,   33,   33,   34,   34,   35,   35,   36,   36,   36,   37,   37,   38,   38,   38,   39,   39,   40,   40,   41,   41,   41,   42,   42,   43,   43,   43,   44,   44,   45,   45,   46,   46,   46,   47,   47,   48,   48,   48,   49,   49,   50,   50,   51,   51,   51,   52,   52,   53,   53,},
  {   26,   27,   27,   27,   28,   28,   29,   29,   30,   30,   30,   31,   31,   32,   32,   32,   33,   33,   34,   34,   34,   35,   35,   36,   36,   37,   37,   37,   38,   38,   39,   39,   39,   40,   40,   41,   41,   41,   42,   42,   43,   43,   43,   44,   44,   45,   45,   46,   46,   46,   47,   47,   48,   48,   48,   49,   49,   50,   50,   50,   51,   51,   52,   52,},
  {   26,   26,   27,   27,   27,   28,   28,   29,   29,   29,   30,   30,   31,   31,   31,   32,   32,   33,   33,   33,   34,   34,   35,   35,   35,   36,   36,   37,   37,   37,   38,   38,   39,   39,   39,   40,   40,   41,   41,   41,   42,   42,   43,   43,   43,   44,   44,   45,   45,   45,   46,   46,   47,   47,   47,   48,   48,   49,   49,   49,   50,   50,   51,   51,},
  {   25,   26,   26,   26,   27,   27,   28,   28,   28,   29,   29,   30,   30,   30,   31,   31,   32,   32,   32,   33,   33,   34,   34,   34,   35,   35,   36,   36,   36,   37,   37,   37,   38,   38,   39,   39,   39,   40,   40,   41,   41,   41,   42,   42,   43,   43,   43,   44,   44,   45,   45,   45,   46,   46,   47,   47,   47,   48,   48,   49,   49,   49,   50,   50,},
  {   25,   25,   26,   26,   26,   27,   27,   27,   28,   28,   29,   29,   29,   30,   30,   31,   31,   31,   32,   32,   32,   33,   33,   34,   34,   34,   35,   35,   36,   36,   36,   37,   37,   38,   38,   38,   39,   39,   39,   40,   40,   41,   41,   41,   42,   42,   43,   43,   43,   44,   44,   44,   45,   45,   46,   46,   46,   47,   47,   48,   48,   48,   49,   49,},
  {   24,   25,   25,   25,   26,   26,   27,   27,   27,   28,   28,   28,   29,   29,   30,   30,   30,   31,   31,   31,   32,   32,   33,   33,   33,   34,   34,   34,   35,   35,   36,   36,   36,   37,   37,   38,   38,   38,   39,   39,   39,   40,   40,   41,   41,   41,   42,   42,   42,   43,   43,   44,   44,   44,   45,   45,   45,   46,   46,   47,   47,   47,   48,   48,},
  {   24,   24,   24,   25,   25,   26,   26,   26,   27,   27,   27,   28,   28,   29,   29,   29,   30,   30,   30,   31,   31,   32,   32,   32,   33,   33,   33,   34,   34,   35,   35,   35,   36,   36,   36,   37,   37,   37,   38,   38,   39,   39,   39,   40,   40,   40,   41,   41,   42,   42,   42,   43,   43,   43,   44,   44,   45,   45,   45,   46,   46,   46,   47,   47,},
  {   23,   24,   24,   24,   25,   25,   25,   26,   26,   27,   27,   27,   28,   28,   28,   29,   29,   29,   30,   30,   31,   31,   31,   32,   32,   32,   33,   33,   33,   34,   34,   35,   35,   35,   36,   36,   36,   37,   37,   37,   38,   38,   39,   39,   39,   40,   40,   40,   41,   41,   41,   42,   42,   43,   43,   43,   44,   44,   44,   45,   45,   45,   46,   46,},
  {   23,   23,   23,   24,   24,   25,   25,   25,   26,   26,   26,   27,   27,   27,   28,   28,   28,   29,   29,   30,   30,   30,   31,   31,   31,   32,   32,   32,   33,   33,   33,   34,   34,   34,   35,   35,   36,   36,   36,   37,   37,   37,   38,   38,   38,   39,   39,   39,   40,   40,   41,   41,   41,   42,   42,   42,   43,   43,   43,   44,   44,   44,   45,   45,},
  {   22,   23,   23,   23,   24,   24,   24,   25,   25,   25,   26,   26,   26,   27,   27,   27,   28,   28,   29,   29,   29,   30,   30,   30,   31,   31,   31,   32,   32,   32,   33,   33,   33,   34,   34,   34,   35,   35,   35,   36,   36,   37,   37,   37,   38,   38,   38,   39,   39,   39,   40,   40,   40,   41,   41,   41,   42,   42,   42,   43,   43,   43,   44,   44,},
  {   22,   22,   22,   23,   23,   23,   24,   24,   24,   25,   25,   25,   26,   26,   27,   27,   27,   28,   28,   28,   29,   29,   29,   30,   30,   30,   31,   31,   31,   32,   32,   32,   33,   33,   33,   34,   34,   34,   35,   35,   35,   36,   36,   36,   37,   37,   37,   38,   38,   38,   39,   39,   39,   40,   40,   40,   41,   41,   41,   42,   42,   42,   43,   43,},
  {   21,   22,   22,   22,   23,   23,   23,   24,   24,   24,   25,   25,   25,   26,   26,   26,   27,   27,   27,   28,   28,   28,   29,   29,   29,   30,   30,   30,   31,   31,   31,   32,   32,   32,   33,   33,   33,   34,   34,   34,   35,   35,   35,   36,   36,   36,   37,   37,   37,   38,   38,   38,   39,   39,   39,   40,   40,   40,   41,   41,   41,   42,   42,   42,},
  {   21,   21,   21,   22,   22,   22,   23,   23,   23,   24,   24,   24,   25,   25,   25,   26,   26,   26,   27,   27,   27,   28,   28,   28,   29,   29,   29,   30,   30,   30,   30,   31,   31,   31,   32,   32,   32,   33,   33,   33,   34,   34,   34,   35,   35,   35,   36,   36,   36,   37,   37,   37,   38,   38,   38,   39,   39,   39,   40,   40,   40,   41,   41,   41,},
  {   20,   21,   21,   21,   22,   22,   22,   22,   23,   23,   23,   24,   24,   24,   25,   25,   25,   26,   26,   26,   27,   27,   27,   28,   28,   28,   28,   29,   29,   29,   30,   30,   30,   31,   31,   31,   32,   32,   32,   33,   33,   33,   34,   34,   34,   34,   35,   35,   35,   36,   36,   36,   37,   37,   37,   38,   38,   38,   39,   39,   39,   40,   40,   40,},
  {   20,   20,   20,   21,   21,   21,   22,   22,   22,   23,   23,   23,   23,   24,   24,   24,   25,   25,   25,   26,   26,   26,   27,   27,   27,   27,   28,   28,   28,   29,   29,   29,   30,   30,   30,   31,   31,   31,   31,   32,   32,   32,   33,   33,   33,   34,   34,   34,   35,   35,   35,   35,   36,   36,   36,   37,   37,   37,   38,   38,   38,   39,   39,   39,},
  {   19,   20,   20,   20,   20,   21,   21,   21,   22,   22,   22,   23,   23,   23,   23,   24,   24,   24,   25,   25,   25,   26,   26,   26,   26,   27,   27,   27,   28,   28,   28,   29,   29,   29,   29,   30,   30,   30,   31,   31,   31,   32,   32,   32,   32,   33,   33,   33,   34,   34,   34,   35,   35,   35,   35,   36,   36,   36,   37,   37,   37,   38,   38,   38,},
  {   19,   19,   19,   20,   20,   20,   21,   21,   21,   21,   22,   22,   22,   23,   23,   23,   23,   24,   24,   24,   25,   25,   25,   25,   26,   26,   26,   27,   27,   27,   28,   28,   28,   28,   29,   29,   29,   30,   30,   30,   30,   31,   31,   31,   32,   32,   32,   33,   33,   33,   33,   34,   34,   34,   35,   35,   35,   35,   36,   36,   36,   37,   37,   37,},
  {   18,   19,   19,   19,   19,   20,   20,   20,   21,   21,   21,   21,   22,   22,   22,   23,   23,   23,   23,   24,   24,   24,   25,   25,   25,   25,   26,   26,   26,   27,   27,   27,   27,   28,   28,   28,   29,   29,   29,   29,   30,   30,   30,   31,   31,   31,   31,   32,   32,   32,   33,   33,   33,   33,   34,   34,   34,   35,   35,   35,   35,   36,   36,   36,},
  {   18,   18,   18,   19,   19,   19,   19,   20,   20,   20,   21,   21,   21,   21,   22,   22,   22,   22,   23,   23,   23,   24,   24,   24,   24,   25,   25,   25,   26,   26,   26,   26,   27,   27,   27,   27,   28,   28,   28,   29,   29,   29,   29,   30,   30,   30,   31,   31,   31,   31,   32,   32,   32,   32,   33,   33,   33,   34,   34,   34,   34,   35,   35,   35,},
  {   17,   18,   18,   18,   18,   19,   19,   19,   19,   20,   20,   20,   20,   21,   21,   21,   22,   22,   22,   22,   23,   23,   23,   23,   24,   24,   24,   25,   25,   25,   25,   26,   26,   26,   26,   27,   27,   27,   27,   28,   28,   28,   29,   29,   29,   29,   30,   30,   30,   30,   31,   31,   31,   32,   32,   32,   32,   33,   33,   33,   33,   34,   34,   34,},
  {   17,   17,   17,   18,   18,   18,   18,   19,   19,   19,   19,   20,   20,   20,   20,   21,   21,   21,   21,   22,   22,   22,   23,   23,   23,   23,   24,   24,   24,   24,   25,   25,   25,   25,   26,   26,   26,   26,   27,   27,   27,   27,   28,   28,   28,   29,   29,   29,   29,   30,   30,   30,   30,   31,   31,   31,   31,   32,   32,   32,   32,   33,   33,   33,},
  {   16,   17,   17,   17,   17,   18,   18,   18,   18,   19,   19,   19,   19,   20,   20,   20,   20,   21,   21,   21,   21,   22,   22,   22,   22,   23,   23,   23,   23,   24,   24,   24,   24,   25,   25,   25,   25,   26,   26,   26,   26,   27,   27,   27,   27,   28,   28,   28,   28,   29,   29,   29,   29,   30,   30,   30,   30,   31,   31,   31,   31,   32,   32,   32,},
  {   16,   16,   16,   16,   17,   17,   17,   17,   18,   18,   18,   18,   19,   19,   19,   19,   20,   20,   20,   20,   21,   21,   21,   21,   22,   22,   22,   22,   23,   23,   23,   23,   24,   24,   24,   24,   25,   25,   25,   25,   26,   26,   26,   26,   27,   27,   27,   27,   28,   28,   28,   28,   29,   29,   29,   29,   30,   30,   30,   30,   31,   31,   31,   31,},
  {   15,   15,   16,   16,   16,   16,   17,   17,   17,   17,   18,   18,   18,   18,   19,   19,   19,   19,   20,   20,   20,   20,   20,   21,   21,   21,   21,   22,   22,   22,   22,   23,   23,   23,   23,   24,   24,   24,   24,   25,   25,   25,   25,   25,   26,   26,   26,   26,   27,   27,   27,   27,   28,   28,   28,   28,   29,   29,   29,   29,   30,   30,   30,   30,},
  {   15,   15,   15,   15,   16,   16,   16,   16,   17,   17,   17,   17,   18,   18,   18,   18,   18,   19,   19,   19,   19,   20,   20,   20,   20,   21,   21,   21,   21,   21,   22,   22,   22,   22,   23,   23,   23,   23,   24,   24,   24,   24,   24,   25,   25,   25,   25,   26,   26,   26,   26,   27,   27,   27,   27,   27,   28,   28,   28,   28,   29,   29,   29,   29,},
  {   14,   14,   15,   15,   15,   15,   16,   16,   16,   16,   16,   17,   17,   17,   17,   18,   18,   18,   18,   18,   19,   19,   19,   19,   20,   20,   20,   20,   20,   21,   21,   21,   21,   22,   22,   22,   22,   22,   23,   23,   23,   23,   24,   24,   24,   24,   24,   25,   25,   25,   25,   26,   26,   26,   26,   26,   27,   27,   27,   27,   28,   28,   28,   28,},
  {   14,   14,   14,   14,   15,   15,   15,   15,   15,   16,   16,   16,   16,   17,   17,   17,   17,   17,   18,   18,   18,   18,   18,   19,   19,   19,   19,   20,   20,   20,   20,   20,   21,   21,   21,   21,   21,   22,   22,   22,   22,   23,   23,   23,   23,   23,   24,   24,   24,   24,   24,   25,   25,   25,   25,   26,   26,   26,   26,   26,   27,   27,   27,   27,},
  {   13,   13,   14,   14,   14,   14,   14,   15,   15,   15,   15,   16,   16,   16,   16,   16,   17,   17,   17,   17,   17,   18,   18,   18,   18,   18,   19,   19,   19,   19,   19,   20,   20,   20,   20,   20,   21,   21,   21,   21,   22,   22,   22,   22,   22,   23,   23,   23,   23,   23,   24,   24,   24,   24,   24,   25,   25,   25,   25,   25,   26,   26,   26,   26,},
  {   13,   13,   13,   13,   14,   14,   14,   14,   14,   15,   15,   15,   15,   15,   16,   16,   16,   16,   16,   17,   17,   17,   17,   17,   18,   18,   18,   18,   18,   19,   19,   19,   19,   19,   20,   20,   20,   20,   20,   21,   21,   21,   21,   21,   22,   22,   22,   22,   22,   23,   23,   23,   23,   23,   24,   24,   24,   24,   24,   25,   25,   25,   25,   25,},
  {   12,   12,   13,   13,   13,   13,   13,   14,   14,   14,   14,   14,   15,   15,   15,   15,   15,   16,   16,   16,   16,   16,   16,   17,   17,   17,   17,   17,   18,   18,   18,   18,   18,   19,   19,   19,   19,   19,   20,   20,   20,   20,   20,   20,   21,   21,   21,   21,   21,   22,   22,   22,   22,   22,   23,   23,   23,   23,   23,   24,   24,   24,   24,   24,},
  {   12,   12,   12,   12,   12,   13,   13,   13,   13,   13,   14,   14,   14,   14,   14,   15,   15,   15,   15,   15,   15,   16,   16,   16,   16,   16,   17,   17,   17,   17,   17,   17,   18,   18,   18,   18,   18,   19,   19,   19,   19,   19,   19,   20,   20,   20,   20,   20,   21,   21,   21,   21,   21,   21,   22,   22,   22,   22,   22,   23,   23,   23,   23,   23,},
  {   11,   11,   12,   12,   12,   12,   12,   12,   13,   13,   13,   13,   13,   14,   14,   14,   14,   14,   14,   15,   15,   15,   15,   15,   15,   16,   16,   16,   16,   16,   17,   17,   17,   17,   17,   17,   18,   18,   18,   18,   18,   18,   19,   19,   19,   19,   19,   20,   20,   20,   20,   20,   20,   21,   21,   21,   21,   21,   21,   22,   22,   22,   22,   22,},
  {   11,   11,   11,   11,   11,   12,   12,   12,   12,   12,   12,   13,   13,   13,   13,   13,   13,   14,   14,   14,   14,   14,   14,   15,   15,   15,   15,   15,   15,   16,   16,   16,   16,   16,   16,   17,   17,   17,   17,   17,   17,   18,   18,   18,   18,   18,   18,   19,   19,   19,   19,   19,   19,   20,   20,   20,   20,   20,   20,   21,   21,   21,   21,   21,},
  {   10,   10,   11,   11,   11,   11,   11,   11,   12,   12,   12,   12,   12,   12,   12,   13,   13,   13,   13,   13,   13,   14,   14,   14,   14,   14,   14,   15,   15,   15,   15,   15,   15,   16,   16,   16,   16,   16,   16,   16,   17,   17,   17,   17,   17,   17,   18,   18,   18,   18,   18,   18,   19,   19,   19,   19,   19,   19,   20,   20,   20,   20,   20,   20,},
  {   10,   10,   10,   10,   10,   11,   11,   11,   11,   11,   11,   11,   12,   12,   12,   12,   12,   12,   12,   13,   13,   13,   13,   13,   13,   14,   14,   14,   14,   14,   14,   14,   15,   15,   15,   15,   15,   15,   16,   16,   16,   16,   16,   16,   16,   17,   17,   17,   17,   17,   17,   18,   18,   18,   18,   18,   18,   18,   19,   19,   19,   19,   19,   19,},
  {    9,    9,   10,   10,   10,   10,   10,   10,   10,   11,   11,   11,   11,   11,   11,   11,   12,   12,   12,   12,   12,   12,   12,   13,   13,   13,   13,   13,   13,   13,   14,   14,   14,   14,   14,   14,   14,   15,   15,   15,   15,   15,   15,   15,   16,   16,   16,   16,   16,   16,   16,   17,   17,   17,   17,   17,   17,   17,   18,   18,   18,   18,   18,   18,},
  {    9,    9,    9,    9,    9,    9,   10,   10,   10,   10,   10,   10,   10,   11,   11,   11,   11,   11,   11,   11,   11,   12,   12,   12,   12,   12,   12,   12,   13,   13,   13,   13,   13,   13,   13,   14,   14,   14,   14,   14,   14,   14,   14,   15,   15,   15,   15,   15,   15,   15,   16,   16,   16,   16,   16,   16,   16,   17,   17,   17,   17,   17,   17,   17,},
  {    8,    8,    9,    9,    9,    9,    9,    9,    9,    9,   10,   10,   10,   10,   10,   10,   10,   10,   11,   11,   11,   11,   11,   11,   11,   11,   12,   12,   12,   12,   12,   12,   12,   13,   13,   13,   13,   13,   13,   13,   13,   14,   14,   14,   14,   14,   14,   14,   14,   15,   15,   15,   15,   15,   15,   15,   15,   16,   16,   16,   16,   16,   16,   16,},
  {    8,    8,    8,    8,    8,    8,    8,    9,    9,    9,    9,    9,    9,    9,    9,   10,   10,   10,   10,   10,   10,   10,   10,   11,   11,   11,   11,   11,   11,   11,   11,   12,   12,   12,   12,   12,   12,   12,   12,   12,   13,   13,   13,   13,   13,   13,   13,   13,   14,   14,   14,   14,   14,   14,   14,   14,   15,   15,   15,   15,   15,   15,   15,   15,},
  {    7,    7,    7,    8,    8,    8,    8,    8,    8,    8,    8,    8,    9,    9,    9,    9,    9,    9,    9,    9,   10,   10,   10,   10,   10,   10,   10,   10,   10,   11,   11,   11,   11,   11,   11,   11,   11,   11,   12,   12,   12,   12,   12,   12,   12,   12,   12,   13,   13,   13,   13,   13,   13,   13,   13,   13,   14,   14,   14,   14,   14,   14,   14,   14,},
  {    7,    7,    7,    7,    7,    7,    7,    7,    8,    8,    8,    8,    8,    8,    8,    8,    8,    9,    9,    9,    9,    9,    9,    9,    9,    9,    9,   10,   10,   10,   10,   10,   10,   10,   10,   10,   11,   11,   11,   11,   11,   11,   11,   11,   11,   11,   12,   12,   12,   12,   12,   12,   12,   12,   12,   13,   13,   13,   13,   13,   13,   13,   13,   13,},
  {    6,    6,    6,    7,    7,    7,    7,    7,    7,    7,    7,    7,    7,    8,    8,    8,    8,    8,    8,    8,    8,    8,    8,    8,    9,    9,    9,    9,    9,    9,    9,    9,    9,    9,   10,   10,   10,   10,   10,   10,   10,   10,   10,   10,   11,   11,   11,   11,   11,   11,   11,   11,   11,   11,   12,   12,   12,   12,   12,   12,   12,   12,   12,   12,},
  {    6,    6,    6,    6,    6,    6,    6,    6,    6,    7,    7,    7,    7,    7,    7,    7,    7,    7,    7,    7,    8,    8,    8,    8,    8,    8,    8,    8,    8,    8,    8,    9,    9,    9,    9,    9,    9,    9,    9,    9,    9,    9,   10,   10,   10,   10,   10,   10,   10,   10,   10,   10,   10,   11,   11,   11,   11,   11,   11,   11,   11,   11,   11,   11,},
  {    5,    5,    5,    5,    6,    6,    6,    6,    6,    6,    6,    6,    6,    6,    6,    6,    7,    7,    7,    7,    7,    7,    7,    7,    7,    7,    7,    7,    8,    8,    8,    8,    8,    8,    8,    8,    8,    8,    8,    8,    9,    9,    9,    9,    9,    9,    9,    9,    9,    9,    9,    9,   10,   10,   10,   10,   10,   10,   10,   10,   10,   10,   10,   10,},
  {    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    6,    6,    6,    6,    6,    6,    6,    6,    6,    6,    6,    6,    6,    7,    7,    7,    7,    7,    7,    7,    7,    7,    7,    7,    7,    7,    7,    8,    8,    8,    8,    8,    8,    8,    8,    8,    8,    8,    8,    8,    9,    9,    9,    9,    9,    9,    9,    9,    9,    9,    9,    9,    9,},
  {    4,    4,    4,    4,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    6,    6,    6,    6,    6,    6,    6,    6,    6,    6,    6,    6,    6,    6,    6,    7,    7,    7,    7,    7,    7,    7,    7,    7,    7,    7,    7,    7,    7,    7,    8,    8,    8,    8,    8,    8,    8,    8,    8,    8,    8,    8,    8,    8,    8,},
  {    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    6,    6,    6,    6,    6,    6,    6,    6,    6,    6,    6,    6,    6,    6,    6,    6,    6,    7,    7,    7,    7,    7,    7,    7,    7,    7,    7,    7,    7,    7,    7,    7,    7,    7,},
  {    3,    3,    3,    3,    3,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    6,    6,    6,    6,    6,    6,    6,    6,    6,    6,    6,    6,    6,    6,    6,    6,    6,    6,    6,},
  {    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,},
  {    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,},
  {    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,},
  {    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,},
  {    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,},
  {    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,},
};

#endif
const uint8_t ProbModelTables::m_RenormTable_32[32] =
{
  6,  5,  4,  4,
  3,  3,  3,  3,
  2,  2,  2,  2,
  2,  2,  2,  2,
  1,  1,  1,  1,
  1,  1,  1,  1,
  1,  1,  1,  1,
  1,  1,  1,  1
};

#if JEM_TOOLS
const uint8_t ProbModelTables::m_RenormTable_128[128] =
{
  7,  6,  5,  5,
  4,  4,  4,  4,
  3,  3,  3,  3,
  3,  3,  3,  3,
  2,  2,  2,  2,
  2,  2,  2,  2,
  2,  2,  2,  2,
  2,  2,  2,  2,
  1,  1,  1,  1,
  1,  1,  1,  1,
  1,  1,  1,  1,
  1,  1,  1,  1,
  1,  1,  1,  1,
  1,  1,  1,  1,
  1,  1,  1,  1,
  1,  1,  1,  1,
  0,  0,  0,  0,
  0,  0,  0,  0,
  0,  0,  0,  0,
  0,  0,  0,  0,
  0,  0,  0,  0,
  0,  0,  0,  0,
  0,  0,  0,  0,
  0,  0,  0,  0,
  0,  0,  0,  0,
  0,  0,  0,  0,
  0,  0,  0,  0,
  0,  0,  0,  0,
  0,  0,  0,  0,
  0,  0,  0,  0,
  0,  0,  0,  0,
  0,  0,  0,  0,
};

#endif
#if JEM_TOOLS
const uint16_t ProbModelTables::m_InistateToCount[128] =
{
    614,    647,    681,    718,    756,    797,    839,    884,    932,    982,   1034,   1089,   1148,   1209,   1274,   1342,
   1414,   1490,   1569,   1653,   1742,   1835,   1933,   2037,   2146,   2261,   2382,   2509,   2643,   2785,   2934,   3091,
   3256,   3430,   3614,   3807,   4011,   4225,   4452,   4690,   4941,   5205,   5483,   5777,   6086,   6412,   6755,   7116,
   7497,   7898,   8320,   8766,   9235,   9729,  10249,  10798,  11375,  11984,  12625,  13300,  14012,  14762,  15551,  16384,
  16384,  17216,  18005,  18755,  19467,  20142,  20783,  21392,  21969,  22518,  23038,  23532,  24001,  24447,  24869,  25270,
  25651,  26012,  26355,  26681,  26990,  27284,  27562,  27826,  28077,  28315,  28542,  28756,  28960,  29153,  29337,  29511,
  29676,  29833,  29982,  30124,  30258,  30385,  30506,  30621,  30730,  30834,  30932,  31025,  31114,  31198,  31277,  31353,
  31425,  31493,  31558,  31619,  31678,  31733,  31785,  31835,  31883,  31928,  31970,  32011,  32049,  32086,  32120,  32153
};

#endif
#if HM_REPRODUCE_CONTEXT_IDX_CALCULATION
const double ProbModelTables::m_StateToProbLPS[64] =
{
  0.50000000, 0.47460857, 0.45050660, 0.42762859, 0.40591239, 0.38529900, 0.36573242, 0.34715948,
  0.32952974, 0.31279528, 0.29691064, 0.28183267, 0.26752040, 0.25393496, 0.24103941, 0.22879875,
  0.21717969, 0.20615069, 0.19568177, 0.18574449, 0.17631186, 0.16735824, 0.15885931, 0.15079198,
  0.14313433, 0.13586556, 0.12896592, 0.12241667, 0.11620000, 0.11029903, 0.10469773, 0.09938088,
  0.09433404, 0.08954349, 0.08499621, 0.08067986, 0.07658271, 0.07269362, 0.06900203, 0.06549791,
  0.06217174, 0.05901448, 0.05601756, 0.05317283, 0.05047256, 0.04790942, 0.04547644, 0.04316702,
  0.04097487, 0.03889405, 0.03691890, 0.03504406, 0.03326442, 0.03157516, 0.02997168, 0.02844963,
  0.02700488, 0.02563349, 0.02433175, 0.02309612, 0.02192323, 0.02080991, 0.01975312, 0.01875000
};
#endif




void BinProbModel_Std::init( int qp, int initId )
{
  int slope     = ( ( initId >>  4 )  * 5 ) - 45;
  int offset    = ( ( initId  & 15 ) << 3 ) - 16;
  int inistate  = ( ( slope   * qp ) >> 4 ) + offset;
  if( inistate >= 64 )
  {
    m_State     = ( std::min( 62, inistate - 64 ) << 1 ) + 1;
  }
  else
  {
    m_State     = ( std::min( 62, 63 - inistate ) << 1 );
  }
}
#if JEM_TOOLS
void BinProbModel_JMP::init( int qp, int initId )
{
  int slope     = ( ( initId >>  4 )  * 5 ) - 45;
  int offset    = ( ( initId  & 15 ) << 3 ) - 16;
  int inistate  = ( ( slope   * qp ) >> 4 ) + offset;
  m_P0 = m_P1   = m_InistateToCount[ inistate < 1 ? 1 : inistate > 126 ? 126 : inistate ];
}

void BinProbModel_JAW::init( int qp, int initId )
{
  int slope         = ( ( initId >>  4 )  * 5 ) - 45;
  int offset        = ( ( initId  & 15 ) << 3 ) - 16;
  int inistate      = ( ( slope   * qp ) >> 4 ) + offset;
  m_P0              = m_InistateToCount[ inistate < 1 ? 1 : inistate > 126 ? 126 : inistate ];
  m_Log2WindowSize0 = m_DefaultLog2WindowSize0;
}

void BinProbModel_JMPAW::init( int qp, int initId )
{
  int slope         = ( ( initId >>  4 )  * 5 ) - 45;
  int offset        = ( ( initId  & 15 ) << 3 ) - 16;
  int inistate      = ( ( slope   * qp ) >> 4 ) + offset;
  m_P0 = m_P1       = m_InistateToCount[ inistate < 1 ? 1 : inistate > 126 ? 126 : inistate ];
  m_Log2WindowSize0 = m_DefaultLog2WindowSize0;
}
#endif




CtxSet::CtxSet( std::initializer_list<CtxSet> ctxSets )
{
  uint16_t  minOffset = std::numeric_limits<uint16_t>::max();
  uint16_t  maxOffset = 0;
  for( auto iter = ctxSets.begin(); iter != ctxSets.end(); iter++ )
  {
    minOffset = std::min<uint16_t>( minOffset, (*iter).Offset              );
    maxOffset = std::max<uint16_t>( maxOffset, (*iter).Offset+(*iter).Size );
  }
  Offset  = minOffset;
  Size    = maxOffset - minOffset;
}





const std::vector<uint8_t>& ContextSetCfg::getInitTable( unsigned initId )
{
  CHECK( initId >= (unsigned)sm_InitTables.size(),
         "Invalid initId (" << initId << "), only " << sm_InitTables.size() << " tables defined." );
  return sm_InitTables[initId];
}


CtxSet ContextSetCfg::addCtxSet( std::initializer_list<std::initializer_list<uint8_t>> initSet2d )
{
  const std::size_t startIdx  = sm_InitTables[0].size();
  const std::size_t numValues = ( *initSet2d.begin() ).size();
        std::size_t setId     = 0;
  for( auto setIter = initSet2d.begin(); setIter != initSet2d.end() && setId < sm_InitTables.size(); setIter++, setId++ )
  {
    const std::initializer_list<uint8_t>& initSet   = *setIter;
    std::vector<uint8_t>&           initTable = sm_InitTables[setId];
    CHECK( initSet.size() != numValues,
           "Number of init values do not match for all sets (" << initSet.size() << " != " << numValues << ")." );
    initTable.resize( startIdx + numValues );
    std::size_t elemId = startIdx;
    for( auto elemIter = ( *setIter ).begin(); elemIter != ( *setIter ).end(); elemIter++, elemId++ )
    {
      initTable[elemId] = *elemIter;
    }
  }
  return CtxSet( (uint16_t)startIdx, (uint16_t)numValues );
}



#define CNU 154 // dummy initialization value for unused context models 'Context model Not Used'
std::vector<std::vector<uint8_t>> ContextSetCfg::sm_InitTables( NUMBER_OF_SLICE_TYPES );

const CtxSet ContextSetCfg::SplitFlag = ContextSetCfg::addCtxSet
({
  {  107, 139, 126, 255,   0,},
  {  107, 139, 126, 255,   0,},
  {  139, 141, 157, 255,   0,},
});

const CtxSet ContextSetCfg::BTSplitFlag = ContextSetCfg::addCtxSet
({
  {  107, 139, 126, 154, 154, 154, 154, 154, 154, 154, 154, 154,},
  {  107, 139, 126, 154, 154, 154, 154, 154, 154, 154, 154, 154,},
  {  139, 141, 157, 154, 154, 154, 154, 154, 154, 154, 154, 154,},
});

const CtxSet ContextSetCfg::SkipFlag = ContextSetCfg::addCtxSet
({
  {  197, 185, 201,},
  {  197, 185, 201,},
  {  CNU, CNU, CNU,},
});

const CtxSet ContextSetCfg::MergeFlag = ContextSetCfg::addCtxSet
({
  {  154,},
  {  110,},
  {  CNU,},
});

const CtxSet ContextSetCfg::MergeIdx = ContextSetCfg::addCtxSet
({
  {  137, CNU, CNU, CNU, CNU,},
  {  122, CNU, CNU, CNU, CNU,},
  {  CNU, CNU, CNU, CNU, CNU,},
});

const CtxSet ContextSetCfg::PartSize = ContextSetCfg::addCtxSet
({
  {  154, 139, 154, 154,},
  {  154, 139, 154, 154,},
  {  184, CNU, CNU, CNU,},
});

const CtxSet ContextSetCfg::PredMode = ContextSetCfg::addCtxSet
({
  {  134,},
  {  149,},
  {  CNU,},
});

const CtxSet ContextSetCfg::IPredMode[] =
{
  ContextSetCfg::addCtxSet
  ({
#if JVET_B0051_NON_MPM_MODE
    {  183, CNU, CNU, CNU, 184 },
    {  154, CNU, CNU, CNU, 184 },
    {  184, CNU, CNU, CNU, 184 },
#else
    {  183, CNU, CNU, CNU,  },
    {  154, CNU, CNU, CNU,  },
    {  184, CNU, CNU, CNU,  },
#endif
  }),
  ContextSetCfg::addCtxSet
  ({
    {  139, 152, 139, 154, 154, 154, 154, 154, 154, 154, 154, 154,  },
    {  139, 152, 139, 154, 154, 154, 154, 154, 154, 154, 154, 154,  },
    {  139,  63, 139, 154, 154, 154, 154, 154, 154, 154, 154, 154,  },
  }),
};

const CtxSet ContextSetCfg::PdpcFlag = ContextSetCfg::addCtxSet
({
  {  107,},
  {  107,},
  {  139,},
});

const CtxSet ContextSetCfg::DeltaQP = ContextSetCfg::addCtxSet
({
  {  154, 154, 154,},
  {  154, 154, 154,},
  {  154, 154, 154,},
});

const CtxSet ContextSetCfg::InterDir = ContextSetCfg::addCtxSet
({
  {   95,  79,  63,  31,  31,},
  {   95,  79,  63,  31,  31,},
  {  CNU, CNU, CNU, CNU, CNU,},
});

const CtxSet ContextSetCfg::RefPic = ContextSetCfg::addCtxSet
({
  {  153, 153,},
  {  153, 153,},
  {  CNU, CNU,},
});

#if JEM_TOOLS
const CtxSet ContextSetCfg::AffineFlag = ContextSetCfg::addCtxSet
({
  {  197, 185, 201,},
  {  197, 185, 201,},
  {  CNU, CNU, CNU,},
});
#endif

const CtxSet ContextSetCfg::Mvd = ContextSetCfg::addCtxSet
({
  {  169, 198,},
  {  140, 198,},
  {  CNU, CNU,},
});

const CtxSet ContextSetCfg::TransSubdivFlag = ContextSetCfg::addCtxSet
({
  {  224, 167, 122, 122, 122},
  {  124, 138,  94,  94,  94},
  {  153, 138, 138, 138, 138},
});

const CtxSet ContextSetCfg::QtRootCbf = ContextSetCfg::addCtxSet
({
  {   79,},
  {   79,},
  {  CNU,},
});

const CtxSet ContextSetCfg::QtCbf[] =
{
  ContextSetCfg::addCtxSet
  ({
    {  153, 111,  },
    {  153, 111,  },
    {  111, 141,  },
  }),
  ContextSetCfg::addCtxSet
  ({
    {  149,  92, 167, 154, 154,  },
    {  149, 107, 167, 154, 154,  },
    {   94, 138, 182, 154, 154,  },
  }),
};

const CtxSet ContextSetCfg::SigCoeffGroup[] =
{
  ContextSetCfg::addCtxSet
  ({
    {  121, 140,  },
    {  121, 140,  },
    {   91, 171,  },
  }),
  ContextSetCfg::addCtxSet
  ({
    {   61, 154,  },
    {   61, 154,  },
    {  134, 141,  },
  }),
  ContextSetCfg::addCtxSet
  ({
    {  122, 143,  },
    {   78, 111,  },
    {  135, 155,  },
  }),
  ContextSetCfg::addCtxSet
  ({
    {   91, 141,  },
    {   60, 140,  },
    {  104, 139,  },
  }),
};

const CtxSet ContextSetCfg::SigFlag[] =
{
  ContextSetCfg::addCtxSet
  ({
    {  170, 154, 139, 153, 139, 123, 123,  63, 124, 166, 183, 140, 136, 153, 154, 166, 183, 140, 136, 153, 154, 166, 183, 140, 136, 153, 154, 140,  },
    {  155, 154, 139, 153, 139, 123, 123,  63, 153, 166, 183, 140, 136, 153, 154, 166, 183, 140, 136, 153, 154, 166, 183, 140, 136, 153, 154, 140,  },
    {  111, 111, 125, 110, 110,  94, 124, 108, 124, 107, 125, 141, 179, 153, 125, 107, 125, 141, 179, 153, 125, 107, 125, 141, 179, 153, 125, 141,  },
  }),
  ContextSetCfg::addCtxSet
  ({
    {  170, 153, 138, 138, 122, 121, 122, 121, 167, 151, 183, 140, 151, 183, 140, 140,  },
    {  170, 153, 123, 123, 107, 121, 107, 121, 167, 151, 183, 140, 151, 183, 140, 140,  },
    {  140, 139, 182, 182, 152, 136, 152, 136, 153, 136, 139, 111, 136, 139, 111, 111,  },
  }),
  ContextSetCfg::addCtxSet
  ({
    {  107, 139, 154, 140, 140, 141, 108, 154, 125, 155, 126, 127, 139, 155, 155, 141, 156, 143, 107, 139, 154, 140, 140, 141, 108, 154, 125, 155, 126, 127, 139, 155, 155, 141, 156, 143, 107, 139, 154, 140, 140, 141, 108, 154, 125, 155, 126, 127, 139, 155, 155, 141, 156, 143,  },
    {  121, 167, 153, 139, 154, 140, 137, 168, 139, 154, 169, 155, 167, 169, 169, 184, 199, 156, 121, 167, 153, 139, 154, 140, 137, 168, 139, 154, 169, 155, 167, 169, 169, 184, 199, 156, 121, 167, 153, 139, 154, 140, 137, 168, 139, 154, 169, 155, 167, 169, 169, 184, 199, 156,  },
    {  152, 139, 154, 154, 169, 155, 182, 154, 169, 184, 155, 141, 168, 214, 199, 170, 170, 171, 152, 139, 154, 154, 169, 155, 182, 154, 169, 184, 155, 141, 168, 214, 199, 170, 170, 171, 152, 139, 154, 154, 169, 155, 182, 154, 169, 184, 155, 141, 168, 214, 199, 170, 170, 171,  },
  }),
  ContextSetCfg::addCtxSet
  ({
    {  137, 154, 154, 155, 155, 156, 124, 185, 156, 171, 142, 158,  },
    {  136, 153, 139, 154, 125, 140, 122, 154, 184, 185, 171, 157,  },
    {  167, 154, 169, 140, 155, 141, 153, 171, 185, 156, 171, 172,  },
  }),
};

const CtxSet ContextSetCfg::LastX[] =
{
  ContextSetCfg::addCtxSet
  ({
    {  125, 110, 124, 110,  95,  94, 125, 111, 111,  79, 125, 126, 111, 111,  79, 126, 111, 111,  79, CNU, CNU, CNU, CNU, CNU, CNU,  },
    {  125, 110,  94, 110,  95,  79, 125, 111, 110,  78, 110, 111, 111,  95,  94, 111, 111,  95,  94, CNU, CNU, CNU, CNU, CNU, CNU,  },
    {  110, 110, 124, 125, 140, 153, 125, 127, 140, 109, 111, 143, 127, 111,  79, 143, 127, 111,  79, CNU, CNU, CNU, CNU, CNU, CNU,  },
  }),
  ContextSetCfg::addCtxSet
  ({
    {  108, 123,  93, 154,  },
    {  108, 123, 108, 154,  },
    {  108, 123,  63, 154,  },
  }),
};

const CtxSet ContextSetCfg::LastY[] =
{
  ContextSetCfg::addCtxSet
  ({
    {  125, 110, 124, 110,  95,  94, 125, 111, 111,  79, 125, 126, 111, 111,  79, 126, 111, 111,  79, CNU, CNU, CNU, CNU, CNU, CNU,  },
    {  125, 110,  94, 110,  95,  79, 125, 111, 110,  78, 110, 111, 111,  95,  94, 111, 111,  95,  94, CNU, CNU, CNU, CNU, CNU, CNU,  },
    {  110, 110, 124, 125, 140, 153, 125, 127, 140, 109, 111, 143, 127, 111,  79, 143, 127, 111,  79, CNU, CNU, CNU, CNU, CNU, CNU,  },
  }),
  ContextSetCfg::addCtxSet
  ({
    {  108, 123,  93, 154,  },
    {  108, 123, 108, 154,  },
    {  108, 123,  63, 154,  },
  }),
};

const CtxSet ContextSetCfg::GreaterOneFlag[] =
{
  ContextSetCfg::addCtxSet
  ({
    {  154, 196, 167, 167,  },
    {  154, 196, 196, 167,  },
    {  140,  92, 137, 138,  },
  }),
  ContextSetCfg::addCtxSet
  ({
    {  154, 152, 167, 182,  },
    {  154, 152, 167, 182,  },
    {  140, 152, 138, 139,  },
  }),
  ContextSetCfg::addCtxSet
  ({
    {  182, 134, 149, 136,  },
    {  182, 134, 149, 136,  },
    {  153,  74, 149,  92,  },
  }),
  ContextSetCfg::addCtxSet
  ({
    {  153, 121, 136, 122,  },
    {  153, 121, 136, 137,  },
    {  139, 107, 122, 152,  },
  }),
  ContextSetCfg::addCtxSet
  ({
    {  169, 208, 166, 167,  },
    {  169, 194, 166, 167,  },
    {  140, 179, 166, 182,  },
  }),
  ContextSetCfg::addCtxSet
  ({
    {  154, 152, 167, 182,  },
    {  154, 167, 137, 182,  },
    {  140, 227, 122, 197,  },
  }),
  ContextSetCfg::addCtxSet
  ({
    {  121, 135, 123, 124, 139, 125,  92, 124, 154, 125, 155, 138, 169, 155, 170, 156,  },
    {  165,  75, 152, 153, 139, 154, 121, 138, 139, 154, 140, 167, 183, 169, 170, 156,  },
    {  196, 105, 152, 153, 139, 154, 136, 138, 139, 169, 140, 196, 183, 169, 170, 171,  },
  }),
  ContextSetCfg::addCtxSet
  ({
    {  166, 152, 140, 170, 171, 157,  },
    {  193, 181, 169, 170, 171, 172,  },
    {  195, 181, 169, 170, 156, 157,  },
  }),
};

const CtxSet ContextSetCfg::GreaterTwoFlag = ContextSetCfg::addCtxSet
({
  {  107, 167,  91, 107, 107, 167,},
  {  107, 167,  91, 122, 107, 167,},
  {  138, 153, 136, 167, 152, 152,},
});

const CtxSet ContextSetCfg::MVPIdx = ContextSetCfg::addCtxSet
({
  {  168,},
  {  168,},
  {  CNU,},
});

const CtxSet ContextSetCfg::SaoMergeFlag = ContextSetCfg::addCtxSet
({
  {  153,},
  {  153,},
  {  153,},
});

const CtxSet ContextSetCfg::SaoTypeIdx = ContextSetCfg::addCtxSet
({
  {  160,},
  {  185,},
  {  200,},
});

#if JEM_TOOLS
const CtxSet ContextSetCfg::AlfCUCtrlFlags = ContextSetCfg::addCtxSet
({
  {  CNU, CNU, CNU,},
  {  CNU, CNU, CNU,},
  {  CNU, CNU, CNU,},
});

const CtxSet ContextSetCfg::AlfUvlcSCModel = ContextSetCfg::addCtxSet
({
  {  154, 140,},
  {  140, 110,},
  {  139, 139,},
});

#endif
const CtxSet ContextSetCfg::TransformSkipFlag = ContextSetCfg::addCtxSet
({
  {  139, 139,},
  {  139, 139,},
  {  139, 139,},
});

const CtxSet ContextSetCfg::TransquantBypassFlag = ContextSetCfg::addCtxSet
({
  {  154,},
  {  154,},
  {  154,},
});

#if JEM_TOOLS
const CtxSet ContextSetCfg::NSSTIdx = ContextSetCfg::addCtxSet
({
  {  107, 107, 107, 107, 107,},
  {  107, 107, 107, 107, 107,},
  {  139, 139, 139, 139, 139,},
});

#endif
const CtxSet ContextSetCfg::RdpcmFlag = ContextSetCfg::addCtxSet
({
  {  139, 139,},
  {  139, 139,},
  {  CNU, CNU,},
});

const CtxSet ContextSetCfg::RdpcmDir = ContextSetCfg::addCtxSet
({
  {  139, 139,},
  {  139, 139,},
  {  CNU, CNU,},
});

#if JEM_TOOLS
const CtxSet ContextSetCfg::EMTTuIndex = ContextSetCfg::addCtxSet
({
  {  CNU, CNU, CNU, CNU,},
  {  CNU, CNU, CNU, CNU,},
  {  CNU, CNU, CNU, CNU,},
});

const CtxSet ContextSetCfg::EMTCuFlag = ContextSetCfg::addCtxSet
({
  {  CNU, CNU, CNU, CNU, CNU, CNU,},
  {  CNU, CNU, CNU, CNU, CNU, CNU,},
  {  CNU, CNU, CNU, CNU, CNU, CNU,},
});

#endif
const CtxSet ContextSetCfg::CrossCompPred = ContextSetCfg::addCtxSet
({
  {  154, 154, 154, 154, 154, 154, 154, 154, 154, 154,},
  {  154, 154, 154, 154, 154, 154, 154, 154, 154, 154,},
  {  154, 154, 154, 154, 154, 154, 154, 154, 154, 154,},
});

const CtxSet ContextSetCfg::ChromaQpAdjFlag = ContextSetCfg::addCtxSet
({
  {  154,},
  {  154,},
  {  154,},
});

const CtxSet ContextSetCfg::ChromaQpAdjIdc = ContextSetCfg::addCtxSet
({
  {  154,},
  {  154,},
  {  154,},
});

#if JEM_TOOLS
const CtxSet ContextSetCfg::ImvFlag = ContextSetCfg::addCtxSet
({
  {  197, 185, 201, 185,},
  {  197, 185, 201, 185,},
  {  CNU, CNU, CNU, CNU,},
});

const CtxSet ContextSetCfg::LICFlag = ContextSetCfg::addCtxSet
({
  {  154,},
  {  154,},
  {  CNU,},
});

const CtxSet ContextSetCfg::ObmcFlag = ContextSetCfg::addCtxSet
({
  {  201,},
  {  201,},
  {  CNU,},
});

const CtxSet ContextSetCfg::FrucFlag = ContextSetCfg::addCtxSet
({
  {  197, 185, 201,},
  {  197, 185, 201,},
  {  CNU, CNU, CNU,},
});

const CtxSet ContextSetCfg::FrucMode = ContextSetCfg::addCtxSet
({
  {  197, 185, 201,},
  {  197, 185, 201,},
  {  CNU, CNU, CNU,},
});
#endif



const unsigned ContextSetCfg::NumberOfContexts = (unsigned)ContextSetCfg::sm_InitTables[0].size();


// combined sets
const CtxSet ContextSetCfg::Sao = { ContextSetCfg::SaoMergeFlag, ContextSetCfg::SaoTypeIdx };



template <class BinProbModel>
CtxStore<BinProbModel>::CtxStore()
  : m_CtxBuffer ()
  , m_Ctx       ( nullptr )
{}

template <class BinProbModel>
CtxStore<BinProbModel>::CtxStore( bool dummy )
  : m_CtxBuffer ( ContextSetCfg::NumberOfContexts )
  , m_Ctx       ( m_CtxBuffer.data() )
{}

template <class BinProbModel>
CtxStore<BinProbModel>::CtxStore( const CtxStore<BinProbModel>& ctxStore )
  : m_CtxBuffer ( ctxStore.m_CtxBuffer )
  , m_Ctx       ( m_CtxBuffer.data() )
{}

template <class BinProbModel>
void CtxStore<BinProbModel>::init( int qp, int initId )
{
  const std::vector<uint8_t>& initTable = ContextSetCfg::getInitTable( initId );
  CHECK( m_CtxBuffer.size() != initTable.size(),
        "Size of init table (" << initTable.size() << ") does not match size of context buffer (" << m_CtxBuffer.size() << ")." );
  int clippedQP = std::min( std::max( 0, qp ), MAX_QP );
  for( std::size_t k = 0; k < m_CtxBuffer.size(); k++ )
  {
    m_CtxBuffer[k].init( clippedQP, initTable[k] );
  }
}

template <class BinProbModel>
void CtxStore<BinProbModel>::setWinSizes( const std::vector<uint8_t>& log2WindowSizes )
{
  CHECK( m_CtxBuffer.size() != log2WindowSizes.size(),
        "Size of window size table (" << log2WindowSizes.size() << ") does not match size of context buffer (" << m_CtxBuffer.size() << ")." );
  for( std::size_t k = 0; k < m_CtxBuffer.size(); k++ )
  {
    m_CtxBuffer[k].setLog2WindowSize( log2WindowSizes[k] );
  }
}

template <class BinProbModel>
void CtxStore<BinProbModel>::loadPStates( const std::vector<uint16_t>& probStates )
{
  CHECK( m_CtxBuffer.size() != probStates.size(),
        "Size of prob states table (" << probStates.size() << ") does not match size of context buffer (" << m_CtxBuffer.size() << ")." );
  for( std::size_t k = 0; k < m_CtxBuffer.size(); k++ )
  {
    m_CtxBuffer[k].setState( probStates[k] );
  }
}

template <class BinProbModel>
void CtxStore<BinProbModel>::savePStates( std::vector<uint16_t>& probStates ) const
{
  probStates.resize( m_CtxBuffer.size(), uint16_t(0) );
  for( std::size_t k = 0; k < m_CtxBuffer.size(); k++ )
  {
    probStates[k] = m_CtxBuffer[k].getState();
  }
}





template class CtxStore<BinProbModel_Std>;
#if JEM_TOOLS
template class CtxStore<BinProbModel_JMP>;
template class CtxStore<BinProbModel_JAW>;
template class CtxStore<BinProbModel_JMPAW>;
#endif





Ctx::Ctx()                                  : m_BPMType( BPM_Undefined )                        {}
Ctx::Ctx( const BinProbModel_Std*   dummy ) : m_BPMType( BPM_Std   ), m_CtxStore_Std  ( true )  {}
#if JEM_TOOLS
Ctx::Ctx( const BinProbModel_JMP*   dummy ) : m_BPMType( BPM_JMP   ), m_CtxStore_JMP  ( true )  {}
Ctx::Ctx( const BinProbModel_JAW*   dummy ) : m_BPMType( BPM_JAW   ), m_CtxStore_JAW  ( true )  {}
Ctx::Ctx( const BinProbModel_JMPAW* dummy ) : m_BPMType( BPM_JMPAW ), m_CtxStore_JMPAW( true )  {}
#endif

Ctx::Ctx( const Ctx& ctx )
  : m_BPMType         ( ctx.m_BPMType )
  , m_CtxStore_Std    ( ctx.m_CtxStore_Std    )
#if JEM_TOOLS
  , m_CtxStore_JMP    ( ctx.m_CtxStore_JMP    )
  , m_CtxStore_JAW    ( ctx.m_CtxStore_JAW    )
  , m_CtxStore_JMPAW  ( ctx.m_CtxStore_JMPAW  )
#endif
#if HM_STORE_FRAC_BITS_AND_USE_ROUNDED_BITS
  , m_FracBitsStore   ( ctx.m_FracBitsStore   )
#endif
{
  ::memcpy( m_GRAdaptStats, ctx.m_GRAdaptStats, sizeof( unsigned ) * RExt__GOLOMB_RICE_ADAPTATION_STATISTICS_SETS );
}

#if JEM_TOOLS

CtxWSizeStore::CtxWSizeStore()
  : m_isInitialized   ( false )
  , m_readWriteBuffer ()
  , m_codeId2ctxId    ()
{
}

void CtxWSizeStore::checkInit( const SPS* sps )
{
  if( !m_isInitialized )
  {
#if JEM_ADAPTIVE_WINDOW_MECHANISM
    const int readWriteBufferSize = 384;
#else
    const int readWriteBufferSize = Ctx::NumberOfContexts;
#endif
    m_readWriteBuffer.resize( readWriteBufferSize, 0 );
    xInitMappingTable       ( sps );
    xSetAllInvalid          ();
    m_isInitialized       = true;
  }
}

void CtxWSizeStore::updateState( const Slice* slice, const bool enc )
{
#if JEM_ADAPTIVE_WINDOW_MECHANISM
#else // bug fix : corrected handling of random access points
  if ( slice->getPendingRasInit() )
  {
    xSetAllInvalid();
  }
  CHECK( slice->getRapPicFlag() && !enc && slice->getCabacWinUpdateMode() == 2, "reusing of cabac window sizes not possible at random access point" );
#endif
  if( !enc && slice->getCabacWinUpdateMode() == 1 )
  {
    xApplyReadWriteBuffer( slice );
  }
}

bool CtxWSizeStore::validWinSizes( const Slice* slice ) const
{
  return m_winSizes[slice->getSliceType()][slice->getSliceQpBase()].isValid();
}

const std::vector<uint8_t>* CtxWSizeStore::getWinSizes( const Slice* slice ) const
{
  if( !slice->getCabacWinUpdateMode() )
  {
    return 0;
  }
  const CtxWSizeSet& set = m_winSizes[slice->getSliceType()][slice->getSliceQpBase()];
  CHECK( !set.isValid(), "win size buffer is invalid" );
  return &set.getWinSizeBuffer();
}

void CtxWSizeStore::xSetAllInvalid()
{
  for( unsigned sliceType = 0; sliceType < NUMBER_OF_SLICE_TYPES; sliceType++ )
  {
    CtxWSizeSet* winSizes = m_winSizes[sliceType];
    for( unsigned QP = 0; QP <= MAX_QP; QP++ )
    {
      winSizes[QP].setInvalid();
    }
  }
}

void CtxWSizeStore::xApplyReadWriteBuffer( const Slice* slice )
{
  const uint8_t         defWinSize  = Ctx::getDefaultWindowSize( slice->getSPS()->getSpsNext().getCABACEngineMode() );
  const uint8_t         minWinSize  = 4;
  const uint8_t         threshold   = uint8_t( defWinSize - minWinSize + 1 );
  CtxWSizeSet&          set         = m_winSizes[slice->getSliceType()][slice->getSliceQpBase()];
  std::vector<uint8_t>& winBuffer   = set.getWinSizeBuffer();
  winBuffer.resize( 0 );
  winBuffer.resize( Ctx::NumberOfContexts, 0 );
  for( std::size_t readId = 0; readId < m_readWriteBuffer.size(); readId++ )
  {
    uint8_t   level   = m_readWriteBuffer[readId];
    if( level )
    {
      uint8_t winSize = minWinSize + ( level < threshold ? level - 1 : level );
      winBuffer[m_codeId2ctxId[readId]] = winSize;
    }
  }
  set.setValidOnly();
}

void CtxWSizeStore::xSetReadWriteBuffer( const Slice* slice )
{
  const uint8_t         defWinSize  = Ctx::getDefaultWindowSize( slice->getSPS()->getSpsNext().getCABACEngineMode() );
  const uint8_t         minWinSize  = 4;
  CtxWSizeSet&          set         = m_winSizes[slice->getSliceType()][slice->getSliceQpBase()];
  CHECK( !set.isValid(), "requested adap. window statistics are invalid" );
  std::vector<uint8_t>& winBuffer   = set.getWinSizeBuffer();
  for( std::size_t codeId = 0; codeId < m_readWriteBuffer.size(); codeId++ )
  {
    uint8_t level = 0;
    int     ctxId = m_codeId2ctxId[codeId];
    if( ctxId >= 0 )
    {
      uint8_t winSize  = winBuffer[ctxId];
      if(     winSize != defWinSize )
      {
        level = 1 + winSize - minWinSize - ( winSize < defWinSize ? 0 : 1 );
      }
    }
    m_readWriteBuffer[codeId] = level;
  }
}

void addCtxSetToMapping( std::vector<int>& mapping, const CtxSet& ctxSet, unsigned startId, unsigned numCtx )
{
  int ctxOffset    = ctxSet(startId);
  int numActualCtx = std::min<int>( (int)numCtx, (int)ctxSet.Size );
  for( int n = 0; n < numActualCtx; n++ )
  {
    mapping.push_back( ctxOffset + n );
  }
  for( int n = numActualCtx; n < (int)numCtx; n++ )
  {
    mapping.push_back( -1 );
  }
}

void addDummySetToMapping( std::vector<int>& mapping, unsigned numCtx )
{
  for( int n = 0; n < (int)numCtx; n++ )
  {
    mapping.push_back( -1 );
  }
}

void CtxWSizeStore::xInitMappingTable( const SPS* sps )
{
#if JEM_ADAPTIVE_WINDOW_MECHANISM
  m_codeId2ctxId.reserve( std::max<std::size_t>( m_readWriteBuffer.size(), std::size_t(Ctx::NumberOfContexts) ) );
  const SPSNext&  spsNext             = sps->getSpsNext();
  unsigned        numCtxSplitFlag     = ( spsNext.getUseLargeCTU() ? 5 : 3 ); // hard-wired in JEM
  unsigned        numCtxMergeIdx      = ( true /* COM16_C806_GEN_MRG_IMPROVEMENT */ ? 5 : 1 );
  unsigned        numCtxIPredLuma     = ( spsNext.getUseIntra65Ang() ? ( false /*JVET_B0051_NON_MPM_MODE */ ? 12 : 9 ) : 1 ); // hard-wired in JEM
  unsigned        numCtxIPredChroma   = ( true /* JVET_E0077_ENHANCED_LM */ ? 8 : 2 ) + ( spsNext.getUseMDMS() ? 4 : 0 );
  unsigned        numCtxTransSubdiv   = ( spsNext.getUseLargeCTU() || spsNext.getUseQTBT() ? 4 : 3 ); // hard-wired in JEM
  unsigned        numCtxLastXY        = ( spsNext.getUseLargeCTU() || spsNext.getUseQTBT() ? 25 : 15 ); // hard-wired in JEM
  unsigned        numCtxEmtCUFlag     = ( spsNext.getUseLargeCTU() || spsNext.getUseQTBT() ? 6 : 4 ); // hard-wired in JEM
#if JEM_TOOLS
  unsigned        numImvCtx           = ( spsNext.getImvMode() > 1 ? 4 : 3 );
#endif
  addCtxSetToMapping      ( m_codeId2ctxId, Ctx::SplitFlag,              0, numCtxSplitFlag );
  if( spsNext.getUseQTBT() )  // hard-wired in JEM
  {
    addCtxSetToMapping    ( m_codeId2ctxId, Ctx::BTSplitFlag,            0, 6 );
  }
  addCtxSetToMapping      ( m_codeId2ctxId, Ctx::SkipFlag,               0, 3 );
  if( false /* VCEG_AZ05_INTRA_MPI */ )
  {
    // dummy placeholder
  }
  if( true /* COM16_C1046_PDPC_INTRA */ )
  {
    addCtxSetToMapping    ( m_codeId2ctxId, Ctx::PdpcFlag,               0, 2 );
  }
  if( true /* VCEG_AZ05_ROT_TR || COM16_C1044_NSST */ )
  {
    addCtxSetToMapping    ( m_codeId2ctxId, Ctx::NSSTIdx,                0, 7 );
  }
  addCtxSetToMapping      ( m_codeId2ctxId, Ctx::MergeFlag,              0, 1 );
  addCtxSetToMapping      ( m_codeId2ctxId, Ctx::MergeIdx,               0, numCtxMergeIdx );
#if JEM_TOOLS
  if( true /* VCEG_AZ07_FRUC_MERGE */ )
  {
    addCtxSetToMapping    ( m_codeId2ctxId, Ctx::FrucFlag,               0, 3 );
    addCtxSetToMapping    ( m_codeId2ctxId, Ctx::FrucMode,               0, 3 );
  }
#endif
  addCtxSetToMapping      ( m_codeId2ctxId, Ctx::PartSize,               0, 4 );
  addCtxSetToMapping      ( m_codeId2ctxId, Ctx::PredMode,               0, 1 );
  addCtxSetToMapping      ( m_codeId2ctxId, Ctx::IPredMode[0],           0, numCtxIPredLuma );
  if( spsNext.getUseMDMS() ) // hard wired in JEM
  {
    addCtxSetToMapping    ( m_codeId2ctxId, Ctx::IPredMode[1],           0, numCtxIPredChroma );
  }
  else
  {
    addCtxSetToMapping    ( m_codeId2ctxId, Ctx::IPredMode[1],           1, numCtxIPredChroma );
  }
  addCtxSetToMapping      ( m_codeId2ctxId, Ctx::DeltaQP,                0, 3 );
  addCtxSetToMapping      ( m_codeId2ctxId, Ctx::InterDir,               0, 5 );
  addCtxSetToMapping      ( m_codeId2ctxId, Ctx::RefPic,                 0, 2 );
  addCtxSetToMapping      ( m_codeId2ctxId, Ctx::Mvd,                    0, 2 );
  addCtxSetToMapping      ( m_codeId2ctxId, Ctx::QtCbf[0],               0, 5 );
  addCtxSetToMapping      ( m_codeId2ctxId, Ctx::QtCbf[1],               0, 5 );
  addCtxSetToMapping      ( m_codeId2ctxId, Ctx::TransSubdivFlag,        0, numCtxTransSubdiv );
  addCtxSetToMapping      ( m_codeId2ctxId, Ctx::QtRootCbf,              0, 1 );
  if( spsNext.getAltResiCompId() == 0 )  // hard-wired in JEM
  {
    addCtxSetToMapping    ( m_codeId2ctxId, Ctx::SigCoeffGroup[0],       0, 2 );
    addCtxSetToMapping    ( m_codeId2ctxId, Ctx::SigCoeffGroup[1],       0, 2 );
    addCtxSetToMapping    ( m_codeId2ctxId, Ctx::SigFlag[0],             0, 28 );
    addCtxSetToMapping    ( m_codeId2ctxId, Ctx::SigFlag[1],             0, 16 );
  }
  else if( spsNext.getAltResiCompId() == 1 )  // hard-wired in JEM
  {
    addCtxSetToMapping    ( m_codeId2ctxId, Ctx::SigCoeffGroup[2],       0, 2 );
    addCtxSetToMapping    ( m_codeId2ctxId, Ctx::SigCoeffGroup[3],       0, 2 );
    addCtxSetToMapping    ( m_codeId2ctxId, Ctx::SigFlag[2],             0, 54 );
    addCtxSetToMapping    ( m_codeId2ctxId, Ctx::SigFlag[3],             0, 12 );
  }
  addCtxSetToMapping      ( m_codeId2ctxId, Ctx::LastX[0],               0, numCtxLastXY );
  addCtxSetToMapping      ( m_codeId2ctxId, Ctx::LastX[1],               0, numCtxLastXY );
  addCtxSetToMapping      ( m_codeId2ctxId, Ctx::LastY[0],               0, numCtxLastXY );
  addCtxSetToMapping      ( m_codeId2ctxId, Ctx::LastY[1],               0, numCtxLastXY );
  if( spsNext.getAltResiCompId() == 0 )  // hard-wired in JEM
  {
    addCtxSetToMapping    ( m_codeId2ctxId, Ctx::GreaterOneFlag[0],      0, 4 );
    addCtxSetToMapping    ( m_codeId2ctxId, Ctx::GreaterOneFlag[1],      0, 4 );
    addCtxSetToMapping    ( m_codeId2ctxId, Ctx::GreaterOneFlag[2],      0, 4 );
    addCtxSetToMapping    ( m_codeId2ctxId, Ctx::GreaterOneFlag[3],      0, 4 );
    addCtxSetToMapping    ( m_codeId2ctxId, Ctx::GreaterOneFlag[4],      0, 4 );
    addCtxSetToMapping    ( m_codeId2ctxId, Ctx::GreaterOneFlag[5],      0, 4 );
    addCtxSetToMapping    ( m_codeId2ctxId, Ctx::GreaterTwoFlag,         0, 6 );
  }
  else if( spsNext.getAltResiCompId() == 1 )  // hard-wired in JEM
  {
    addCtxSetToMapping    ( m_codeId2ctxId, Ctx::GreaterOneFlag[6],      0, 16 );
    addCtxSetToMapping    ( m_codeId2ctxId, Ctx::GreaterOneFlag[7],      0, 6 );
  }
  addCtxSetToMapping      ( m_codeId2ctxId, Ctx::MVPIdx,                 0, 1 );
  addCtxSetToMapping      ( m_codeId2ctxId, Ctx::SaoMergeFlag,           0, 1 );
  addCtxSetToMapping      ( m_codeId2ctxId, Ctx::SaoTypeIdx,             0, 1 );
  addCtxSetToMapping      ( m_codeId2ctxId, Ctx::TransformSkipFlag,      0, 2 );
  if( false /* VCEG_AZ08_KLT_COMMON */ )
  {
    // dummy placeholder -- MTT contexts, not used in JEM compatibility modus
#if JEM_COMP
    addCtxSetToMapping    ( m_codeId2ctxId, Ctx::BTSplitFlag,            6, 2 );
#endif
  }
  addCtxSetToMapping      ( m_codeId2ctxId, Ctx::TransquantBypassFlag,   0, 1 );
  addCtxSetToMapping      ( m_codeId2ctxId, Ctx::RdpcmFlag,              0, 2 );
  addCtxSetToMapping      ( m_codeId2ctxId, Ctx::RdpcmDir,               0, 2 );
  addCtxSetToMapping      ( m_codeId2ctxId, Ctx::CrossCompPred,          0, 10 );
  addCtxSetToMapping      ( m_codeId2ctxId, Ctx::ChromaQpAdjFlag,        0, 1 );
  addCtxSetToMapping      ( m_codeId2ctxId, Ctx::ChromaQpAdjIdc,         0, 1 );
#if JEM_TOOLS
  if( true /* COM16_C806_OBMC */ )
  {
    addCtxSetToMapping    ( m_codeId2ctxId, Ctx::ObmcFlag,               0, 1 );
  }
  if( true /* VCEG_AZ07_IMV */ )
  {
    addCtxSetToMapping    ( m_codeId2ctxId, Ctx::ImvFlag,                0, numImvCtx );
  }
  if( spsNext.getLICEnabled() ) // hard-wired in JEM
  {
    addCtxSetToMapping    ( m_codeId2ctxId, Ctx::LICFlag,                0, 1 );
  }
#endif
  if( true /* ALF_HM3_REFACTOR */ )
  {
    if( spsNext.getGALFEnabled() )
    {
      addCtxSetToMapping  ( m_codeId2ctxId, Ctx::AlfCUCtrlFlags,         0, 1 );
      addCtxSetToMapping  ( m_codeId2ctxId, Ctx::AlfUvlcSCModel,         0, 2 );
    }
    else
    {
      addCtxSetToMapping  ( m_codeId2ctxId, Ctx::AlfCUCtrlFlags,         0, 3 );
      addDummySetToMapping( m_codeId2ctxId,                                 1 );
      addCtxSetToMapping  ( m_codeId2ctxId, Ctx::AlfUvlcSCModel,         0, 2 );
      addDummySetToMapping( m_codeId2ctxId,                                 3 );
    }
  }
  if( true /* COM16_C806_EMT */ )
  {
    addCtxSetToMapping    ( m_codeId2ctxId, Ctx::EMTTuIndex,             0, 4 );
    addCtxSetToMapping    ( m_codeId2ctxId, Ctx::EMTCuFlag,              0, numCtxEmtCUFlag );
  }
#if JEM_TOOLS
  if( true /* COM16_C1016_AFFINE */ )
  {
    addCtxSetToMapping    ( m_codeId2ctxId, Ctx::AffineFlag,             0, 3 );
  }
#endif
  if( spsNext.getUseQTBT() )
  {
    addCtxSetToMapping    ( m_codeId2ctxId, Ctx::BTSplitFlag,            6, Ctx::BTSplitFlag.Size-6 );
  }
  m_codeId2ctxId.resize   ( m_readWriteBuffer.size(), -1 );
#else
  m_codeId2ctxId.resize   ( Ctx::NumberOfContexts );
  for( std::size_t n = 0; n < Ctx::NumberOfContexts; n++ )
  {
    m_codeId2ctxId[n] = (int)n;
  }
#endif
}
#endif
